/*
 * Copyright (c) 2010 LAAS/CNRS                       --  Oct 2010
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
/*------------------------------------------------------------
 * 
 *                     --  Module  LWR  --
 * 
 *  Description:
 *  Creation date : Wed May 19 22:35:29 2010 
 *  Author: Xavier Broquere
 *  LAAS-CNRS, Toulouse FRANCE
 *------------------------------------------------------------*/

/**
 ** lwrCmdCodels.cc
 **
 ** Codels called by execution task lwrCmd
 **
 ** Author: 
 ** Date: 
 **
 **/

#ifdef LWR_FRI
#include "kukaFri-libs/friremote.h"
#endif
#include "server/lwrHeader.h"
extern "C" {
#include "Proto_gb.h"
#include "Proto_gbModeles.h"   // LWR MGI
}
#include "libeuler/pomEuler.h"
#include "coldman.h"
#include "softMotion/softMotion.h"
#include "softMotion/softMotionStruct.h"
#include "softMotion/softMotionStructGenom.h"
#include <pthread.h>
#include <sys/time.h>

#include <taskLib.h>
#include <h2timerLib.h>

#include "lwrConst.h"
#include "lwrFunctions.h"



#ifdef LWR_FRI
friRemote * friInstLeft = NULL;
friRemote * friInstRight = NULL;
//pthread_t threadFriComm = NULL;
int lwrGetArmInf(friRemote *friInst, LWR_ARM_INST *arm, char* armName);
int lwrGetInf();
int lwrDoDataExchange();
#endif

static int LWR_JOINTS_USER;
static int LWR_CART_USER;
bool initIsDone = false;
static bool closingConnect = false;
LWR_MOTION_STR currentMotion;

GENMANIP_POSITION_ARM_SAHAND_TRACK_STR  lwrPositionsPoster;
SM_TRAJ_STR smTraj;
float initialCartPos[LWR_CART_FRM_DIM];
static POSTER_ID lwr_poster_q = NULL;
static POSTER_ID posterMhpCollideId = NULL;
static POSTER_ID posterSAHandId = NULL;
static POSTER_ID posterHumDistId = NULL;

FILE* fileptr = NULL;
STATUS lwrCmd();
int lwrMonitorAndUpdateCntrlMode(LWR_ARM_INST *armInst, char* message) ;


#define MAX(a,b) ((a)>(b)?(a):(b))

#ifdef LWR_FRI
static H2TIMER_ID threadFriCommTimerId;

int threadFriCommCode() {
  if (h2timerPause(threadFriCommTimerId) != OK) {
    printf("threadFriCommCode: h2timerPause error\n");
    //lwrCmdSuspend (FALSE);
  }
  if(SDI_F->status.simulation == GEN_TRUE) {    
    lwrCmd();
  } else {
    if(closingConnect) {  //?? move to reat time
      return OK;
    }
    if ((SDI_F->armLeft.is_used == GEN_TRUE) && (friInstLeft == NULL)) {
      return OK;
    }
    if((SDI_F->armRight.is_used == GEN_TRUE) && (friInstRight == NULL)) {
      return OK;
    }
    lwrCmd();
    lwrDoDataExchange();
  }
  // WARNING  names ("R7ArmLeft") come from jido.geom for POM...
  lwrUpdateArmStatus(friInstLeft , &SDI_F->armLeft , &SDI_F->pomArmLeftPos , (char *)"R7ArmLeft"); //?? move to reat time
  lwrUpdateArmStatus(friInstRight, &SDI_F->armRight, &SDI_F->pomArmRightPos, (char *)"R7ArmRight");//?? move to reat time
  lwrGetInf();  //?? move real time
  lwrUpdateMotion();  //?? move real time
  return 0;
}

void *threadFriCommFct(void *) {
  do{
    threadFriCommCode();
  } while(closingConnect == false);
  printf("threadFriComm comm exited\n");
  return 0;
}
#endif

int lwrGetInf() {
#ifdef LWR_FRI
  /************************/
  /* Info about LEFT ARM  */
  /************************/
  if(SDI_F->armLeft.is_used == GEN_TRUE && friInstLeft != NULL) {
    lwrGetArmInf(friInstLeft, &(SDI_F->armLeft), (char *)"Left");
  }    
  /*************************/
  /* Info about RIGHT ARM  */
  /*************************/
  if(SDI_F->armRight.is_used ==  GEN_TRUE && friInstRight != NULL) {
    lwrGetArmInf(friInstRight, &(SDI_F->armRight), (char *)"Right");
  }
#endif
  return 0;
}


int lwrGenerateDigestStatus() {
#ifdef LWR_FRI
  // Prepare an human-readable resume of robot status and configuration
  if(SDI_F->armLeft.is_used == GEN_TRUE && SDI_F->armRight.is_used == GEN_TRUE ) {
    snprintf(SDI_F->status.status_str.str, 1024,
	     "Simulation:\t%s\n"			    \
	     "Arm power:\tLeft: %s\tRight: %s\n"		\
	     "FRI status:\tLeft: %s\tRight: %s\n"		\
	     "Control scheme:\tLeft: %s\tRight: %s\n"	\
	     "Link quality:\tLeft: %s\tRight:%s\n"		     \
	     "Configuration\n\t"					      \
	     "q1:\t%3.2f\n\tq2:\t%3.2f\n\tq3:\t%3.2f\n\tq4:\t"		\
	     "%3.2f\n\tq5:\t%3.2f\n\tq6:\t%3.2f\n\tq7:\t%3.2f\n"        \
	     "Fx: %4.2f Fy: %4.2f Fz: %4.2f Mx: %4.2f My: %4.2f Mz: %4.2f",
	     SDI_F->status.simulation?"yes":"no",
	     SDI_F->armLeft.power_on?"on":"off",SDI_F->armRight.power_on?"on":"off",
	     friStateStr(SDI_F->armLeft.fri_state), friStateStr(SDI_F->armRight.fri_state),
	     friSchemeStr(SDI_F->armLeft.fri_scheme), friSchemeStr(SDI_F->armRight.fri_scheme),
	     friQualityStr(SDI_F->armLeft.fri_quality),friQualityStr(SDI_F->armRight.fri_quality),
	     rad2deg(SDI_F->motion. q[0]), rad2deg(SDI_F->motion. q[1]),
	     rad2deg(SDI_F->motion. q[2]), rad2deg(SDI_F->motion. q[3]),
	     rad2deg(SDI_F->motion. q[4]), rad2deg(SDI_F->motion. q[5]),
	     rad2deg(SDI_F->motion. q[6]),
	     SDI_F->armLeft.estExtTcpFT[0], SDI_F->armLeft.estExtTcpFT[1], SDI_F->armLeft.estExtTcpFT[2],
	     SDI_F->armLeft.estExtTcpFT[3], SDI_F->armLeft.estExtTcpFT[4], SDI_F->armLeft.estExtTcpFT[5]);



  } else if( SDI_F->armLeft.is_used == GEN_TRUE ) {
    snprintf(SDI_F->status.status_str.str, 1024,
	     "Simulation:\t%s\n"			    \
	     "Arm power:\tLeft: %s\n"	    \
	     "FRI status:\tLeft: %s\n"		\
	     "Control scheme:\tLeft: %s\n"	\
	     "Link quality:\tLeft: %s\n"		     \
	     "Configuration\n\t"					      \
	     "q1:\t%3.2f\n\tq2:\t%3.2f\n\tq3:\t%3.2f\n\tq4:\t"		\
	     "%3.2f\n\tq5:\t%3.2f\n\tq6:\t%3.2f\n\tq7:\t%3.2f\n"        \
	     "Fx: %4.2f Fy: %4.2f Fz: %4.2f Mx: %4.2f My: %4.2f Mz: %4.2f",
	     SDI_F->status.simulation?"yes":"no",
	     SDI_F->armLeft.power_on?"on":"off",
	     friStateStr(SDI_F->armLeft.fri_state),
	     friSchemeStr(SDI_F->armLeft.fri_scheme),
	     friQualityStr(SDI_F->armLeft.fri_quality),
	     rad2deg(SDI_F->motion. q[0]), rad2deg(SDI_F->motion. q[1]),
	     rad2deg(SDI_F->motion. q[2]), rad2deg(SDI_F->motion. q[3]),
	     rad2deg(SDI_F->motion. q[4]), rad2deg(SDI_F->motion. q[5]),
	     rad2deg(SDI_F->motion. q[6]),
	     SDI_F->armLeft.estExtTcpFT[0], SDI_F->armLeft.estExtTcpFT[1], SDI_F->armLeft.estExtTcpFT[2],
	     SDI_F->armLeft.estExtTcpFT[3], SDI_F->armLeft.estExtTcpFT[4], SDI_F->armLeft.estExtTcpFT[5]);
  } else if ( SDI_F->armRight.is_used == GEN_TRUE ) {
    snprintf(SDI_F->status.status_str.str, 1024,
	     "Simulation:\t%s\n"			    \
	     "Arm power:\tRight: %s\n"	    \
	     "FRI status:\tRight: %s\n"		\
	     "Control scheme:\tRight: %s\n"	\
	     "Link quality:\tRight: %s\n"		     \
	     "Configuration\n\t"					      \
	     "q1:\t%3.2f\n\tq2:\t%3.2f\n\tq3:\t%3.2f\n\tq4:\t"		\
	     "%3.2f\n\tq5:\t%3.2f\n\tq6:\t%3.2f\n\tq7:\t%3.2f\n"        \
	     "Fx: %4.2f Fy: %4.2f Fz: %4.2f Mx: %4.2f My: %4.2f Mz: %4.2f",
	     SDI_F->status.simulation?"yes":"no",
	     SDI_F->armRight.power_on?"on":"off",
	     friStateStr(SDI_F->armRight.fri_state),
	     friSchemeStr(SDI_F->armRight.fri_scheme),
	     friQualityStr(SDI_F->armRight.fri_quality),
	     rad2deg(SDI_F->motion. q[LWR_JOINTS+0]), rad2deg(SDI_F->motion. q[LWR_JOINTS+1]),
	     rad2deg(SDI_F->motion. q[LWR_JOINTS+2]), rad2deg(SDI_F->motion. q[LWR_JOINTS+3]),
	     rad2deg(SDI_F->motion. q[LWR_JOINTS+4]), rad2deg(SDI_F->motion. q[LWR_JOINTS+5]),
	     rad2deg(SDI_F->motion. q[LWR_JOINTS+6]),
	     SDI_F->armRight.estExtTcpFT[0], SDI_F->armRight.estExtTcpFT[1], SDI_F->armRight.estExtTcpFT[2],
	     SDI_F->armRight.estExtTcpFT[3], SDI_F->armRight.estExtTcpFT[4], SDI_F->armRight.estExtTcpFT[5]);
  }
#else
  if(SDI_F->armLeft.is_used == GEN_TRUE || SDI_F->armRight.is_used == GEN_TRUE ) {
    // Prepare an human-readable resume of robot status and configuration
    snprintf(SDI_F->status.status_str.str, 1024,
	     "Simulation:\t%s\n"			    \
	     "Configuration armLeft:\n\t"					\
	     "q1:\t%3.2f\n\tq2:\t%3.2f\n\tq3:\t%3.2f\n\tq4:\t"		\
	     "%3.2f\n\tq5:\t%3.2f\n\tq6:\t%3.2f\n\tq7:\t%3.2f\n"
	     "Configuration armRight:\n\t"					\
	     "q1:\t%3.2f\n\tq2:\t%3.2f\n\tq3:\t%3.2f\n\tq4:\t"		\
	     "%3.2f\n\tq5:\t%3.2f\n\tq6:\t%3.2f\n\tq7:\t%3.2f\n",
	     "yes",
	     rad2deg(SDI_F->motion. q[0]), rad2deg(SDI_F->motion. q[1]),
	     rad2deg(SDI_F->motion. q[2]), rad2deg(SDI_F->motion. q[3]),
	     rad2deg(SDI_F->motion. q[4]), rad2deg(SDI_F->motion. q[5]),
	     rad2deg(SDI_F->motion. q[6]),
	     rad2deg(SDI_F->motion. q[LWR_JOINTS+0]), rad2deg(SDI_F->motion. q[LWR_JOINTS+1]),
	     rad2deg(SDI_F->motion. q[LWR_JOINTS+2]), rad2deg(SDI_F->motion. q[LWR_JOINTS+3]),
	     rad2deg(SDI_F->motion. q[LWR_JOINTS+4]), rad2deg(SDI_F->motion. q[LWR_JOINTS+5]),
	     rad2deg(SDI_F->motion. q[LWR_JOINTS+6])
	     );
  } else if( SDI_F->armLeft.is_used == GEN_TRUE ) {
    snprintf(SDI_F->status.status_str.str, 1024,
	     "Simulation:\t%s\n"			    \
	     "Configuration arm Left\n\t"					\
	     "q1:\t%3.2f\n\tq2:\t%3.2f\n\tq3:\t%3.2f\n\tq4:\t"		\
	     "%3.2f\n\tq5:\t%3.2f\n\tq6:\t%3.2f\n\tq7:\t%3.2f\n",
	     "yes",
	     rad2deg(SDI_F->motion. q[0]), rad2deg(SDI_F->motion. q[1]),
	     rad2deg(SDI_F->motion. q[2]), rad2deg(SDI_F->motion. q[3]),
	     rad2deg(SDI_F->motion. q[4]), rad2deg(SDI_F->motion. q[5]),
	     rad2deg(SDI_F->motion. q[6]));

  } else if ( SDI_F->armRight.is_used == GEN_TRUE ) {
    snprintf(SDI_F->status.status_str.str, 1024,
	     "Simulation:\t%s\n"			    \
	     "Configuration armRight\n\t"					\
	     "q1:\t%3.2f\n\tq2:\t%3.2f\n\tq3:\t%3.2f\n\tq4:\t"		\
	     "%3.2f\n\tq5:\t%3.2f\n\tq6:\t%3.2f\n\tq7:\t%3.2f\n",
	     "yes",
	     rad2deg(SDI_F->motion. q[LWR_JOINTS+0]), rad2deg(SDI_F->motion. q[LWR_JOINTS+1]),
	     rad2deg(SDI_F->motion. q[LWR_JOINTS+2]), rad2deg(SDI_F->motion. q[LWR_JOINTS+3]),
	     rad2deg(SDI_F->motion. q[LWR_JOINTS+4]), rad2deg(SDI_F->motion. q[LWR_JOINTS+5]),
	     rad2deg(SDI_F->motion. q[LWR_JOINTS+6]));
  }
#endif
  return 0;
}



#ifdef LWR_FRI

friRemote * getFriRemote(LWR_ARM_ID_ENUM arm_id) {
  if(arm_id == LWR_ARM_LEFT) {
    return friInstLeft;
  } else if (arm_id == LWR_ARM_RIGHT) {
    return friInstRight;
  } else {
    printf("ERROR: getFriRemote wrong arm id: %d\n", arm_id);
    return NULL;
  }
}

int lwrDoDataExchange() {
  bool flagLeft = false, flagRight = false, flag = false;
  int sockLeft = 0, sockRight = 0;
  fd_set rfds;
  int retval = -1;
     
  if(friInstLeft == NULL && SDI_F->armLeft.is_used == GEN_TRUE) {
    return 0;
  }
  if( friInstRight == NULL && SDI_F->armRight.is_used == GEN_TRUE ) {
    return 0;
  }
  if( SDI_F->armLeft.is_used == GEN_TRUE &&  SDI_F->armRight.is_used == GEN_TRUE) {
    do{    
      FD_ZERO(&rfds);
      FD_SET(SDI_F->armLeft.sock, &rfds);
      FD_SET(SDI_F->armRight.sock, &rfds);    
      retval = select( MAX(SDI_F->armLeft.sock, SDI_F->armRight.sock)+1 , &rfds, NULL, NULL, NULL);
      if (retval == -1) {
	perror("select()");
      } else if (retval) {
	if(FD_ISSET(SDI_F->armLeft.sock, &rfds)){
	  flagLeft = true;
	  friInstLeft->doDataExchange();
	}
	if(FD_ISSET(SDI_F->armRight.sock, &rfds)){
	  flagRight = true;
	  friInstRight->doDataExchange();
	}
      }
    } while(flagLeft==false || flagRight==false );

  } else if ( SDI_F->armLeft.is_used == GEN_TRUE) {
    friInstLeft->doDataExchange();
  } else if ( SDI_F->armRight.is_used == GEN_TRUE) {
    friInstRight->doDataExchange();
  } else {
    printf("WARNING: lwrDoDataExchange: Any arm is used\n");
  }
  return 0;
}

int lwrGetArmInf(friRemote *friInst, LWR_ARM_INST *arm, char* armName) {
  GEN_BOOL power_on;
  // Arm power on/off
  power_on = friInst->isPowerOn()?GEN_TRUE:GEN_FALSE;
  if (arm->power_on != power_on) {
    if (friInst->isPowerOn() == false) {
      printf("WARN: Arm %s power is off\n",armName);
    }
    arm->power_on = power_on;
  }
  //  FRI mode monitor/command
  if (arm->fri_state != friInst->getState()) {
    if (friInst->getState() != FRI_STATE_CMD) {
      printf("WARN: FRI %s is in monitor mode\n",armName);
    }
    arm->fri_state = friInst->getState();
  }
  // Quality change: report as a warning as it can imply errors on following commands
  if (friInst->getQuality() != arm->fri_quality) {
    printf("%s: Arm %s Communication quality has %s (new value is %d)\n",
	   (friInst->getQuality() > arm->fri_quality)?"INFO":"WARN", armName,
	   (friInst->getQuality() > arm->fri_quality)?"improved":"worsened",
	   friInst->getQuality());
    arm->fri_quality = friInst->getQuality();
  }
    
  memcpy(arm->estExtTcpFT,  friInst->getMsrEstExtTcpFT(), sizeof(float)*FRI_CART_VEC);
  memcpy(arm->estExtJntTrq,  friInst->getMsrEstExtJntTrq(), sizeof(float)*LWR_JOINTS);
  memcpy(arm->msrJntTrq,  friInst->getMsrJntTrq(), sizeof(float)*LWR_JOINTS);


  /* */
  for(int v=0; v<(JNT_TRQ_NB_TIC_MEAN-1); v++) {
    for(int w=0; w< LWR_JOINTS; w++) {  
      arm->estExtJntTrqHistory[w][v] = arm->estExtJntTrqHistory[w][v+1];
    }
  }
  for(int w=0; w< LWR_JOINTS; w++) {  
    arm->estExtJntTrqHistory[w][JNT_TRQ_NB_TIC_MEAN-1] = arm->estExtJntTrq[w];
  }

  for(int v=0; v< LWR_JOINTS; v++) { 
    arm->estMeanExtJntTrq[v] =  0.0;
    for(int w=0; w<JNT_TRQ_NB_TIC_MEAN; w++) { 
      arm->estMeanExtJntTrq[v] +=  arm->estExtJntTrqHistory[v][w];
    }
    arm->estMeanExtJntTrq[v] = arm->estMeanExtJntTrq[v] / (double)JNT_TRQ_NB_TIC_MEAN;
  }
  /* */
  for(int v=0; v<(LWR_FORCES_WINDOW-1); v++) {
    for(int u=0; u < LWR_CART_VEC; u++) {
      arm->estExtTcpFTHistory[v][u] = arm->estExtTcpFTHistory[v+1][u];
    }
  }
  for(int w=0; w< LWR_CART_VEC; w++) {  
    arm->estExtTcpFTHistory[LWR_FORCES_WINDOW-1][w] = arm->estExtTcpFT[w];
  }
  /* */



  //  FRI command scheme POSITION/CART_IMP/JOINT_IMP/GRAV_COMP
  arm->fri_scheme = (LWR_CTRL_SCHEME_ENUM)friInst->getCurrentControlScheme();
  arm->timeStamp = friInst->getTimestamp();
  return 0;
}

ACTIVITY_EVENT
requestFriStateCmd(int *report, LWR_ARM_ID_ENUM arm_id,  LWR_CTRL_SCHEME_ENUM ctrlScheme)
{
  static int iter = 0;
  static int max_iter = MAX_REQ_CMD;

  friRemote * friInst = getFriRemote(arm_id);
  if(friInst == NULL) {
    printf("ERROR requestFriStateCmd: friInst == NULL\n");
    return ETHER;
  }

  if (SDI_F->status.simulation == GEN_TRUE) {
    /* No verifications required on simulation mode */
    return EXEC;
  }

  if (friInst->getState() != FRI_STATE_CMD)
    syncFriWithMeasuredPose(friInst, ctrlScheme);

  if (iter == max_iter) {
    /* FRI doesn't enter in command state after enough tries*/
    friInst->setToKRLInt(CMD_CHANNEL, FRI_IDLE);
    printf("ERROR: FRI doesn't start after %d iterations on arm %d \n", iter, arm_id);
    *report = S_lwr_NOT_INITIALIZED;
    iter = 0;
    return ETHER;
  }
  else if (ctrlScheme != (LWR_CTRL_SCHEME_ENUM)friInst->getCurrentControlScheme()) {
    /* Not in desired control scheme; request an scheme change */
    if (iter == 0) {
      max_iter = 3*MAX_REQ_CMD;    // provide extra time for the switch
      printf("WARN: Switching control scheme to %s. This can take several seconds\n",
	     friSchemeStr(ctrlScheme));
    }
    
    switchCtrlScheme(friInst, ctrlScheme);
    
    iter++;
    return START;
  }
  else if (friInst->isPowerOn() == false) {
    /* Arm power is off; request switching it on */
    friInst->setToKRLInt(CMD_CHANNEL, POWER_ON);
    iter++;
    return START;
  }
  else if (friInst->getState() != FRI_STATE_CMD) {
    /* FRI is still not ready to accept commands; retry */
    friInst->setToKRLInt(CMD_CHANNEL, FRI_START);
    iter++;
    return START;
  }
  else {
    /* all ready; proceed to send command */
    friInst->setToKRLInt(CMD_CHANNEL, FRI_IDLE);
    *report = OK;
    iter = 0;
    return EXEC;
  }
}

ACTIVITY_EVENT
requestFriStateIdle(LWR_ARM_ID_ENUM arm_id, int *report)
{
  friRemote * friInst = getFriRemote(arm_id);
  if(friInst == NULL) {
    return ETHER;
  }

  /* Force entering on idle state */
  if(friInst->getFrmKRLInt(CMD_CHANNEL) != FRI_IDLE) {
    friInst->setToKRLInt(CMD_CHANNEL, FRI_IDLE);
    return END;
  }

  return ETHER;
}

#endif  /* ifndef LWR_FRI */

/*------------------------------------------------------------------------
 *
 * lwrCmdStart  --  Initialization codel (fIDS, ...)
 *
 * Description: 
 * 
 * Returns:    OK or ERROR
 */

STATUS
lwrCmdStart(int *report)
{
  
  return OK;
}

/*------------------------------------------------------------------------
 *
 * lwrCmdEnd  --  Termination codel
 *
 * Description: 
 * 
 * Returns:    OK or ERROR
 */
STATUS
lwrCmdEnd(void)
{
#ifdef LWR_FRI
  if(SDI_F->armLeft.is_used == GEN_TRUE) {
    printf("CLOSING CONNECTION FOR LWR LEFT ... \n");
    /* Send close connection code to KRL */
    //do {
    //  friInstLeft->setToKRLInt(CMD_CHANNEL, FRI_CLOSE);
    //  usleep(friInstLeft->getSampleTime()*1000);
    //}
    //while (friInstLeft->getFrmKRLInt(CMD_CHANNEL) != FRI_CLOSE);
      
    closingConnect = true;
    //usleep(friInstLeft->getSampleTime()*2000);
    delete friInstLeft;
    friInstLeft = NULL;
    printf(" LWR LEFT CLOSED!\n");
  } 
  if(SDI_F->armRight.is_used == GEN_TRUE) {
    printf("CLOSING CONNECTION FOR LWR RIGHT IN ... \n");
    /* Send close connection code to KRL */
    //do {
    //  friInstRight->setToKRLInt(CMD_CHANNEL, FRI_CLOSE);
    //  usleep(friInstRight->getSampleTime()*1000);
    //}
    //while (friInstRight->getFrmKRLInt(CMD_CHANNEL) != FRI_CLOSE);
      
    closingConnect = true;
    //usleep(friInstRight->getSampleTime()*2000);
    delete friInstRight;
    friInstRight = NULL;
    printf(" LWR RIGHT CLOSED!\n");
  }
 // if(threadFriComm) {
 //   pthread_exit();
 // }
#endif
  return OK;
}

/*------------------------------------------------------------------------
 *
 * lwrCmdPerm  --  1st codel of permanent activity (called BEFORE other activities)
 *
 * Description: 
 * 
 * Reports:      OK
 * 
 * Returns:    OK or ERROR
 */
static FILE* file = 0;
int pbFriCompteur = 0;

STATUS
lwrCmdPerm(int *report)
{
  if (!initIsDone) {
    return OK;
  }

#ifdef LWR_FRI
#else
  lwrCmd();
  lwrUpdateArmStatus( &SDI_F->armLeft , &SDI_F->pomArmLeftPos , (char *)"R7ArmLeft");
  lwrUpdateArmStatus( &SDI_F->armRight, &SDI_F->pomArmRightPos, (char *)"R7ArmRight");
  lwrGetInf();
  lwrUpdateMotion();
#endif
  lwrGenerateDigestStatus(); //?? not real time
  return OK;
}

STATUS lwrCmd() {

  bool flagLeft = false;
  bool flagRight = false;

  if (currentMotion.type != LWR_NOT_MOVE) {

    if(SDI_F->saveMotionFile == GEN_TRUE) {
      if(currentMotion.curTime == 0.0) {
	if ((fileptr = fopen("lwrMotion.dat","wb"))==NULL) {
	  printf("ERROR Open File lwrMotion.dat\n");
	} else {
	  if(fileptr) {
	    fprintf(fileptr,"# i ");
	    for(int u=1; u<=LWR_JOINTS_USER; u++) {
	      //fprintf(fileptr,"q%dCmd.a q%dCmd.v q%dCmd.x  q%dMsr.a q%dMsr.v q%dMsr.x ");
	      fprintf(fileptr,"q%dCmd.x q%dMsr.a q%dMsr.v q%dMsr.x ", u, u, u, u);
	    }
	    fprintf(fileptr," ;\n");
	  }
	}
      } /*  if(currentMotion.curTime == 0.0) */
    } /* if(SDI_F->saveMotionFile == GEN_TRUE) */


    /* check for predicted auto collision with colman library */
    if (SDI_F->status.collisionMode == GEN_TRUE) {
      for (int i = 0; i < SDI_F->motion.size; i++) {
	if (coldmanSetRobotQi(i, SDI_F->motion.q[i],
			      SDI_F->motion.vq[i],
			      SDI_F->motion.aq[i]) != 0) {
	  printf("ERROR coldmanSetRobotQi on axis %d\n",i);
	}
      }
	
      /* predict collision within 0.03 sec */
      if (coldmanPredictCollision(0.03) == 1) {
	printf("WARN: Coldman has reported collision risk. Motion aborted\n");
	setMotionError(LWR_COLLISION_RISK);   // this will set state as LWR_ABORTED
      } 
    } /*  (SDI_F->status.collisionMode == GEN_TRUE) */
      
      
    if(SDI_F->status.collisionEnvMode == GEN_TRUE) {	  
      if(lwrGENMANIP_COLLISION_STATE_STRPosterRead(posterMhpCollideId, &SDI_F->collisionEnvState) == ERROR) {
	printf("Collision reported cannot read poster of mhp \n");
      } else {
	if(SDI_F->collisionEnvState.state == GENMANIP_COLLISION_COLLISION) {
	  printf("Collision reported by env module: lwr stoped \n");
	  //setMotionError(LWR_COLLISION_RISK);  
	  setMotionEnd();
	}
      }
    } /*  (SDI_F->status.collisionEnvMode == GEN_TRUE) */
      
    // Add here your custom watch
      

#ifndef LWR_FRI
    
    if(SDI_F->armLeft.is_used == GEN_TRUE){	
      lwrCommandArm(SDI_F->isTrqCntrlSofMotion, LWR_ARM_LEFT, &SDI_F->armLeft, &(SDI_F->motion), &currentMotion);
    }
    if(SDI_F->armRight.is_used == GEN_TRUE){
      lwrCommandArm(SDI_F->isTrqCntrlSofMotion, LWR_ARM_RIGHT, &SDI_F->armRight, &(SDI_F->motion), &currentMotion);
    }
    
#else
    if(SDI_F->armLeft.is_used == GEN_TRUE) {
      if(friInstLeft->getState() != FRI_STATE_CMD) {
	flagLeft = true;
      }
    } else {
      flagLeft = true;
    }
       
    if(SDI_F->armRight.is_used == GEN_TRUE) {
      if(friInstRight->getState() != FRI_STATE_CMD) {
	flagRight = true;
      }
    } else {
      flagRight = true;
    }
    
    if ( flagRight && flagLeft && (SDI_F->status.simulation != GEN_TRUE)) {
      // The LWR has revoqued the FRI command state due to an exceptional situation
      printf("ERROR: FRI command state revoqued. Motion aborted flagRight: %d flagLeft: %d\n", flagRight, flagLeft);
      setMotionError(LWR_FRI_MONITOR_MODE);
    } // TODO CHECK IT FOR THE TIRGHT ARM
    else {


      if(SDI_F->armLeft.is_used == GEN_TRUE){	
	lwrCommandArm(friInstLeft, SDI_F->status.simulation, SDI_F->isTrqCntrlSofMotion, LWR_ARM_LEFT, &SDI_F->armLeft, &(SDI_F->motion), &currentMotion);
      }
    
      if(SDI_F->armRight.is_used == GEN_TRUE){
	lwrCommandArm(friInstRight, SDI_F->status.simulation, SDI_F->isTrqCntrlSofMotion, LWR_ARM_RIGHT, &SDI_F->armRight, &(SDI_F->motion), &currentMotion);
      }


#endif
      if(SDI_F->saveMotionFile == GEN_TRUE) {
	/*save motion into the file */
	if(fileptr) {
	  fprintf(fileptr,"%f ", SDI_F->armLeft.timeStamp);
	  for(int u=0; u<LWR_JOINTS_USER; u++) {
	    if(u < LWR_JOINTS) {
	      fprintf(fileptr,"%f %f %f %f %f %f %f %f ", SDI_F->armLeft.estMeanExtJntTrq[u], 
		      SDI_F->armLeft.currentJntVals[u],  SDI_F->armLeft.currentMsrJntVals[u],
		      SDI_F->motion.vq[u],  SDI_F->motion.aq[u],
		      SDI_F->armLeft.currentCmdVelJntVals[u],SDI_F->armLeft.currentCmdAccJntVals[u],
		      currentMotion.q_V_max[u]);
	    } else {
	      fprintf(fileptr,"%f %f %f %f %f %f %f %f ", SDI_F->armRight.estMeanExtJntTrq[u-LWR_JOINTS], 
		      SDI_F->armLeft.currentJntVals[u],  SDI_F->armLeft.currentMsrJntVals[u],
		      SDI_F->motion.vq[u],  SDI_F->motion.aq[u],
		      SDI_F->armRight.currentCmdVelJntVals[u-LWR_JOINTS], SDI_F->armRight.currentCmdAccJntVals[u-LWR_JOINTS],
		      currentMotion.q_V_max[u]);
	    }
	  }
	  fprintf(fileptr," \n");
	}
      }


      if(SDI_F->armLeft.cntrl_mode == LWR_OPEN_LOOP_CNTRL || SDI_F->armLeft.cntrl_mode == LWR_OPEN_LOOP_TRACK_Q_CNTRL || SDI_F->armLeft.cntrl_mode == LWR_TRAJECTORY_CNTRL || 
	 SDI_F->armRight.cntrl_mode == LWR_OPEN_LOOP_CNTRL || SDI_F->armRight.cntrl_mode == LWR_OPEN_LOOP_TRACK_Q_CNTRL ||  SDI_F->armRight.cntrl_mode == LWR_TRAJECTORY_CNTRL ) {
	if (currentMotion.curTime < currentMotion.duration) {

	  /* Update the hand command */
	  if(SDI_F->armRight.isHandUsed == GEN_TRUE) {
	    SDI_F->sahand.handRight = SDI_F->armRight.hand;
       	  }
	  SDI_F->sahand.cpt = currentMotion.sample;

	  if((posterHumDistId != NULL) && (SDI_F->humanDistMode == GEN_TRUE)) {
	    if(lwrGENHUM_HUMAN_DISTANCEPosterRead(posterHumDistId, &SDI_F->humanDist) == ERROR) {
	      printf("lwrGENHUM_HUMAN_DISTANCEPosterRead CANNOT read poster\n");
	    } else {
	      SDI_F->timeScale = 1.0 - SDI_F->humanDist.costDistRobToHum;
	    }
	  }
	 
	  SM_LIMITS limitsGoto;
	  limitsGoto.maxJerk = 9.0;
	  limitsGoto.maxAcc  = 3.0;
	  limitsGoto.maxVel  = 1.0;
	  lwrComputeSmoothedTimeScale(SDI_F->timeScale, limitsGoto, &currentMotion.timeScaleCond);


	  currentMotion.curTime = currentMotion.curTime + (currentMotion.timeScaleCond.v * LWR_PERIOD);
	  // toto
	  //currentMotion.curTime = currentMotion.curTime + (SDI_F->timeScale*LWR_PERIOD);
	  //printf("%f \n",currentMotion.curTime);

	  currentMotion.sample ++;

	  /* Check curTime validity (it can be > 0 since there is the timeScale factor */
	  if(currentMotion.curTime < 0.0 || currentMotion.curTime > currentMotion.duration) {
	    setMotionEnd();
	    if(SDI_F->saveMotionFile == GEN_TRUE) {
	      if(fileptr) {
		fclose(fileptr);
		printf("File lwrMotion.dat is written !\n");
		fileptr = NULL;
	      }
	    }
	    printf("Motion completed\n");
	  }
	} else {
	  
	  if(SDI_F->armLeft.cntrl_mode == LWR_FORCE_CNTRL || SDI_F->armRight.cntrl_mode == LWR_FORCE_CNTRL) {
	    SDI_F->armLeft.cntrl_mode = LWR_FORCE_CNTRL ;
	    SDI_F->armRight.cntrl_mode = LWR_FORCE_CNTRL;
	    printf("WARN: All Arm switch to LWR_FORCE_CNTRL \n");
	  } else {
	    setMotionEnd();
	    if(SDI_F->saveMotionFile == GEN_TRUE) {
	      if(fileptr) {
		fclose(fileptr);
		printf("File lwrMotion.dat is written !\n");
		fileptr = NULL;
	      }
	    }
	    printf("Motion completed\n");
	  }
	}
      } 
      if(SDI_F->armLeft.cntrl_mode == LWR_FORCE_CNTRL && SDI_F->armRight.cntrl_mode == LWR_FORCE_CNTRL) {
	if(SDI_F->armLeft.isStabilized == GEN_TRUE &&  SDI_F->armRight.isStabilized == GEN_TRUE) {
	  setMotionEnd();
	  if(SDI_F->saveMotionFile == GEN_TRUE) {
	    if(fileptr) {
	      fclose(fileptr);
	      printf("File lwrMotion.dat is written !\n");
	      fileptr = NULL;
	    }
	  }
	  printf("Motion completed\n");
	}
      }

      if(currentMotion.type  != LWR_NOT_MOVE && SDI_F->isMonitoringForces == GEN_TRUE) {
	lwrMonitorAndUpdateCntrlMode(&(SDI_F->armLeft), (char*)"LEFT");
	lwrMonitorAndUpdateCntrlMode(&(SDI_F->armRight), (char*)"RIGHT");
      }


#ifdef LWR_FRI
    } /* else { --> flagRight && flagLeft && (SDI_F->status.simulation != GEN_TRUE) */
#endif
  } else {
    if(SDI_F->status.sahandUpdate == GEN_TRUE) {	  
      if(lwrGENMANIP_SAHAND_STRPosterRead(posterSAHandId, &SDI_F->sahandPos) == ERROR) {
	printf("Collision reported cannot read poster of sahand \n");
      } else {
	SDI_F->sahand.handRight.thumbRotation  = SDI_F->sahandPos.handRight.thumbRotation ;
	SDI_F->sahand.handRight.thumb.base     = SDI_F->sahandPos.handRight.thumb.base    ;
	SDI_F->sahand.handRight.thumb.phaProx  = SDI_F->sahandPos.handRight.thumb.phaProx ;
	SDI_F->sahand.handRight.thumb.phaMid   = SDI_F->sahandPos.handRight.thumb.phaMid  ;
	SDI_F->sahand.handRight.thumb.phaDist  = SDI_F->sahandPos.handRight.thumb.phaDist ;
	SDI_F->sahand.handRight.fore.base      = SDI_F->sahandPos.handRight.fore.base     ;
	SDI_F->sahand.handRight.fore.phaProx   = SDI_F->sahandPos.handRight.fore.phaProx  ;
	SDI_F->sahand.handRight.fore.phaMid    = SDI_F->sahandPos.handRight.fore.phaMid   ;
	SDI_F->sahand.handRight.fore.phaDist   = SDI_F->sahandPos.handRight.fore.phaDist  ;
	SDI_F->sahand.handRight.middle.base    = SDI_F->sahandPos.handRight.middle.base   ;
	SDI_F->sahand.handRight.middle.phaProx = SDI_F->sahandPos.handRight.middle.phaProx;
	SDI_F->sahand.handRight.middle.phaMid  = SDI_F->sahandPos.handRight.middle.phaMid ;
	SDI_F->sahand.handRight.middle.phaDist = SDI_F->sahandPos.handRight.middle.phaDist ;
	SDI_F->sahand.handRight.ring.base      = SDI_F->sahandPos.handRight.ring.base     ;
	SDI_F->sahand.handRight.ring.phaProx   = SDI_F->sahandPos.handRight.ring.phaProx  ;
	SDI_F->sahand.handRight.ring.phaMid    = SDI_F->sahandPos.handRight.ring.phaMid   ;
	SDI_F->sahand.handRight.ring.phaDist   = SDI_F->sahandPos.handRight.ring.phaDist  ;
	SDI_F->sahand.cpt = SDI_F->sahandPos.cpt;
      }
    } /*  (SDI_F->status.sahandUpdate == GEN_TRUE) */


  }
  return OK;
}

int lwrMonitorAndUpdateCntrlMode(LWR_ARM_INST *armInst, char* message) {
  double minForce[LWR_JOINTS];
  double coef = 5.0;
  double force = 0.0;
  minForce[0] = coef *1.0;
  minForce[1] = coef *1.0;
  minForce[2] = coef *1.0;
  minForce[3] = coef *1.0;
  minForce[4] = coef *0.5;
  minForce[5] = coef *0.5;
  minForce[6] = coef *0.5;
  unsigned int i = 0;

  if(armInst->is_used == GEN_FALSE) {
    armInst->isStabilized = GEN_TRUE;
    return 0;
  }
  for( i=0 ; i<LWR_JOINTS; i++) { 
    force = armInst->estMeanExtJntTrq[i];

    if(ABS(force) > ABS(minForce[i]) ) {
      //if(armInst->cntrl_mode != LWR_FORCE_CNTRL) {
      //  printf("WARN: Arm %s switch to LWR_FORCE_CNTRL \n", message);
      //}
      //armInst->cntrl_mode = LWR_FORCE_CNTRL;
      setMotionEnd();
      if(SDI_F->saveMotionFile == GEN_TRUE) {
	if(fileptr) {
	  fclose(fileptr);
	  printf("File lwrMotion.dat is written !\n");
	  fileptr = NULL;
	}
      }
      printf("Motion stopped due to external forces\n");

      break;
    }
  }
  return 0;
}

int lwrUpdateMotion() {
  SDI_F->motion.size = LWR_JOINTS_USER+13;
  SDI_F->motion.q[0] =  SDI_F->armLeft.pos.q.q1;
  SDI_F->motion.q[1] =  SDI_F->armLeft.pos.q.q2;
  SDI_F->motion.q[2] =  SDI_F->armLeft.pos.q.q3;
  SDI_F->motion.q[3] =  SDI_F->armLeft.pos.q.q4;
  SDI_F->motion.q[4] =  SDI_F->armLeft.pos.q.q5;
  SDI_F->motion.q[5] =  SDI_F->armLeft.pos.q.q6;
  SDI_F->motion.q[6] =  SDI_F->armLeft.pos.q.q7;
  SDI_F->motion.q[7]  =  SDI_F->armRight.pos.q.q1;
  SDI_F->motion.q[8]  =  SDI_F->armRight.pos.q.q2;
  SDI_F->motion.q[9]  =  SDI_F->armRight.pos.q.q3;
  SDI_F->motion.q[10] =  SDI_F->armRight.pos.q.q4;
  SDI_F->motion.q[11] =  SDI_F->armRight.pos.q.q5;
  SDI_F->motion.q[12] =  SDI_F->armRight.pos.q.q6;
  SDI_F->motion.q[13] =  SDI_F->armRight.pos.q.q7;

  SDI_F->motion.vq[0] = SDI_F->armLeft.pos.vq.q1;
  SDI_F->motion.vq[1] = SDI_F->armLeft.pos.vq.q2;
  SDI_F->motion.vq[2] = SDI_F->armLeft.pos.vq.q3;
  SDI_F->motion.vq[3] = SDI_F->armLeft.pos.vq.q4;
  SDI_F->motion.vq[4] = SDI_F->armLeft.pos.vq.q5;
  SDI_F->motion.vq[5] = SDI_F->armLeft.pos.vq.q6;
  SDI_F->motion.vq[6] = SDI_F->armLeft.pos.vq.q7;
  SDI_F->motion.vq[7]  = SDI_F->armRight.pos.vq.q1;
  SDI_F->motion.vq[8]  = SDI_F->armRight.pos.vq.q2;
  SDI_F->motion.vq[9]  = SDI_F->armRight.pos.vq.q3;
  SDI_F->motion.vq[10] = SDI_F->armRight.pos.vq.q4;
  SDI_F->motion.vq[11] = SDI_F->armRight.pos.vq.q5;
  SDI_F->motion.vq[12] = SDI_F->armRight.pos.vq.q6;
  SDI_F->motion.vq[13] = SDI_F->armRight.pos.vq.q7;

  SDI_F->motion.aq[0] = SDI_F->armLeft.pos.aq.q1;
  SDI_F->motion.aq[1] = SDI_F->armLeft.pos.aq.q2;
  SDI_F->motion.aq[2] = SDI_F->armLeft.pos.aq.q3;
  SDI_F->motion.aq[3] = SDI_F->armLeft.pos.aq.q4;
  SDI_F->motion.aq[4] = SDI_F->armLeft.pos.aq.q5;
  SDI_F->motion.aq[5] = SDI_F->armLeft.pos.aq.q6;
  SDI_F->motion.aq[6] = SDI_F->armLeft.pos.aq.q7;
  SDI_F->motion.aq[7]  = SDI_F->armRight.pos.aq.q1;
  SDI_F->motion.aq[8]  = SDI_F->armRight.pos.aq.q2;
  SDI_F->motion.aq[9]  = SDI_F->armRight.pos.aq.q3;
  SDI_F->motion.aq[10] = SDI_F->armRight.pos.aq.q4;
  SDI_F->motion.aq[11] = SDI_F->armRight.pos.aq.q5;
  SDI_F->motion.aq[12] = SDI_F->armRight.pos.aq.q6;
  SDI_F->motion.aq[13] = SDI_F->armRight.pos.aq.q7;

  int indexHandRight = 14;
  SDI_F->motion.q[indexHandRight]    = SDI_F->sahand.handRight.thumbRotation ;
  SDI_F->motion.q[indexHandRight+1]  = SDI_F->sahand.handRight.thumb.base    ;
  SDI_F->motion.q[indexHandRight+2]  = SDI_F->sahand.handRight.thumb.phaProx ;
  SDI_F->motion.q[indexHandRight+3]  = SDI_F->sahand.handRight.thumb.phaMid  ;
  SDI_F->motion.q[indexHandRight+4]  = SDI_F->sahand.handRight.fore.base     ;
  SDI_F->motion.q[indexHandRight+5]  = SDI_F->sahand.handRight.fore.phaProx  ;
  SDI_F->motion.q[indexHandRight+6]  = SDI_F->sahand.handRight.fore.phaMid   ;
  SDI_F->motion.q[indexHandRight+7]  = SDI_F->sahand.handRight.middle.base   ;
  SDI_F->motion.q[indexHandRight+8]  = SDI_F->sahand.handRight.middle.phaProx;
  SDI_F->motion.q[indexHandRight+9]  = SDI_F->sahand.handRight.middle.phaMid ;
  SDI_F->motion.q[indexHandRight+10] = SDI_F->sahand.handRight.ring.base     ;
  SDI_F->motion.q[indexHandRight+11] = SDI_F->sahand.handRight.ring.phaProx  ;
  SDI_F->motion.q[indexHandRight+12] = SDI_F->sahand.handRight.ring.phaMid   ;
  return 0;
}
#ifdef LWR_FRI  
int lwrInitArm(LWR_INIT_ARM_STR arm, friRemote *friInst, LWR_ARM_INST *armInst, char* message) {

  armInst->is_used = GEN_FALSE;

  armInst->currentJntVals[0] = 0.0;
  armInst->currentJntVals[1] = 0.0;
  armInst->currentJntVals[2] = 0.0;
  armInst->currentJntVals[3] = 0.0;
  armInst->currentJntVals[4] = 0.0;
  armInst->currentJntVals[5] = 0.0;
  armInst->currentJntVals[6] = 0.0;
  
  /* Setup default cartesian stiffness and damping factors */
  armInst->cartStiff.x  = LWR_DEF_X_STIFFNESS;
  armInst->cartStiff.y  = LWR_DEF_Y_STIFFNESS;
  armInst->cartStiff.z  = LWR_DEF_Z_STIFFNESS;
  armInst->cartStiff.rx = LWR_DEF_A_STIFFNESS;
  armInst->cartStiff.ry = LWR_DEF_B_STIFFNESS;
  armInst->cartStiff.rz = LWR_DEF_C_STIFFNESS;
  armInst->cartDamp.x   = LWR_DEF_X_DAMPING;
  armInst->cartDamp.y   = LWR_DEF_Y_DAMPING;
  armInst->cartDamp.z   = LWR_DEF_Z_DAMPING;
  armInst->cartDamp.rx  = LWR_DEF_A_DAMPING;
  armInst->cartDamp.ry  = LWR_DEF_B_DAMPING;
  armInst->cartDamp.rz  = LWR_DEF_C_DAMPING;

  armInst->s_date0_tv_sec  = 0.0;
  armInst->s_date0_tv_usec = 0.0;
  armInst->date1 = -1;
  armInst->date2 = -1;
    
  if(arm.is_used == GEN_TRUE) {

    /* Setup transformation matrix from reference system 7 to R tool */
    armInst->toolCfg.distance  = -1;
    armInst->toolCfg.tool_type = arm.tool_type;
    setupTh7Tool(friInst, armInst->toolCfg.tool_type, armInst->toolCfg.distance, &(armInst->th_7_tool), &(armInst->th_8_tool));

    /* Setup transformation matrix from world to base reference systems */
    armInst->baseCfg.base_type = arm.base_type;
    setupThBase0(friInst, armInst->baseCfg.base_type, &(armInst->th_base_0));

    // Ideally, this should avoid a FRI interpolation error on the first command
    if (friInst->getQuality() >= FRI_QUALITY_BAD){
      syncFriWithMeasuredPose(friInst);
    }

    armInst->sock = friInst->getSock();
    printf("INFO: FRI %s connected at socket %d\n", message, friInst->getSock());
    armInst->is_used = GEN_TRUE;
    printf("INFO: FRI %s pointer %p\n", message,friInst);
  }
  return 0;
}
#else
int lwrInitArm(LWR_INIT_ARM_STR arm,  LWR_ARM_INST *armInst, char* message) {

  armInst->is_used = GEN_FALSE;

  armInst->currentJntVals[0] = 0.0;
  armInst->currentJntVals[1] = 0.0;
  armInst->currentJntVals[2] = 0.0;
  armInst->currentJntVals[3] = 0.0;
  armInst->currentJntVals[4] = 0.0;
  armInst->currentJntVals[5] = 0.0;
  armInst->currentJntVals[6] = 0.0;
  
  /* Setup default cartesian stiffness and damping factors */
  armInst->cartStiff.x  = LWR_DEF_X_STIFFNESS;
  armInst->cartStiff.y  = LWR_DEF_Y_STIFFNESS;
  armInst->cartStiff.z  = LWR_DEF_Z_STIFFNESS;
  armInst->cartStiff.rx = LWR_DEF_A_STIFFNESS;
  armInst->cartStiff.ry = LWR_DEF_B_STIFFNESS;
  armInst->cartStiff.rz = LWR_DEF_C_STIFFNESS;
  armInst->cartDamp.x   = LWR_DEF_X_DAMPING;
  armInst->cartDamp.y   = LWR_DEF_Y_DAMPING;
  armInst->cartDamp.z   = LWR_DEF_Z_DAMPING;
  armInst->cartDamp.rx  = LWR_DEF_A_DAMPING;
  armInst->cartDamp.ry  = LWR_DEF_B_DAMPING;
  armInst->cartDamp.rz  = LWR_DEF_C_DAMPING;

  armInst->s_date0_tv_sec  = 0.0;
  armInst->s_date0_tv_usec = 0.0;
  armInst->date1 = -1;
  armInst->date2 = -1;
    
  if(arm.is_used == GEN_TRUE) {
    /* create an instance of FRI interface; the connection socket is automatically opened */
 
    /* Setup transformation matrix from reference system 7 to R tool */
    armInst->toolCfg.distance  = -1;
    armInst->toolCfg.tool_type = arm.tool_type;

    setupTh7Tool( armInst->toolCfg.tool_type, armInst->toolCfg.distance, &(armInst->th_7_tool), &(armInst->th_8_tool));

    /* Setup transformation matrix from world to base reference systems */
    armInst->baseCfg.base_type = arm.base_type;
    setupThBase0( armInst->baseCfg.base_type, &(armInst->th_base_0));

    armInst->is_used = GEN_TRUE;
  }
  return 0;
}

#endif


/*------------------------------------------------------------------------
 * Init
 *
 * Description: 
 *
 * Reports:      OK
 *              S_lwr_FILE_NOT_FOUND
 *              S_lwr_UNKNOW_ERROR
 */

/* lwrInitMain  -  codel EXEC of Init
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrInitMain(LWR_INIT_STR *initStr, int *report)
{
  int errcode = 0;
  printf("init lwr\n");
  if (initIsDone) {
    printf("Init is already done! --> Do nothing\n");
    return ETHER;
  }

  if(initStr->armLeft.is_used == GEN_FALSE && initStr->armRight.is_used == GEN_FALSE) {
    printf("ERROR: There is no arm used \n");  
    SDI_F->status.simulation = GEN_TRUE;
    return ETHER;
  }
#ifdef LWR_FRI  
  if(initStr->armLeft.is_used == GEN_TRUE) {
    printf("INFO: connecting to the left arm ...\n"); 
    friInstLeft = new friRemote(initStr->armLeft.lwr_port, initStr->armLeft.lwr_host.name);
    printf("INFO: FRI %s connected at %s:%d\n", (char*)"LEFT", initStr->armLeft.lwr_host.name, initStr->armLeft.lwr_port);
  }
  lwrInitArm( initStr->armLeft, friInstLeft, &(SDI_F->armLeft), (char*)"LEFT"); 

  if(initStr->armRight.is_used == GEN_TRUE) {
    printf("INFO: connecting to the right arm ...\n"); 
    friInstRight = new friRemote(initStr->armRight.lwr_port, initStr->armRight.lwr_host.name);
    printf("INFO: FRI %s connected at %s:%d\n", (char*)"RIGHT", initStr->armRight.lwr_host.name, initStr->armRight.lwr_port);
  }
  lwrInitArm( initStr->armRight,friInstRight, &(SDI_F->armRight), (char*)"RIGHT");  
#else 
  lwrInitArm( initStr->armLeft,  &(SDI_F->armLeft), (char*)"LEFT"); 
  lwrInitArm( initStr->armRight, &(SDI_F->armRight), (char*)"RIGHT");  
#endif
  /***********************************/
  /*  Initialize global structures   */                 
  /***********************************/
 
  memset(&currentMotion, 0, sizeof(currentMotion));
  SDI_F->status.sahandUpdate = GEN_FALSE;  
  SDI_F->status.collisionMode = GEN_TRUE;
  SDI_F->status.collisionEnvMode = GEN_FALSE;
  SDI_F->status.simulation = GEN_FALSE;

  SDI_F->timeScale = 1.0;

  LWR_JOINTS_USER =  2*LWR_JOINTS;
  LWR_CART_USER = 2*7;

  /* Setup v, a, j limits */
  SDI_F->speedLimit = LWR_DEF_SPEED_LIMIT;
  setupMaxVAJ(SDI_F->speedLimit, currentMotion.q_V_max, currentMotion.q_A_max, currentMotion.q_J_max, currentMotion.x_V_max, currentMotion.x_A_max, currentMotion.x_J_max);

  /* Setup collision detection manager */
  if (coldmanEnvInit(initStr->world_path.name, initStr->world_file.name) != 0) {
    *report = S_lwr_FILE_NOT_FOUND;
    return ETHER;
  }
  printf("INFO: coldman model loaded from file %s/%s\n", initStr->world_path.name, initStr->world_file.name);

  //TODO Check if all sampling time are equal (armLeft, armRight and LWR_PERIOD)
#ifdef LWR_FRI  
  printf("##########################################################\n");
  printf("### INFO: Now, launch the FRI on the arm(s) controller ###\n");
  printf("##########################################################\n");
  // The next line is to wait the Kuka initialisation
  lwrDoDataExchange();

  if ((threadFriCommTimerId = h2timerAlloc()) == NULL) {
    printf("ERROR: lwrInitMain: h2timerAlloc\n");
    return ETHER;
  }
  if (h2timerStart (threadFriCommTimerId, LWR_TIC, 0) != OK) {
    printf("ERROR: lwrInitMain: h2timerStart\n");
    return ETHER;
  }

  /* create the FriComm thread */
  //if (errcode=pthread_create(&threadFriComm, NULL, threadFriCommFct, NULL)) {
  //  printf("ERROR: pthread_create cannot create thread err: %d\n",errcode);
  //  return ETHER;
  //}
  // taskSpawn : create a "vxWorks" thread
  if (taskSpawn("lwrDataExchangeTask",
		1 /* priorite */,
		VX_FP_TASK,
		4096 /*size*/, 
		(FUNCPTR)threadFriCommFct) == ERROR) {
     printf("lwrInitMain: cannot spawn lwrDataExchangeTask\n");
     return ETHER;
  }
#endif

  if(SDI_F->armLeft.is_used == GEN_TRUE && SDI_F->armRight.is_used == GEN_TRUE) {
    SDI_F->armLeft.indexInTraj =  0;
    SDI_F->armRight.indexInTraj = LWR_JOINTS;
    SDI_F->armRight.indexHandInTraj = 2*LWR_JOINTS;
    SDI_F->armLeft.isHandUsed = GEN_FALSE;
    SDI_F->armRight.isHandUsed = GEN_TRUE;   
   
  } else if (SDI_F->armLeft.is_used == GEN_TRUE) {
    SDI_F->armLeft.indexInTraj =  0;
    SDI_F->armLeft.isHandUsed = GEN_FALSE;
    SDI_F->armRight.isHandUsed = GEN_FALSE;   
  } else if (SDI_F->armRight.is_used == GEN_TRUE) {
    SDI_F->armRight.indexInTraj =  0;
    SDI_F->armRight.indexHandInTraj = LWR_JOINTS;
    SDI_F->armLeft.isHandUsed = GEN_FALSE;
    SDI_F->armRight.isHandUsed = GEN_TRUE;   
  } else {
    printf("ERROR: No arm is used\n");
    return ETHER;
  }
    
  SDI_F->armLeft.isLeft = GEN_TRUE;
  SDI_F->armRight.isLeft = GEN_FALSE;
  SDI_F->isTrqCntrlSofMotion = GEN_FALSE;
  SDI_F->isMonitoringForces = GEN_FALSE;

  currentMotion.timeScaleCond.a = 0.0;
  currentMotion.timeScaleCond.v = SDI_F->timeScale;
  currentMotion.timeScaleCond.x = 0.0;

  SDI_F->saveMotionFile = GEN_FALSE;
  initIsDone = true;
  printf("lwr Init OK\n");
  return ETHER;
}


/*------------------------------------------------------------------------
 * SetCartStiff
 *
 * Description: Set current cartesian stiffness values
 *
 * Reports:      OK
 *              S_lwr_OutOfRange
 */

/* lwrSetCartStiffStart  -  codel START of SetCartStiff
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetCartStiffStart(LWR_CART_STIFF *cartStiff, int *report)
{
  /* TODO verify bounds */
  return EXEC;
}

/* lwrSetCartStiffMain  -  codel EXEC of SetCartStiff
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetCartStiffMain(LWR_CART_STIFF *cartStiff, int *report)
{
#ifdef LWR_FRI
  friRemote * friInst = getFriRemote(cartStiff->arm_id);
  if(friInst == NULL) {
    return END;
  }

  friInst->setToKRLReal(0, cartStiff->cartStiff.x);
  friInst->setToKRLReal(1, cartStiff->cartStiff.y);
  friInst->setToKRLReal(2, cartStiff->cartStiff.z);
  friInst->setToKRLReal(3, cartStiff->cartStiff.rx);
  friInst->setToKRLReal(4, cartStiff->cartStiff.ry);
  friInst->setToKRLReal(5, cartStiff->cartStiff.rz);
  friInst->setToKRLInt(CMD_CHANNEL, CART_STIFF);

  printf("Cartesian stiffness set to: %f, %f, %f, %f, %f, %f\n",
         cartStiff->cartStiff.x,  cartStiff->cartStiff.y,  cartStiff->cartStiff.z,
         cartStiff->cartStiff.rx, cartStiff->cartStiff.ry, cartStiff->cartStiff.rz);
#endif
  return END;
}

/* lwrSetCartStiffEnd  -  codel END of SetCartStiff
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetCartStiffEnd(LWR_CART_STIFF *cartStiff, int *report)
{
#ifdef LWR_FRI
  friRemote * friInst = getFriRemote(cartStiff->arm_id);
  if(friInst == NULL) {
    return END;
  }
  // Return arm script to idle status
  friInst->setToKRLInt(CMD_CHANNEL, FRI_IDLE);
#endif
  return ETHER;
}

/*------------------------------------------------------------------------
 * SetCartDamp
 *
 * Description: Set current cartesian damping values
 *
 * Reports:      OK
 *              S_lwr_OutOfRange
 */

/* lwrSetCartDampStart  -  codel START of SetCartDamp
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetCartDampStart(LWR_CART_DAMP *cartDamp, int *report)
{
  /* TODO verify bounds */
  return EXEC;
}

/* lwrSetCartDampMain  -  codel EXEC of SetCartDamp
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetCartDampMain(LWR_CART_DAMP *cartDamp, int *report)
{
#ifdef LWR_FRI
  friRemote * friInst = getFriRemote(cartDamp->arm_id);
  if(friInst == NULL) {
    return END;
  }

  friInst->setToKRLReal(0, cartDamp->cartDamp.x);
  friInst->setToKRLReal(1, cartDamp->cartDamp.y);
  friInst->setToKRLReal(2, cartDamp->cartDamp.z);
  friInst->setToKRLReal(3, cartDamp->cartDamp.rx);
  friInst->setToKRLReal(4, cartDamp->cartDamp.ry);
  friInst->setToKRLReal(5, cartDamp->cartDamp.rz);
  friInst->setToKRLInt(CMD_CHANNEL, CART_DAMP);

  printf("Cartesian damping set to: %f, %f, %f, %f, %f, %f\n",
         cartDamp->cartDamp.x,  cartDamp->cartDamp.y,  cartDamp->cartDamp.z,
         cartDamp->cartDamp.rx, cartDamp->cartDamp.ry, cartDamp->cartDamp.rz);
#endif
  return END;
}

/* lwrSetCartDampEnd  -  codel END of SetCartDamp
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetCartDampEnd(LWR_CART_DAMP *cartDamp, int *report)
{
#ifdef LWR_FRI
  friRemote * friInst = getFriRemote(cartDamp->arm_id);
  if(friInst == NULL) {
    return END;
  }

  // Return arm script to idle status
  friInst->setToKRLInt(CMD_CHANNEL, FRI_IDLE);
#endif 
  return ETHER;
}



/*------------------------------------------------------------------------
 * GotoQ
 *
 * Description: 
 *
 * Reports:      OK
 *              S_lwr_NOT_INITIALIZED
 *              S_lwr_WRONG_VALUE
 *              S_lwr_SOFTMOTION_ERROR
 *              S_lwr_FRI_MONITOR_MODE
 *              S_lwr_COLLISION_RISK
 */

/* lwrGotoQStart  -  codel START of GotoQ
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoQStart(LWR_GOTO_CONF_STR *gotoConf, int *report)
{

  currentMotion.timeScaleCond.a = 0.0;
  currentMotion.timeScaleCond.v = SDI_F->timeScale;
  currentMotion.timeScaleCond.x = 0.0;

  if(SDI_F->status.simulation == GEN_TRUE) {
    return EXEC;
  }
#ifdef LWR_FRI
  ACTIVITY_EVENT eventLeft, eventRight;
  if(SDI_F->armLeft.is_used == GEN_TRUE) {
    eventLeft = requestFriStateCmd(report, LWR_ARM_LEFT, LWR_CTRL_POSITION);
  } else {
    eventLeft = EXEC;
  }
  if(SDI_F->armRight.is_used == GEN_TRUE) {
    eventRight = requestFriStateCmd(report, LWR_ARM_RIGHT, LWR_CTRL_POSITION);
  } else {
    eventRight = EXEC;
  }

  if(eventLeft == EXEC && eventRight == EXEC) {
    printf("lwrGotoQStart OK\n");
    return EXEC;
  }
  if(eventLeft == ETHER || eventRight == ETHER) { 
    return ETHER;
  }
  if(eventLeft == START || eventRight == START) {
    return START;
  }
#else
  return EXEC;
#endif
}

/* lwrGotoQMain  -  codel EXEC of GotoQ
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoQMain(LWR_GOTO_CONF_STR *gotoConf, int *report)
{
  /* Send new absolute angles for all robot joints */



  memcpy(&(SDI_F->gotoConf), gotoConf, sizeof(LWR_GOTO_CONF_STR));

  double criticalLength;
  SM_MOTION_MONO motion[LWR_JOINTS_USER];
  SM_LIMITS limitsGoto;
  int TrajectoryType[LWR_JOINTS_USER];
  SM_TIMES T_Jerk[LWR_JOINTS_USER];
  double maxTime = 0.0;
  double sum = 0.0;
  double maxAxis = 0;
 

  memset(motion, 0, LWR_JOINTS_USER*sizeof(SM_MOTION_MONO));
  for(int i=0; i<LWR_JOINTS_USER; i++) {
    motion[i].IC.x = SDI_F->motion.q[i] ;
    motion[i].IC.v = SDI_F->motion.vq[i] ;
    motion[i].IC.a = SDI_F->motion.aq[i] ;
  }

  if(SDI_F->armLeft.is_used == GEN_TRUE) {
    printf("lwrGotoQAbsMain Arm Left is_used\n");
    motion[0].FC.x = SDI_F->gotoConf.confLeft.q1;
    motion[1].FC.x = SDI_F->gotoConf.confLeft.q2;
    motion[2].FC.x = SDI_F->gotoConf.confLeft.q3;
    motion[3].FC.x = SDI_F->gotoConf.confLeft.q4;
    motion[4].FC.x = SDI_F->gotoConf.confLeft.q5;
    motion[5].FC.x = SDI_F->gotoConf.confLeft.q6;
    motion[6].FC.x = SDI_F->gotoConf.confLeft.q7;
  } else {
    motion[0].FC.x = motion[0].IC.x;
    motion[1].FC.x = motion[1].IC.x;
    motion[2].FC.x = motion[2].IC.x;
    motion[3].FC.x = motion[3].IC.x;
    motion[4].FC.x = motion[4].IC.x;
    motion[5].FC.x = motion[5].IC.x;
    motion[6].FC.x = motion[6].IC.x;
  }
  if(SDI_F->armRight.is_used == GEN_TRUE) {
    printf("lwrGotoQAbsMain Arm Right is_used\n");
    motion[7].FC.x  = SDI_F->gotoConf.confRight.q1;
    motion[8].FC.x  = SDI_F->gotoConf.confRight.q2;
    motion[9].FC.x  = SDI_F->gotoConf.confRight.q3;
    motion[10].FC.x = SDI_F->gotoConf.confRight.q4;
    motion[11].FC.x = SDI_F->gotoConf.confRight.q5;
    motion[12].FC.x = SDI_F->gotoConf.confRight.q6;
    motion[13].FC.x = SDI_F->gotoConf.confRight.q7;
  } else {
    motion[7].FC.x  =  motion[7].IC.x  ;
    motion[8].FC.x  =  motion[8].IC.x  ;
    motion[9].FC.x  =  motion[9].IC.x  ;
    motion[10].FC.x =  motion[10].IC.x ;
    motion[11].FC.x =  motion[11].IC.x ;
    motion[12].FC.x =  motion[12].IC.x ;
    motion[13].FC.x =  motion[13].IC.x ;
  }

  for(int i=0; i<LWR_JOINTS_USER; i++) {
    motion[i].FC.v = 0.0;
    motion[i].FC.a = 0.0;
  }

#ifdef LWR_DEBUG
  //std::cout << "IC : ";
  //for (int i = 0; i < LWR_JOINTS_USER; i++)  std::cout << motion[i].IC.x << "   ";
  //std::cout << std::endl;
  //
  //std::cout << "FC : ";
  //for (int i = 0; i < LWR_JOINTS_USER; i++)  std::cout << motion[i].FC.x << "   ";
  //std::cout << std::endl;
#endif

  // Compute soft motion between initial and final conditions
  SM_STATUS resp =   sm_ComputeSoftMotionPointToPoint_gen(LWR_JOINTS_USER, currentMotion.q_J_max, currentMotion.q_A_max, currentMotion.q_V_max, motion);
  if (resp != SM_OK) {
    printf("ERROR: Q interpolation failed (sm_ComputeSoftMotionPointToPoint_gen funcion)\n");
    *report = S_lwr_SOFTMOTION_ERROR;
    return ETHER;
  }


  
  SM_COND IC[LWR_JOINTS_USER][SM_NB_SEG];
  double  Time[LWR_JOINTS_USER][SM_NB_SEG];
  double  Jerk[LWR_JOINTS_USER][SM_NB_SEG];
  SM_SEG seg;
  std::vector<SM_SEG> traj_i;
  std::vector<double> I(3);
  std::vector<double> T(SM_NB_SEG);
  std::vector<double> J(SM_NB_SEG);
  std::vector<double> t(1);
  std::vector<double> a(1);
  std::vector<double> v(1);
  std::vector<double> x(1);

  for (int i = 0; i < LWR_JOINTS_USER; i++)
    {

      I[0] = motion[i].IC.a;
      I[1] = motion[i].IC.v;
      I[2] = motion[i].IC.x;

      T[0] = motion[i].Times.Tjpa;
      T[1] = motion[i].Times.Taca;
      T[2] = motion[i].Times.Tjna;
      T[3] = motion[i].Times.Tvc;
      T[4] = motion[i].Times.Tjnb;
      T[5] = motion[i].Times.Tacb;
      T[6] = motion[i].Times.Tjpb;

      J[0] =   motion[i].Dir*motion[i].jerk.J1;
      J[1] =   0.0;
      J[2] = - motion[i].Dir*motion[i].jerk.J1;
      J[3] =   0.0;
      J[4] = - motion[i].Dir*motion[i].jerk.J1;
      J[5] =   0.0;
      J[6] =   motion[i].Dir*motion[i].jerk.J1;

      Time[i][0] = T[0];
      Time[i][1] = T[1];
      Time[i][2] = T[2];
      Time[i][3] = T[3];
      Time[i][4] = T[4];
      Time[i][5] = T[5];
      Time[i][6] = T[6];

      Jerk[i][0] = J[0];
      Jerk[i][1] = J[1];
      Jerk[i][2] = J[2];
      Jerk[i][3] = J[3];
      Jerk[i][4] = J[4];
      Jerk[i][5] = J[5];
      Jerk[i][6] = J[6];

      IC[i][0].a = I[0];
      IC[i][0].v = I[1];
      IC[i][0].x = I[2];
      t[0] = 0.0;
      for (int smp = 0; smp < SM_NB_SEG ; smp++) {
	resp = sm_AVX_TimeVar(I, T, J, t, a, v, x);
	IC[i][smp].a = a[0];
	IC[i][smp].v = v[0];
	IC[i][smp].x = x[0];     
	if (resp != SM_OK) {
	  printf("ERROR: Q interpolation failed (sm_AVX_TimeVar funcion)\n");
	  *report = S_lwr_SOFTMOTION_ERROR;
	  return ETHER;
	}
	t[0] += Time[i][smp];
      }
    }

  currentMotion.traj.clear();
  for (int i = 0; i < LWR_JOINTS_USER; i++) {
    traj_i.clear();
    for (int j = 0; j < SM_NB_SEG; j++) {
      seg.IC = IC[i][j];
      seg.time = Time[i][j];
      seg.jerk = Jerk[i][j];
      traj_i.push_back(seg);     
    }
    currentMotion.traj.traj.push_back(traj_i);
    currentMotion.traj.qStart.push_back(motion[i].IC.x);
    currentMotion.traj.qGoal.push_back(motion[i].FC.x);
  }
  currentMotion.traj.computeTimeOnTraj();

  currentMotion.duration = currentMotion.traj.getDuration();

#ifdef LWR_DEBUG
  //currentMotion.traj.print();
#endif
  //  currentMotion.cntrlMode = LWR_OPEN_LOOP_CNTRL;
  SDI_F->armLeft.cntrl_mode = LWR_TRAJECTORY_CNTRL;
  SDI_F->armRight.cntrl_mode = LWR_TRAJECTORY_CNTRL;
  setMotionStart(LWR_POSITION);

  printf("lwrGotoQ EXEC OK\n");
  printf("currentMotion.timeScaleCond.a %f currentMotion.timeScaleCond.v %f currentMotion.timeScaleCond.x %f\n",currentMotion.timeScaleCond.a, currentMotion.timeScaleCond.v, currentMotion.timeScaleCond.x);

  return END;
}

/* lwrGotoQEnd  -  codel END of GotoQ
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoQEnd(LWR_GOTO_CONF_STR *gotoConf, int *report)
{
  if ((gotoConf->wait_done == GEN_TRUE) &&
      ((currentMotion.state == LWR_STARTED) || (currentMotion.state == LWR_MOVING)))
    return END;
  
  fillMotionReport(currentMotion.error, report);
  return ETHER;
}


int lwrComputeSoftMotionCmd() {
  double criticalLength;
  SM_MOTION_MONO motion[LWR_JOINTS_USER];
  SM_LIMITS limitsGoto;
  int TrajectoryType[LWR_JOINTS_USER];
  SM_TIMES T_Jerk[LWR_JOINTS_USER];
  double maxTime = 0.0;
  double sum = 0.0;
  double maxAxis = 0;

  memset(motion, 0, LWR_JOINTS_USER*sizeof(SM_MOTION_MONO));
  for(int i=0; i<LWR_JOINTS_USER; i++) {
    motion[i].IC.x = SDI_F->motion.q[i] ;
    motion[i].IC.v = SDI_F->motion.vq[i] ;
    motion[i].IC.a = SDI_F->motion.aq[i] ;
  }

  if(SDI_F->armLeft.is_used == GEN_TRUE) {
    printf("lwrGotoQAbsMain Arm Left is_used\n");
    motion[0].FC.x = SDI_F->gotoConf.confLeft.q1;
    motion[1].FC.x = SDI_F->gotoConf.confLeft.q2;
    motion[2].FC.x = SDI_F->gotoConf.confLeft.q3;
    motion[3].FC.x = SDI_F->gotoConf.confLeft.q4;
    motion[4].FC.x = SDI_F->gotoConf.confLeft.q5;
    motion[5].FC.x = SDI_F->gotoConf.confLeft.q6;
    motion[6].FC.x = SDI_F->gotoConf.confLeft.q7;
  } else {
    motion[0].FC.x = motion[0].IC.x;
    motion[1].FC.x = motion[1].IC.x;
    motion[2].FC.x = motion[2].IC.x;
    motion[3].FC.x = motion[3].IC.x;
    motion[4].FC.x = motion[4].IC.x;
    motion[5].FC.x = motion[5].IC.x;
    motion[6].FC.x = motion[6].IC.x;
  }
  if(SDI_F->armRight.is_used == GEN_TRUE) {
    printf("lwrGotoQAbsMain Arm Right is_used\n");
    motion[7].FC.x  = SDI_F->gotoConf.confRight.q1;
    motion[8].FC.x  = SDI_F->gotoConf.confRight.q2;
    motion[9].FC.x  = SDI_F->gotoConf.confRight.q3;
    motion[10].FC.x = SDI_F->gotoConf.confRight.q4;
    motion[11].FC.x = SDI_F->gotoConf.confRight.q5;
    motion[12].FC.x = SDI_F->gotoConf.confRight.q6;
    motion[13].FC.x = SDI_F->gotoConf.confRight.q7;
  } else {
    motion[7].FC.x  =  motion[7].IC.x  ;
    motion[8].FC.x  =  motion[8].IC.x  ;
    motion[9].FC.x  =  motion[9].IC.x  ;
    motion[10].FC.x =  motion[10].IC.x ;
    motion[11].FC.x =  motion[11].IC.x ;
    motion[12].FC.x =  motion[12].IC.x ;
    motion[13].FC.x =  motion[13].IC.x ;
  }

  for(int i=0; i<LWR_JOINTS_USER; i++) {
    motion[i].FC.v = 0.0;
    motion[i].FC.a = 0.0;
  }


  maxTime = 0.0;
  sum = 0.0;
  maxAxis = 0;

  for(int i=0; i<LWR_JOINTS_USER; i++) {
    limitsGoto.maxJerk = currentMotion.q_J_max[i];
    limitsGoto.maxAcc  = currentMotion.q_A_max[i];
    limitsGoto.maxVel  = currentMotion.q_V_max[i];

    SM_STATUS resp = sm_ComputeSoftMotion(motion[i].IC, motion[i].FC, limitsGoto, &(T_Jerk[i]), &(TrajectoryType[i]));
    if (resp != SM_OK) {
      printf("ERROR sm_ComputeSoftMotion\n");
    }
    sm_sum_motionTimes(&(T_Jerk[i]), &sum);

    if(sum > maxTime) {
      maxTime = sum;
      maxAxis = i;
    }
  }
  for(int i=0; i<LWR_JOINTS_USER; i++) {
    sm_adjustMotionWith3seg(motion[i].IC, motion[i].FC, maxTime, &(motion[i]));
  }
  

}

/*------------------------------------------------------------------------
 * GotoQRel
 *
 * Description: Set relative deplacements for all joints
 *
 * Reports:      OK
 *              S_lwr_NOT_INITIALIZED
 *              S_lwr_WRONG_VALUE
 *              S_lwr_SOFTMOTION_ERROR
 *              S_lwr_FRI_MONITOR_MODE
 */

/* lwrGotoQRelStart  -  codel START of GotoQRel
   Returns:  START EXEC END ETHER FAIL ZOMBIE */

ACTIVITY_EVENT
lwrGotoQRelStart(LWR_GOTO_CONF_STR *gotoConf, int *report)
{
  if(SDI_F->status.simulation == GEN_TRUE) {
    return EXEC;
  }
  
#ifdef LWR_FRI
  ACTIVITY_EVENT eventLeft, eventRight;
  if(SDI_F->armLeft.is_used == GEN_TRUE) {
    eventLeft = requestFriStateCmd(report, LWR_ARM_LEFT, LWR_CTRL_POSITION);
  } else {
    eventLeft = EXEC;
  }
  if(SDI_F->armRight.is_used == GEN_TRUE) {
    eventRight = requestFriStateCmd(report, LWR_ARM_RIGHT, LWR_CTRL_POSITION);
  } else {
    eventRight = EXEC;
  }

  if(eventLeft == EXEC && eventRight == EXEC) {
    printf("lwrGotoQRelStart OK\n");
    return EXEC;
  }
  if(eventLeft == ETHER || eventRight == ETHER) { 
    return ETHER;
  }
  if(eventLeft == START || eventRight == START) {
    return START;
  }
#else
  return EXEC;
#endif
}

/* lwrGotoQRelMain  -  codel EXEC of GotoQRel
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoQRelMain(LWR_GOTO_CONF_STR *gotoConf, int *report)
{
  /* Send new absolute angles for all robot joints */
  if(SDI_F->armLeft.is_used == GEN_TRUE) {
    gotoConf->confLeft.q1 += SDI_F->armLeft.currConf.q1;
    gotoConf->confLeft.q2 += SDI_F->armLeft.currConf.q2;
    gotoConf->confLeft.q3 += SDI_F->armLeft.currConf.q3;
    gotoConf->confLeft.q4 += SDI_F->armLeft.currConf.q4;
    gotoConf->confLeft.q5 += SDI_F->armLeft.currConf.q5;
    gotoConf->confLeft.q6 += SDI_F->armLeft.currConf.q6;
    gotoConf->confLeft.q7 += SDI_F->armLeft.currConf.q7;
  } else {
    gotoConf->confLeft.q1 = SDI_F->armLeft.currConf.q1;
    gotoConf->confLeft.q2 = SDI_F->armLeft.currConf.q2;
    gotoConf->confLeft.q3 = SDI_F->armLeft.currConf.q3;
    gotoConf->confLeft.q4 = SDI_F->armLeft.currConf.q4;
    gotoConf->confLeft.q5 = SDI_F->armLeft.currConf.q5;
    gotoConf->confLeft.q6 = SDI_F->armLeft.currConf.q6;
    gotoConf->confLeft.q7 = SDI_F->armLeft.currConf.q7;
  }

  if(SDI_F->armRight.is_used == GEN_TRUE) {
    gotoConf->confRight.q1 += SDI_F->armRight.currConf.q1;
    gotoConf->confRight.q2 += SDI_F->armRight.currConf.q2;
    gotoConf->confRight.q3 += SDI_F->armRight.currConf.q3;
    gotoConf->confRight.q4 += SDI_F->armRight.currConf.q4;
    gotoConf->confRight.q5 += SDI_F->armRight.currConf.q5;
    gotoConf->confRight.q6 += SDI_F->armRight.currConf.q6;
    gotoConf->confRight.q7 += SDI_F->armRight.currConf.q7;
  } else {
    gotoConf->confRight.q1 = SDI_F->armRight.currConf.q1;
    gotoConf->confRight.q2 = SDI_F->armRight.currConf.q2;
    gotoConf->confRight.q3 = SDI_F->armRight.currConf.q3;
    gotoConf->confRight.q4 = SDI_F->armRight.currConf.q4;
    gotoConf->confRight.q5 = SDI_F->armRight.currConf.q5;
    gotoConf->confRight.q6 = SDI_F->armRight.currConf.q6;
    gotoConf->confRight.q7 = SDI_F->armRight.currConf.q7;
  }
  return lwrGotoQMain(gotoConf, report);
}

/* lwrGotoQRelEnd  -  codel END of GotoQRel
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoQRelEnd(LWR_GOTO_CONF_STR *gotoConf, int *report)
{
  if ((gotoConf->wait_done == GEN_TRUE) &&
      ((currentMotion.state == LWR_STARTED) || (currentMotion.state == LWR_MOVING)))
    return END;

  fillMotionReport(currentMotion.error, report);
  return ETHER;
}


/*------------------------------------------------------------------------
 * GotoXAbs
 *
 * Description: Send a new absolute position/orientation for the tool
 *
 * Reports:      OK
 *              S_lwr_NOT_INITIALIZED
 *              S_lwr_WRONG_VALUE
 *              S_lwr_SOFTMOTION_ERROR
 *              S_lwr_FRI_MONITOR_MODE
 *              S_lwr_MGI_CALC_ERROR
 *              S_lwr_MGI_SINGULARITY
 *              S_lwr_JNT_OUT_OF_RANGE
 *              S_lwr_COLLISION_RISK
 */

/* lwrGotoXAbsStart  -  codel START of GotoXAbs
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoXAbsStart(LWR_GOTO_POSE_STR *gotoPose, int *report)
{
  if(SDI_F->status.simulation == GEN_TRUE) {
    return EXEC;
  }
  
#ifdef LWR_FRI
  ACTIVITY_EVENT eventLeft, eventRight;
  if(SDI_F->armLeft.is_used == GEN_TRUE) {
    eventLeft = requestFriStateCmd(report, LWR_ARM_LEFT, LWR_CTRL_CART_IMP);
  } else {
    eventLeft = EXEC;
  }
  if(SDI_F->armRight.is_used == GEN_TRUE) {
    eventRight = requestFriStateCmd(report, LWR_ARM_RIGHT, LWR_CTRL_CART_IMP);
  } else {
    eventRight = EXEC;
  }

  if(eventLeft == EXEC && eventRight == EXEC) {
    printf("lwrGotoXAbsStart OK\n");
    return EXEC;
  }
  if(eventLeft == ETHER || eventRight == ETHER) { 
    return ETHER;
  }
  if(eventLeft == START || eventRight == START) {
    if(eventLeft == START ) {
      printf("lwrGotoXAbsStart waiting for arm Left\n");
    }
    if(eventRight == START ) {
      printf("lwrGotoXAbsStart waiting for arm Right\n");
    }
    return START;
  }
#else
  return EXEC;
#endif
}

/* lwrGotoXAbsMain  -  codel EXEC of GotoXAbs
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoXAbsMain(LWR_GOTO_POSE_STR *gotoPose, int *report)
{
   Gb_dep gotoDep = gotoPose->poseRight;
   LWR_GOTO_CONF_STR gotoConf;
   if (! normalize( gotoDep.rx,  gotoDep.ry,  gotoDep.rz,
                   &gotoDep.rx, &gotoDep.ry, &gotoDep.rz)) {
     printf("ERROR: null vector provided for tool orientation\n");
     *report = S_lwr_WRONG_VALUE;
     return ETHER;
   }
   
   Gb_dep dep_Rt; 
   Gb_th  th_goto;
   Gb_th th_Rt, th_R8, th_8_tool_inv;
   Gb_th th_invBase;

   Gb_dep_th(&gotoDep, &th_goto);
   Gb_th_print(&th_goto, "Th_goto");
   Gb_th_inverse(&SDI_F->armRight.th_base_0, &th_invBase);
   Gb_th_produit(&th_invBase, &th_goto, &th_Rt);


   Gb_th_inverse(&SDI_F->armRight.th_8_tool, &th_8_tool_inv);
   Gb_th_produit(&th_Rt, &th_8_tool_inv, &th_R8);
   
   Gb_th_print(&th_Rt, "Trt");
   Gb_th_print(&th_R8, "Tr8");
  
  int e1, e2, e3;
  kukaLBR_gete1e2e3(LWR_R3, LWR_R5, &SDI_F->armRight.currConf, &e1, &e2, &e3);
  
  Gb_q7 q;
  Gb_statusMGI result =
    kukaLBR_mgi_q_e(&th_R8, &SDI_F->armRight.currConf, LWR_R3, LWR_R5, LWR_MGI_EPSILON,
		    e1, e2, e3, &q);
  
  switch (result) {
  case MGI_ERROR:
    printf("WARN: MGI calculation failed. Coldman model is probably wrong\n");
   return ETHER;
    break;
  case MGI_SINGULAR:
    printf("WARN: Singularity on MGI calculation. Colman model is probably wrong\n");
    return ETHER;
    break;
  case MGI_APPROXIMATE:
    printf("WARN: Approximate MGI calculation. Colman model can be inexact\n");
    return ETHER;
    break;
  case MGI_OK:
    printf("INFO: MGI OK\n");       
  }

  /* Send new absolute angles for all robot joints */
  if(SDI_F->armLeft.is_used == GEN_TRUE) {
    gotoConf.confLeft.q1 = SDI_F->armLeft.currConf.q1;
    gotoConf.confLeft.q2 = SDI_F->armLeft.currConf.q2;
    gotoConf.confLeft.q3 = SDI_F->armLeft.currConf.q3;
    gotoConf.confLeft.q4 = SDI_F->armLeft.currConf.q4;
    gotoConf.confLeft.q5 = SDI_F->armLeft.currConf.q5;
    gotoConf.confLeft.q6 = SDI_F->armLeft.currConf.q6;
    gotoConf.confLeft.q7 = SDI_F->armLeft.currConf.q7;
  } else {
    gotoConf.confLeft.q1 = SDI_F->armLeft.currConf.q1;
    gotoConf.confLeft.q2 = SDI_F->armLeft.currConf.q2;
    gotoConf.confLeft.q3 = SDI_F->armLeft.currConf.q3;
    gotoConf.confLeft.q4 = SDI_F->armLeft.currConf.q4;
    gotoConf.confLeft.q5 = SDI_F->armLeft.currConf.q5;
    gotoConf.confLeft.q6 = SDI_F->armLeft.currConf.q6;
    gotoConf.confLeft.q7 = SDI_F->armLeft.currConf.q7;
  }

  if(SDI_F->armRight.is_used == GEN_TRUE) {
    gotoConf.confRight.q1 = q.q1;
    gotoConf.confRight.q2 = q.q2;
    gotoConf.confRight.q3 = q.q3;
    gotoConf.confRight.q4 = q.q4;
    gotoConf.confRight.q5 = q.q5;
    gotoConf.confRight.q6 = q.q6;
    gotoConf.confRight.q7 = q.q7;
  } else {
    gotoConf.confRight.q1 = SDI_F->armRight.currConf.q1;
    gotoConf.confRight.q2 = SDI_F->armRight.currConf.q2;
    gotoConf.confRight.q3 = SDI_F->armRight.currConf.q3;
    gotoConf.confRight.q4 = SDI_F->armRight.currConf.q4;
    gotoConf.confRight.q5 = SDI_F->armRight.currConf.q5;
    gotoConf.confRight.q6 = SDI_F->armRight.currConf.q6;
    gotoConf.confRight.q7 = SDI_F->armRight.currConf.q7;
  }
  return lwrGotoQMain(&gotoConf, report);

//   
//   // Get current pose
//   Gb_dep currDep;
//   //   fri2dep_cartPos(SDI_F->armRigh.currentCartPos, &currDep);
//
//   currDep = SDI_F->armRight.currPose;
//   // Send a new absolute position/orientation for the tool
//   SM_MOTION_MONO motion[LWR_SM_CART_COORDS];
//   memset(motion, 0, LWR_SM_CART_COORDS*sizeof(SM_MOTION_MONO));
// 
//   // IC.a, IC.v, FC.a and FC.v are already 0, as desired
//   motion[0].IC.x = currDep.x;
//   motion[1].IC.x = currDep.y;
//   motion[2].IC.x = currDep.z;
//   motion[3].IC.x = currDep.a * currDep.rx;
//   motion[4].IC.x = currDep.a * currDep.ry;
//   motion[5].IC.x = currDep.a * currDep.rz;
// 
//   motion[0].FC.x = gotoDep.x;
//   motion[1].FC.x = gotoDep.y;
//   motion[2].FC.x = gotoDep.z;
//   motion[3].FC.x = gotoDep.a * gotoDep.rx;
//   motion[4].FC.x = gotoDep.a * gotoDep.ry;
//   motion[5].FC.x = gotoDep.a * gotoDep.rz;
// 
// #ifdef LWR_DEBUG
//   std::cout << "IC : ";
//   for (int i = 0; i < 6; i++)  std::cout << motion[i].IC.x << "   ";
//   std::cout << std::endl;
// 
//   std::cout << "FC : ";
//   for (int i = 0; i < 6; i++)  std::cout << motion[i].FC.x << "   ";
//   std::cout << std::endl;
// #endif
//  
//    // Compute soft motion between initial and final conditions
//    SM_STATUS resp =
//      sm_ComputeSoftMotionPointToPoint_gen(LWR_SM_CART_COORDS,  currentMotion.x_J_max,  currentMotion.x_A_max,  currentMotion.x_V_max, motion);
//    if (resp != SM_OK) {
//      printf("ERROR: X interpolation failed (sm_ComputeSoftMotionPointToPoint_gen funcion)\n");
//      *report = S_lwr_SOFTMOTION_ERROR;
//      return ETHER;
//    }
//  
//    // Compute sample time, total time, samples number and time evolution
//    computeTimming(friInst->getSampleTime(), motion[0].MotionDuration, &currentMotion);
//  
//    std::vector<double> I(3);
//    std::vector<double> T(SM_NB_SEG);
//    std::vector<double> J(SM_NB_SEG);
//    std::vector<double> t(1), a(1), v(1), x(1);
//  
//    for (int i = 0; i < LWR_SM_CART_COORDS; i++)
//    {
//      I[0] = motion[i].IC.a;
//      I[1] = motion[i].IC.v;
//      I[2] = motion[i].IC.x;
//  
//      T[0] = motion[i].Times.Tjpa;
//      T[1] = motion[i].Times.Taca;
//      T[2] = motion[i].Times.Tjna;
//      T[3] = motion[i].Times.Tvc;
//      T[4] = motion[i].Times.Tjnb;
//      T[5] = motion[i].Times.Tacb;
//      T[6] = motion[i].Times.Tjpb;
//  
//      J[0] =   motion[i].Dir*motion[i].jerk.J1;
//      J[1] =   0.0;
//      J[2] = - motion[i].Dir*motion[i].jerk.J1;
//      J[3] =   0.0;
//      J[4] = - motion[i].Dir*motion[i].jerk.J1;
//      J[5] =   0.0;
//      J[6] =   motion[i].Dir*motion[i].jerk.J1;
//  
//      for (int smp = 0; smp < currentMotion.samplesNb; smp++) {
//  
//        if (smp%50 == 0) friInst->doDataExchange();  // avoid bad quality
//  
//        t[0] = currentMotion.t[smp];
//  
//        resp = sm_AVX_TimeVar(I, T, J, t, a, v, x);
//  
//        currentMotion.a[i][smp] = a[0];
//        currentMotion.v[i][smp] = v[0];
//        currentMotion.x[i][smp] = x[0];
//  
//        if (resp != SM_OK) {
//          printf("ERROR: X interpolation failed (sm_AVX_TimeVar funcion)\n");
//          *report = S_lwr_SOFTMOTION_ERROR;
//          return ETHER;
//        }
//      } // for each sample...
//    } // for each cartesian coordinate (x, y, z, a, b, c)...
//  
//    setMotionStart(LWR_CARTESIAN);
//  
//    printf("INFO: Going to absolute X:  %.2f, %.2f, %.2f, %.2f, %.2f, %.2f, %.2f...\n",
//           gotoPose->poseRight.x,  gotoPose->poseRight.y,  gotoPose->poseRight.z,
//           gotoPose->poseRight.rx, gotoPose->poseRight.ry, gotoPose->poseRight.rz, gotoPose->poseRight.a);
  
  return END;
}

/* lwrGotoXAbsEnd  -  codel END of GotoXAbs
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoXAbsEnd(LWR_GOTO_POSE_STR *gotoPose, int *report)
{
  if ((gotoPose->wait_done == GEN_TRUE) &&
      ((currentMotion.state == LWR_STARTED) || (currentMotion.state == LWR_MOVING)))
    return END;
  
  fillMotionReport(currentMotion.error, report);
  return ETHER;
}


/*------------------------------------------------------------------------
 * GotoXRel
 *
 * Description: Send a relative change on tool position/orientation
 *
 * Reports:      OK
 *              S_lwr_NOT_INITIALIZED
 *              S_lwr_WRONG_VALUE
 *              S_lwr_SOFTMOTION_ERROR
 *              S_lwr_FRI_MONITOR_MODE
 *              S_lwr_MGI_CALC_ERROR
 *              S_lwr_MGI_SINGULARITY
 *              S_lwr_JNT_OUT_OF_RANGE
 */

/* lwrGotoXRelStart  -  codel START of GotoXRel
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoXRelStart(LWR_GOTO_POSE_STR *gotoPose, int *report)
{
  // return requestFriStateCmd(report, LWR_CTRL_CART_IMP);
}

/* lwrGotoXRelMain  -  codel EXEC of GotoXRel
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoXRelMain(LWR_GOTO_POSE_STR *gotoPose, int *report)
{
  //  Gb_dep gotoDep, currDep;
  //
  //  gotoDep = gotoPose->pose;
  //  //multip. por v3 en lugar de dep  >>>   solo aplicar la rotacion de th_base_0, no la transalt
  //  //GotoHome >> usar GotoQAbs, con una config. propia para cada base
  //    
  //  // Convert the given pose (in base coordinates) into R0 system
  //  Gb_th_x_dep(&th_base_0, &gotoDep, &gotoDep);
  //Gb_dep_print(&gotoDep);
  //  
  //  // Get and normalize the current pose
  //  fri2dep_cartPos(currentCartPos, &currDep);
  //  normalize( currDep.rx,  currDep.ry,  currDep.rz,
  //            &currDep.rx, &currDep.ry, &currDep.rz);
  //  gotoDep.rx = currDep.rx;
  //  gotoDep.ry = currDep.ry;
  //  gotoDep.rz = currDep.rz;
  //  
  //  // Undo the tool transformation (from R tool to R8 system)
  //  Gb_th th_Rt, th_R8, th_8_tool_inv;
  //  Gb_dep_th(&gotoDep, &th_Rt);
  //  Gb_th_inverse(&th_8_tool, &th_8_tool_inv);
  //  Gb_th_produit(&th_Rt, &th_8_tool_inv, &th_R8);
  //  Gb_th_dep(&th_R8, &gotoDep);
  //Gb_dep_print(&gotoDep);
  //  // Send a relative change on tool position/orientation
  //  gotoDep.x += currDep.x;
  //  gotoDep.y += currDep.y;
  //  gotoDep.z += currDep.z;
  //  gotoDep.a += currDep.a;
  //Gb_dep_print(&gotoDep);
  ///*
  //  // Normalize rotation to add it to current rotation (already normalized)
  //  if (! normalize( gotoDep.rx,  gotoDep.ry,  gotoDep.rz,
  //                  &gotoDep.rx, &gotoDep.ry, &gotoDep.rz) ||
  //                 ! gotoDep.a)
  //  {
  //    // Null rotation vector and/or angle; retain the current rotation
  //    gotoDep.rx  = currDep.rx;
  //    gotoDep.ry  = currDep.ry;
  //    gotoDep.rz  = currDep.rz;
  //  }
  //  else
  //  {
  //    gotoDep.rx += currDep.rx;
  //    gotoDep.ry += currDep.ry;
  //    gotoDep.rz += currDep.rz;
  //  }
  //  */
  //  // Send a new absolute position/orientation for the tool; from now
  //  // the algorithm is the same that on the case of absolute X-motion
  //  SM_MOTION_MONO motion[LWR_SM_CART_COORDS];
  //  memset(motion, 0, LWR_SM_CART_COORDS*sizeof(SM_MOTION_MONO));
  //
  //  // IC.a, IC.v, FC.a and FC.v are already 0, as desired
  //  motion[0].IC.x = currDep.x;
  //  motion[1].IC.x = currDep.y;
  //  motion[2].IC.x = currDep.z;
  //  motion[3].IC.x = currDep.a * currDep.rx;
  //  motion[4].IC.x = currDep.a * currDep.ry;
  //  motion[5].IC.x = currDep.a * currDep.rz;
  //
  //  motion[0].FC.x = gotoDep.x;
  //  motion[1].FC.x = gotoDep.y;
  //  motion[2].FC.x = gotoDep.z;
  //  motion[3].FC.x = gotoDep.a * gotoDep.rx;
  //  motion[4].FC.x = gotoDep.a * gotoDep.ry;
  //  motion[5].FC.x = gotoDep.a * gotoDep.rz;
  //
  //#ifdef LWR_DEBUG
  //  std::cout << "IC : ";
  //  for (int i = 0; i < 6; i++)  std::cout << motion[i].IC.x << "   ";
  //  std::cout << std::endl;
  //
  //  std::cout << "FC : ";
  //  for (int i = 0; i < 6; i++)  std::cout << motion[i].FC.x << "   ";
  //  std::cout << std::endl;
  //#endif
  //
  //  // Compute soft motion between initial and final conditions
  //  SM_STATUS resp =
  //      sm_ComputeSoftMotionPointToPoint_gen(LWR_SM_CART_COORDS, x_J_max, x_A_max, x_V_max, motion);
  //  if (resp != SM_OK) {
  //    printf("ERROR: X interpolation failed (sm_ComputeSoftMotionPointToPoint_gen funcion)\n");
  //    *report = S_lwr_SOFTMOTION_ERROR;
  //    return ETHER;
  //  }
  //
  //  // Compute sample time, total time, samples number and time evolution
  //  computeTimming(friInst->getSampleTime(), motion[0].MotionDuration, &currentMotion);
  //
  //  std::vector<double> I(3);
  //  std::vector<double> T(SM_NB_SEG);
  //  std::vector<double> J(SM_NB_SEG);
  //  std::vector<double> t(1), a(1), v(1), x(1);
  //
  //  for (int i = 0; i < LWR_SM_CART_COORDS; i++)
  //  {
  //    I[0] = motion[i].IC.a;
  //    I[1] = motion[i].IC.v;
  //    I[2] = motion[i].IC.x;
  //
  //    T[0] = motion[i].Times.Tjpa;
  //    T[1] = motion[i].Times.Taca;
  //    T[2] = motion[i].Times.Tjna;
  //    T[3] = motion[i].Times.Tvc;
  //    T[4] = motion[i].Times.Tjnb;
  //    T[5] = motion[i].Times.Tacb;
  //    T[6] = motion[i].Times.Tjpb;
  //
  //    J[0] =   motion[i].Dir*motion[i].jerk.J1;
  //    J[1] =   0.0;
  //    J[2] = - motion[i].Dir*motion[i].jerk.J1;
  //    J[3] =   0.0;
  //    J[4] = - motion[i].Dir*motion[i].jerk.J1;
  //    J[5] =   0.0;
  //    J[6] =   motion[i].Dir*motion[i].jerk.J1;
  //
  //    for (int smp = 0; smp < currentMotion.samplesNb; smp++) {
  //
  //      if (smp%50 == 49) friInst->doDataExchange();  // avoid bad quality
  //
  //      t[0] = currentMotion.t[smp];
  //
  //      resp = sm_AVX_TimeVar(I, T, J, t, a, v, x);
  //
  //      currentMotion.a[i][smp] = a[0];
  //      currentMotion.v[i][smp] = v[0];
  //      currentMotion.x[i][smp] = x[0];
  //
  //      if (resp != SM_OK) {
  //        printf("ERROR: X interpolation failed (sm_AVX_TimeVar funcion)\n");
  //        *report = S_lwr_SOFTMOTION_ERROR;
  //        return ETHER;
  //      }
  //    } // for each sample...
  //  } // for each cartesian coordinate (x, y, z, a, b, c)...
  //
  //  setMotionStart(LWR_CARTESIAN);
  //
  //  printf("INFO: Going to absolute X:  %.2f, %.2f, %.2f, %.2f, %.2f, %.2f, %.2f...\n",
  //         gotoPose->pose.x,  gotoPose->pose.y,  gotoPose->pose.z,
  //         gotoPose->pose.rx, gotoPose->pose.ry, gotoPose->pose.rz, gotoPose->pose.a);
  //
  //  return END;
}

/* lwrGotoXRelEnd  -  codel END of GotoXRel
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoXRelEnd(LWR_GOTO_POSE_STR *gotoPose, int *report)
{
  //  if ((gotoPose->wait_done == GEN_TRUE) &&
  //      ((currentMotion.state == LWR_STARTED) || (currentMotion.state == LWR_MOVING)))
  //    return END;
  //
  //  fillMotionReport(currentMotion.error, report);
  return ETHER;
}

/*------------------------------------------------------------------------
 * Stop
 *
 * Description: Gently stop arm
 *
 * Reports:      OK
 */

/* lwrStopMain  -  codel EXEC of Stop
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrStopMain(int *report)
{
  /* Gently stop arm by forcing the end of the current motion */
  setMotionEnd();
  *report = OK;
  printf("INFO: Motion stoped\n");

  return ETHER;
}


/*------------------------------------------------------------------------
 * TrackQ
 *
 * Description: 
 *
 * Reports:      OK
 *              S_lwr_NOT_INITIALIZED
 *              S_lwr_FRI_MONITOR_MODE
 *              S_lwr_POSTER_NOT_FOUND
 *              S_lwr_POSTER_READ_ERROR
 *              S_lwr_CMD_STOP_TRACK
 *              S_lwr_NB_POINTS
 *              S_lwr_ERROR_POSITION
 *              S_lwr_BAD_MODE
 *              S_lwr_FILE_NOT_FOUND
 *              S_lwr_ERROR_INIT_POSITION
 *              S_lwr_Q_LIMITED
 */

/* lwrTrackQStart  -  codel START of TrackQ
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrTrackQStart(LWR_TRACK_STR *trackStr, int *report)
{
  if(SDI_F->status.simulation == GEN_TRUE) {
    return EXEC;
  }
  
#ifdef LWR_FRI
  ACTIVITY_EVENT eventLeft, eventRight;
  if(SDI_F->armLeft.is_used == GEN_TRUE) {
    eventLeft = requestFriStateCmd(report, LWR_ARM_LEFT, LWR_CTRL_POSITION);
  } else {
    eventLeft = EXEC;
  }
  if(SDI_F->armRight.is_used == GEN_TRUE) {
    eventRight = requestFriStateCmd(report, LWR_ARM_RIGHT, LWR_CTRL_POSITION);
  } else {
    eventRight = EXEC;
  }

  if(eventLeft == EXEC && eventRight == EXEC) {
    printf("lwrGotoQRelStart OK\n");
    return EXEC;
  }
  if(eventLeft == ETHER || eventRight == ETHER) { 
    return ETHER;
  }
  if(eventLeft == START || eventRight == START) {
    if(eventLeft == START ) {
      printf("lwrGotoQRelStart waiting for arm Left\n");
    }
    if(eventRight == START ) {
      printf("lwrGotoQRelStart waiting for arm Right\n");
    }
    return START;
  }
#else
  return EXEC;
#endif
}

#ifdef LWR_FRI
int * fctDataExchange()
{
  lwrDoDataExchange();
  return 0;
}
#endif

/* lwrTrackQExec  -  codel EXEC of TrackQ
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrTrackQExec(LWR_TRACK_STR *trackStr, int *report)
{

  currentMotion.timeScaleCond.a = 0.0;
  currentMotion.timeScaleCond.v = SDI_F->timeScale;
  currentMotion.timeScaleCond.x = 0.0;

  switch(trackStr->trackMode) {
  case LWR_TRACK_FILE:

#ifdef LWR_FRI
    currentMotion.traj.load(trackStr->posterName.name, fctDataExchange);
#else
    currentMotion.traj.load(trackStr->posterName.name, NULL);
#endif
    
    break;

  case LWR_TRACK_POSTER:
    if(lwrSM_TRAJ_STRPosterFind (trackStr->posterName.name, &lwr_poster_q) != OK) {
      *report = S_lwr_POSTER_NOT_FOUND;
      return ETHER;
    }
    if(lwrSM_TRAJ_STRPosterRead (lwr_poster_q, &smTraj) != OK) {
      *report = S_lwr_POSTER_READ_ERROR;
      printf("ERROR: %s poster read failed", trackStr->posterName.name);
      return ETHER;
    }
    currentMotion.traj.clear();
    currentMotion.traj.importFromSM_TRAJ_STR(&smTraj);

    break;
  default:
    *report = S_lwr_BAD_MODE;
    return ETHER;
  }
 
  /* check initial position */
  Gb_q7 errorMax;
  std::vector<SM_COND> cond;
  errorMax.q1 = LWR_A1_MAX_VELOCITY/10.0;
  errorMax.q2 = LWR_A2_MAX_VELOCITY/10.0;
  errorMax.q3 = LWR_E1_MAX_VELOCITY/10.0;
  errorMax.q4 = LWR_A3_MAX_VELOCITY/10.0;
  errorMax.q5 = LWR_A4_MAX_VELOCITY/10.0;
  errorMax.q6 = LWR_A5_MAX_VELOCITY/10.0;
  errorMax.q7 = LWR_A6_MAX_VELOCITY/10.0;

  if(SDI_F->timeScale < 0.0) {
    SDI_F->timeScale = 1.0;
    printf("WARN: lwr::TrackQ set timeScale to 1.0\n");
  }
  currentMotion.traj.getMotionCond(0.0, cond);
  /* check Left arm */
  if(SDI_F->armLeft.is_used == GEN_TRUE) {
    if ((fabs(SDI_F->armLeft.currConf.q1 - cond[0+SDI_F->armLeft.indexInTraj].x) > errorMax.q1) ||
	(fabs(SDI_F->armLeft.currConf.q2 - cond[1+SDI_F->armLeft.indexInTraj].x) > errorMax.q2) ||
	(fabs(SDI_F->armLeft.currConf.q3 - cond[2+SDI_F->armLeft.indexInTraj].x) > errorMax.q3) ||
	(fabs(SDI_F->armLeft.currConf.q4 - cond[3+SDI_F->armLeft.indexInTraj].x) > errorMax.q4) ||
	(fabs(SDI_F->armLeft.currConf.q5 - cond[4+SDI_F->armLeft.indexInTraj].x) > errorMax.q5) ||
	(fabs(SDI_F->armLeft.currConf.q6 - cond[5+SDI_F->armLeft.indexInTraj].x) > errorMax.q6) ||
	(fabs(SDI_F->armLeft.currConf.q7 - cond[6+SDI_F->armLeft.indexInTraj].x) > errorMax.q7)) {
      *report = S_lwr_ERROR_INIT_POSITION;
      return ETHER;
    }
  }
  if(SDI_F->armRight.is_used == GEN_TRUE) {
    if ((fabs(SDI_F->armRight.currConf.q1 - cond[0+SDI_F->armRight.indexInTraj].x) > errorMax.q1) ||
	(fabs(SDI_F->armRight.currConf.q2 - cond[1+SDI_F->armRight.indexInTraj].x) > errorMax.q2) ||
	(fabs(SDI_F->armRight.currConf.q3 - cond[2+SDI_F->armRight.indexInTraj].x) > errorMax.q3) ||
	(fabs(SDI_F->armRight.currConf.q4 - cond[3+SDI_F->armRight.indexInTraj].x) > errorMax.q4) ||
	(fabs(SDI_F->armRight.currConf.q5 - cond[4+SDI_F->armRight.indexInTraj].x) > errorMax.q5) ||
	(fabs(SDI_F->armRight.currConf.q6 - cond[5+SDI_F->armRight.indexInTraj].x) > errorMax.q6) ||
	(fabs(SDI_F->armRight.currConf.q7 - cond[6+SDI_F->armRight.indexInTraj].x) > errorMax.q7)) {
      *report = S_lwr_ERROR_INIT_POSITION;
      return ETHER;
    }
  }
  // currentMotion.traj.print();
  currentMotion.duration = currentMotion.traj.getDuration(); 
  currentMotion.curTime = 0.0;
  SDI_F->armLeft.cntrl_mode = LWR_OPEN_LOOP_CNTRL;
  SDI_F->armRight.cntrl_mode = LWR_OPEN_LOOP_CNTRL;
  setMotionStart(LWR_POSITION);
  return END;
}

/* lwrTrackQInter  -  codel INTER of TrackQ
   Returns:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrTrackQInter(LWR_TRACK_STR *trackStr, int *report)
{
  /* ... add your code here ... */
  return ETHER;
}

/* lwrTrackQEnd  -  codel END of TrackQ
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrTrackQEnd(LWR_TRACK_STR *trackStr, int *report)
{
  if ((currentMotion.state == LWR_STARTED) || (currentMotion.state == LWR_MOVING)){
    return END;
  }
  fillMotionReport(currentMotion.error, report);
  return ETHER;
}


/*------------------------------------------------------------------------
 * PlayFile
 *
 * Description: 
 *
 * Reports:      OK
 *              S_lwr_NOT_INITIALIZED
 *              S_lwr_FRI_MONITOR_MODE
 *              S_lwr_POSTER_NOT_FOUND
 *              S_lwr_POSTER_READ_ERROR
 *              S_lwr_CMD_STOP_TRACK
 *              S_lwr_NB_POINTS
 *              S_lwr_ERROR_POSITION
 *              S_lwr_BAD_MODE
 *              S_lwr_FILE_NOT_FOUND
 *              S_lwr_ERROR_INIT_POSITION
 *              S_lwr_Q_LIMITED
 */

/*
 * Utility function to read points from file
 */
STATUS lwrPlayFileReadFile(char* fileName, int *report)
{
  FILE *fileptr = NULL;
  int indexTrajectory = 0;

  /* Open File Pointer */
  if ((fileptr = fopen(fileName, "r")) == NULL) {
    printf("ERROR: open track file %s failed\n", fileName);
    *report = S_lwr_FILE_NOT_FOUND;
    return ERROR;
  }
  /* Read File Variables */
  while (!feof(fileptr)) {
    // in the file there are 7dof (arm) + 17dof (hand) + 1 int localPathId*/
    fscanf(fileptr, "%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%d\n",
	   &(lwrPositionsPoster.armRight[indexTrajectory].q1),
	   &(lwrPositionsPoster.armRight[indexTrajectory].q2),
	   &(lwrPositionsPoster.armRight[indexTrajectory].q3),
	   &(lwrPositionsPoster.armRight[indexTrajectory].q4),
	   &(lwrPositionsPoster.armRight[indexTrajectory].q5),
	   &(lwrPositionsPoster.armRight[indexTrajectory].q6),
	   &(lwrPositionsPoster.armRight[indexTrajectory].q7),
	   &(lwrPositionsPoster.handRight[indexTrajectory].thumbRotation),
	   &(lwrPositionsPoster.handRight[indexTrajectory].thumb.base),
	   &(lwrPositionsPoster.handRight[indexTrajectory].thumb.phaProx),
	   &(lwrPositionsPoster.handRight[indexTrajectory].thumb.phaMid),
	   &(lwrPositionsPoster.handRight[indexTrajectory].thumb.phaDist),
	   &(lwrPositionsPoster.handRight[indexTrajectory].fore.base),
	   &(lwrPositionsPoster.handRight[indexTrajectory].fore.phaProx),
	   &(lwrPositionsPoster.handRight[indexTrajectory].fore.phaMid),
	   &(lwrPositionsPoster.handRight[indexTrajectory].fore.phaDist),
	   &(lwrPositionsPoster.handRight[indexTrajectory].middle.base),
	   &(lwrPositionsPoster.handRight[indexTrajectory].middle.phaProx),
	   &(lwrPositionsPoster.handRight[indexTrajectory].middle.phaMid),
	   &(lwrPositionsPoster.handRight[indexTrajectory].middle.phaDist),
	   &(lwrPositionsPoster.handRight[indexTrajectory].ring.base),
	   &(lwrPositionsPoster.handRight[indexTrajectory].ring.phaProx),
	   &(lwrPositionsPoster.handRight[indexTrajectory].ring.phaMid),
	   &(lwrPositionsPoster.handRight[indexTrajectory].ring.phaDist),
	   &(lwrPositionsPoster.lp[indexTrajectory]));

    /* Testing limits with an epsilon would probably be better */
    if ((lwrPositionsPoster.armRight[indexTrajectory].q1 < LWR_A1_RANGE_MIN) ||
        (lwrPositionsPoster.armRight[indexTrajectory].q1 > LWR_A1_RANGE_MAX) ||
        (lwrPositionsPoster.armRight[indexTrajectory].q2 < LWR_A2_RANGE_MIN) ||
        (lwrPositionsPoster.armRight[indexTrajectory].q2 > LWR_A2_RANGE_MAX) ||
        (lwrPositionsPoster.armRight[indexTrajectory].q3 < LWR_E1_RANGE_MIN) ||
        (lwrPositionsPoster.armRight[indexTrajectory].q3 > LWR_E1_RANGE_MAX) ||
        (lwrPositionsPoster.armRight[indexTrajectory].q4 < LWR_A3_RANGE_MIN) ||
        (lwrPositionsPoster.armRight[indexTrajectory].q4 > LWR_A3_RANGE_MAX) ||
        (lwrPositionsPoster.armRight[indexTrajectory].q5 < LWR_A4_RANGE_MIN) ||
        (lwrPositionsPoster.armRight[indexTrajectory].q5 > LWR_A4_RANGE_MAX) ||
        (lwrPositionsPoster.armRight[indexTrajectory].q6 < LWR_A5_RANGE_MIN) ||
        (lwrPositionsPoster.armRight[indexTrajectory].q6 > LWR_A5_RANGE_MAX) ||
        (lwrPositionsPoster.armRight[indexTrajectory].q7 < LWR_A6_RANGE_MIN) ||
        (lwrPositionsPoster.armRight[indexTrajectory].q7 > LWR_A6_RANGE_MAX) ) {
      printf("ERROR: Q value out of bounds");
      *report= S_lwr_Q_LIMITED;
      return ERROR;
    }

    indexTrajectory++;
    if (indexTrajectory >= LWR_TRACK_MAX_SIZE) {
      printf("ERROR: Track file too long");
      *report= S_lwr_NB_POINTS;
      fclose(fileptr);
      return ERROR;
    }
#ifdef LWR_FRI
    if (indexTrajectory%100 == 99) lwrDoDataExchange();
#endif
  }
  fclose(fileptr);
  lwrPositionsPoster.nb_points = indexTrajectory;

  /* We force the choice of the error parameters ... */
  lwrPositionsPoster.errorMax.q1 = 0.005;
  lwrPositionsPoster.errorMax.q2 = 0.005;
  lwrPositionsPoster.errorMax.q3 = 0.005;
  lwrPositionsPoster.errorMax.q4 = 0.005;
  lwrPositionsPoster.errorMax.q5 = 0.005;
  lwrPositionsPoster.errorMax.q6 = 0.005;
  lwrPositionsPoster.errorMax.q7 = 0.005;
  /*
    fprintf(stdout, "lwrTrackQReadFile: list of read positions:\n");
    for (int i = 0; i < indexTrajectory; i++) {
    fprintf(stdout,"%4d %2.5f %2.5f %2.5f %2.5f %2.5f %2.5f %2.5f\n",
    i,
    lwrPositionsPoster.positions[i].q1,
    lwrPositionsPoster.positions[i].q2,
    lwrPositionsPoster.positions[i].q3,
    lwrPositionsPoster.positions[i].q4,
    lwrPositionsPoster.positions[i].q5,
    lwrPositionsPoster.positions[i].q6,
    lwrPositionsPoster.positions[i].q7);
    }*/

  return OK;
}




/* lwrPlayFileStart  -  codel START of PlayFile
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrPlayFileStart(LWR_TRACK_STR *trackStr, int *report)
{
  if(SDI_F->status.simulation == GEN_TRUE) {
    return EXEC;
  }
  
#ifdef LWR_FRI
  ACTIVITY_EVENT eventLeft, eventRight;
  if(SDI_F->armLeft.is_used == GEN_TRUE) {
    eventLeft = requestFriStateCmd(report, LWR_ARM_LEFT, LWR_CTRL_POSITION);
  } else {
    eventLeft = EXEC;
  }
  if(SDI_F->armRight.is_used == GEN_TRUE) {
    eventRight = requestFriStateCmd(report, LWR_ARM_RIGHT, LWR_CTRL_POSITION);
  } else {
    eventRight = EXEC;
  }

  if(eventLeft == EXEC && eventRight == EXEC) {
    printf("lwrGotoQRelStart OK\n");
    return EXEC;
  }
  if(eventLeft == ETHER || eventRight == ETHER) { 
    return ETHER;
  }
  if(eventLeft == START || eventRight == START) {
    if(eventLeft == START ) {
      printf("lwrGotoQRelStart waiting for arm Left\n");
    }
    if(eventRight == START ) {
      printf("lwrGotoQRelStart waiting for arm Right\n");
    }
    return START;
  }
#else
  return EXEC;
#endif
}

/* lwrPlayFileMain  -  codel EXEC of PlayFile
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrPlayFileMain(LWR_TRACK_STR *trackStr, int *report)
{
  Gb_q7* qPoster = NULL;
  STATUS status;
  long counter;

  switch(trackStr->trackMode) {
  case LWR_TRACK_FILE:
    if (lwrPlayFileReadFile(trackStr->posterName.name, report) != OK) {
      return ETHER;
    }
    printf("lwrTrackQReadFile OK\n");

    lwrPositionsPoster.errorMax.q1 = LWR_A1_MAX_VELOCITY/100.0;
    lwrPositionsPoster.errorMax.q2 = LWR_A2_MAX_VELOCITY/100.0; 
    lwrPositionsPoster.errorMax.q3 = LWR_E1_MAX_VELOCITY/100.0; 
    lwrPositionsPoster.errorMax.q4 = LWR_A3_MAX_VELOCITY/100.0; 
    lwrPositionsPoster.errorMax.q5 = LWR_A4_MAX_VELOCITY/100.0; 
    lwrPositionsPoster.errorMax.q6 = LWR_A5_MAX_VELOCITY/100.0;
    lwrPositionsPoster.errorMax.q7 = LWR_A6_MAX_VELOCITY/100.0;
    lwr_poster_q = (POSTER_ID)1;
    break;
  case LWR_TRACK_POSTER:
    *report = S_lwr_BAD_MODE;
    return ETHER;
    break;
  default:
    *report = S_lwr_BAD_MODE;
    return ETHER;
  }

  lwrPositionsPoster.track_position_index = 0;
  if (lwrPositionsPoster.nb_points <  0 ||
      lwrPositionsPoster.nb_points >= LWR_TRACK_MAX_SIZE) {
    *report = S_lwr_NB_POINTS;
    return ETHER;
  }

  lwrPositionsPoster.errorMax.q1 = LWR_A1_MAX_VELOCITY/10.0;
  lwrPositionsPoster.errorMax.q2 = LWR_A2_MAX_VELOCITY/10.0;
  lwrPositionsPoster.errorMax.q3 = LWR_E1_MAX_VELOCITY/10.0;
  lwrPositionsPoster.errorMax.q4 = LWR_A3_MAX_VELOCITY/10.0;
  lwrPositionsPoster.errorMax.q5 = LWR_A4_MAX_VELOCITY/10.0;
  lwrPositionsPoster.errorMax.q6 = LWR_A5_MAX_VELOCITY/10.0;
  lwrPositionsPoster.errorMax.q7 = LWR_A6_MAX_VELOCITY/10.0;

  /* Check the initial positions */
  qPoster = &(lwrPositionsPoster.armRight[lwrPositionsPoster.track_position_index]);
  if ((fabs(SDI_F->armRight.currConf.q1 - qPoster->q1) > lwrPositionsPoster.errorMax.q1) ||
      (fabs(SDI_F->armRight.currConf.q2 - qPoster->q2) > lwrPositionsPoster.errorMax.q2) ||
      (fabs(SDI_F->armRight.currConf.q3 - qPoster->q3) > lwrPositionsPoster.errorMax.q3) ||
      (fabs(SDI_F->armRight.currConf.q4 - qPoster->q4) > lwrPositionsPoster.errorMax.q4) ||
      (fabs(SDI_F->armRight.currConf.q5 - qPoster->q5) > lwrPositionsPoster.errorMax.q5) ||
      (fabs(SDI_F->armRight.currConf.q6 - qPoster->q6) > lwrPositionsPoster.errorMax.q6) ||
      (fabs(SDI_F->armRight.currConf.q7 - qPoster->q7) > lwrPositionsPoster.errorMax.q7)) {
    *report = S_lwr_ERROR_INIT_POSITION;
    return ETHER;
  }

  /* Testing limits with an epsilon would probably be better */
  if ((qPoster->q1 < LWR_A1_RANGE_MIN) || (qPoster->q1 > LWR_A1_RANGE_MAX) ||
      (qPoster->q2 < LWR_A2_RANGE_MIN) || (qPoster->q2 > LWR_A2_RANGE_MAX) ||
      (qPoster->q3 < LWR_E1_RANGE_MIN) || (qPoster->q3 > LWR_E1_RANGE_MAX) ||
      (qPoster->q4 < LWR_A3_RANGE_MIN) || (qPoster->q4 > LWR_A3_RANGE_MAX) ||
      (qPoster->q5 < LWR_A4_RANGE_MIN) || (qPoster->q5 > LWR_A4_RANGE_MAX) ||
      (qPoster->q6 < LWR_A5_RANGE_MIN) || (qPoster->q6 > LWR_A5_RANGE_MAX) ||
      (qPoster->q7 < LWR_A6_RANGE_MIN) || (qPoster->q7 > LWR_A6_RANGE_MAX)) {
    *report= S_lwr_Q_LIMITED;
    return ETHER;
  }

  /* Fill currentMotion */
  // Compute sample time, total time, samples number and time evolution

#ifndef LWR_FRI
  computeTimming(LWR_PERIOD, 
		 LWR_PERIOD*lwrPositionsPoster.nb_points, &currentMotion);
#else
  if(friInstRight) {
    computeTimming(friInstRight->getSampleTime(), 
		   friInstRight->getSampleTime()*lwrPositionsPoster.nb_points, &currentMotion);
  } else {
    computeTimming(friInstLeft->getSampleTime(), 
		   friInstLeft->getSampleTime()*lwrPositionsPoster.nb_points, &currentMotion);
  }
#endif
  currentMotion.samplesNb =  currentMotion.samplesNb -1;
 

  if(currentMotion.samplesNb != lwrPositionsPoster.nb_points) {
    printf("ERROR currentMotion.samplesNb != lwrPositionsPoster.nb_points %f != %f \n",
	   currentMotion.samplesNb,lwrPositionsPoster.nb_points );
    return ETHER;
  }
  
  int OFFSET_ARM = 7;
  for (int smp = 0; smp < currentMotion.samplesNb; smp++) {
#ifdef LWR_FRI
    if (smp%50 == 49) {
      if(friInstRight) {
	friInstRight->doDataExchange();  // avoid bad quality
      }
      if(friInstLeft) {
	friInstLeft->doDataExchange();  // avoid bad quality
      }
    }

#endif
    currentMotion.lp[smp] =  lwrPositionsPoster.lp[smp];

    currentMotion.armLeft[smp].q1 =  SDI_F->armLeft.currConf.q1 ;
    currentMotion.armLeft[smp].q2 =  SDI_F->armLeft.currConf.q2 ;
    currentMotion.armLeft[smp].q3 =  SDI_F->armLeft.currConf.q3 ;
    currentMotion.armLeft[smp].q4 =  SDI_F->armLeft.currConf.q4 ;
    currentMotion.armLeft[smp].q5 =  SDI_F->armLeft.currConf.q5 ;
    currentMotion.armLeft[smp].q6 =  SDI_F->armLeft.currConf.q6 ;
    currentMotion.armLeft[smp].q7 =  SDI_F->armLeft.currConf.q7 ;
 
    currentMotion.armRight[smp].q1 =  lwrPositionsPoster.armRight[smp].q1 ;
    currentMotion.armRight[smp].q2 =  lwrPositionsPoster.armRight[smp].q2 ;
    currentMotion.armRight[smp].q3 =  lwrPositionsPoster.armRight[smp].q3 ;
    currentMotion.armRight[smp].q4 =  lwrPositionsPoster.armRight[smp].q4 ;
    currentMotion.armRight[smp].q5 =  lwrPositionsPoster.armRight[smp].q5 ;
    currentMotion.armRight[smp].q6 =  lwrPositionsPoster.armRight[smp].q6 ;
    currentMotion.armRight[smp].q7 =  lwrPositionsPoster.armRight[smp].q7 ;

    currentMotion.handRight[smp].thumbRotation  = lwrPositionsPoster.handRight[smp].thumbRotation  ;
    currentMotion.handRight[smp].thumb.base     = lwrPositionsPoster.handRight[smp].thumb.base     ;
    currentMotion.handRight[smp].thumb.phaProx  = lwrPositionsPoster.handRight[smp].thumb.phaProx  ;
    currentMotion.handRight[smp].thumb.phaMid   = lwrPositionsPoster.handRight[smp].thumb.phaMid   ;
    currentMotion.handRight[smp].thumb.phaDist  = lwrPositionsPoster.handRight[smp].thumb.phaDist  ;
    currentMotion.handRight[smp].fore.base      = lwrPositionsPoster.handRight[smp].fore.base      ;
    currentMotion.handRight[smp].fore.phaProx   = lwrPositionsPoster.handRight[smp].fore.phaProx   ;
    currentMotion.handRight[smp].fore.phaMid    = lwrPositionsPoster.handRight[smp].fore.phaMid    ;
    currentMotion.handRight[smp].fore.phaDist   = lwrPositionsPoster.handRight[smp].fore.phaDist   ;
    currentMotion.handRight[smp].middle.base    = lwrPositionsPoster.handRight[smp].middle.base    ;
    currentMotion.handRight[smp].middle.phaProx = lwrPositionsPoster.handRight[smp].middle.phaProx ;
    currentMotion.handRight[smp].middle.phaMid  = lwrPositionsPoster.handRight[smp].middle.phaMid  ;
    currentMotion.handRight[smp].middle.phaDist = lwrPositionsPoster.handRight[smp].middle.phaDist ;
    currentMotion.handRight[smp].ring.base      = lwrPositionsPoster.handRight[smp].ring.base      ;
    currentMotion.handRight[smp].ring.phaProx   = lwrPositionsPoster.handRight[smp].ring.phaProx   ;
    currentMotion.handRight[smp].ring.phaMid    = lwrPositionsPoster.handRight[smp].ring.phaMid    ;
    currentMotion.handRight[smp].ring.phaDist   = lwrPositionsPoster.handRight[smp].ring.phaDist   ;
    
  } // for each sample...
 
  currentMotion.duration = (currentMotion.samplesNb-1) * LWR_PERIOD;
 
  SDI_F->armLeft.cntrl_mode = LWR_OPEN_LOOP_TRACK_Q_CNTRL;
  SDI_F->armRight.cntrl_mode = LWR_OPEN_LOOP_TRACK_Q_CNTRL; 
  setMotionStart(LWR_POSITION);
  
  return END;
}

/* lwrPlayFileEnd  -  codel END of PlayFile
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrPlayFileEnd(LWR_TRACK_STR *trackStr, int *report)
{
  if ((currentMotion.state == LWR_STARTED) || (currentMotion.state == LWR_MOVING)){
    return END;
  }
  fillMotionReport(currentMotion.error, report);
  return ETHER;
}





/*------------------------------------------------------------------------
 * TrackX
 *
 * Description: 
 *
 * Reports:      OK
 *              S_lwr_NOT_INITIALIZED
 *              S_lwr_FRI_MONITOR_MODE
 *              S_lwr_POSTER_NOT_FOUND
 *              S_lwr_POSTER_READ_ERROR
 *              S_lwr_CMD_STOP_TRACK
 *              S_lwr_NB_POINTS
 *              S_lwr_ERROR_POSITION
 *              S_lwr_BAD_MODE
 *              S_lwr_FILE_NOT_FOUND
 *              S_lwr_ERROR_INIT_POSITION
 *              S_lwr_Q_LIMITED
 */


/*
 * Utility function to read points from file
 */
STATUS lwrTrackXReadFile(char* fileName, int *report)
{
  FILE *fileptr = NULL;
  int indexTrajectory = 0;
  
  /* Open File Pointer */
  if ((fileptr = fopen(fileName, "r")) == NULL) {
    printf("ERROR: open track file %s failed\n", fileName);
    *report = S_lwr_FILE_NOT_FOUND;
    return ERROR;
  }
  /* Read File Variables */
  while (!feof(fileptr)) {
    fscanf(fileptr, "%lf\t%lf\t%lf\n",
	   &(lwrPositionsPoster.armRight[indexTrajectory].q1),
	   &(lwrPositionsPoster.armRight[indexTrajectory].q2),
	   &(lwrPositionsPoster.armRight[indexTrajectory].q3));
    
    indexTrajectory++;
    if (indexTrajectory >= LWR_TRACK_MAX_SIZE) {
      printf("ERROR: Track file too long");
      *report= S_lwr_NB_POINTS;
      fclose(fileptr);
      return ERROR;
    }
 
#ifdef LWR_FRI   
    if (indexTrajectory%50 == 0) lwrDoDataExchange();  // avoid bad quality
#endif
  }
  fclose(fileptr);
  lwrPositionsPoster.nb_points = indexTrajectory;
  /*
    fprintf(stdout, "lwrTrackQReadFile: list of read positions:\n");
    for (int i = 0; i < indexTrajectory; i++) {
    if ((i%10 == 0) || (i == (indexTrajectory - 1)))
    fprintf(stdout,"%4d %2.5f %2.5f %2.5f\n",
    i,
    lwrPositionsPoster.armRight[i].q1,
    lwrPositionsPoster.armRight[i].q2,
    lwrPositionsPoster.armRight[i].q3);
    }
  */
  return OK;
}

/* lwrTrackXStart  -  codel START of TrackX
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrTrackXStart(LWR_TRACK_STR *trackStr, int *report)
{
  STATUS status;
  Gb_q7* qPoster;
  long counter;

#ifdef LWR_FRI     
  ACTIVITY_EVENT event = requestFriStateCmd(report, LWR_ARM_RIGHT, LWR_CTRL_CART_IMP);
  if (event != EXEC)
    return event;
#endif
  
  switch(trackStr->trackMode)
    {
    case LWR_TRACK_FILE:
      if (lwrTrackXReadFile(trackStr->posterName.name, report) != OK)
        return ETHER;
      
      lwrPositionsPoster.errorMax.q1 = LWR_MAX_LIN_VELOCITY/100.0;
      lwrPositionsPoster.errorMax.q2 = LWR_MAX_LIN_VELOCITY/100.0;
      lwrPositionsPoster.errorMax.q3 = LWR_MAX_LIN_VELOCITY/100.0;
      break;
      
    case LWR_TRACK_POSTER:
    default:
      *report = S_lwr_BAD_MODE;
      return ETHER;
    }
  
    lwrPositionsPoster.track_position_index = 0;
    if (lwrPositionsPoster.nb_points <  0 ||
        lwrPositionsPoster.nb_points >= LWR_TRACK_MAX_SIZE) {
      *report = S_lwr_NB_POINTS;
      return ETHER;
    }
  
    memcpy(initialCartPos, SDI_F->armRight.currentCartPos, sizeof(float)*LWR_CART_FRM_DIM);
  
  return EXEC;
}

float commandCartPos[LWR_CART_FRM_DIM];

/* lwrTrackXExec  -  codel EXEC of TrackX
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrTrackXExec(LWR_TRACK_STR *trackStr, int *report)
{
  GENMANIP_POSITION7_TRACK_STR positions_poster;
  Gb_q7* qPoster = NULL;
  long counter;
  STATUS status;
  Gb_th th_Rt, th_R8;
  Gb_v3 v3, v3_R8;
  
  if (lwrPositionsPoster.track_position_index >= lwrPositionsPoster.nb_points)
    return END;
  
  qPoster = &(lwrPositionsPoster.armRight[lwrPositionsPoster.track_position_index]);
  if ((fabs(qPoster->q1 - lwrPositionsPoster.armRight[lwrPositionsPoster.track_position_index-1].q1) > lwrPositionsPoster.errorMax.q1) ||
      (fabs(qPoster->q2 - lwrPositionsPoster.armRight[lwrPositionsPoster.track_position_index-1].q2) > lwrPositionsPoster.errorMax.q2) ||
      (fabs(qPoster->q3 - lwrPositionsPoster.armRight[lwrPositionsPoster.track_position_index-1].q3) > lwrPositionsPoster.errorMax.q3)) {
    printf("lwr_ERROR_POSITION = %f %f %f (max %f)\n",
	   fabs(qPoster->q1 - lwrPositionsPoster.armRight[lwrPositionsPoster.track_position_index-1].q1),
	   fabs(qPoster->q2 - lwrPositionsPoster.armRight[lwrPositionsPoster.track_position_index-1].q2),
	   fabs(qPoster->q3 - lwrPositionsPoster.armRight[lwrPositionsPoster.track_position_index-1].q3),
	   lwrPositionsPoster.errorMax.q3);
    *report = S_lwr_ERROR_POSITION;
    return END;
  }
  
  fri2th_cartPos(initialCartPos, &th_R8);
  
  v3.x = qPoster->q1;
  v3.y = qPoster->q2;
  v3.z = qPoster->q3;
  
  Gb_th_x_v3(&th_R8, &v3, &v3_R8);
  
  memcpy(commandCartPos, initialCartPos, sizeof(float)*LWR_CART_FRM_DIM);
  commandCartPos[ 3] -= v3_R8.x;   // rotation around axe Y to avoid specular
  commandCartPos[ 7] += v3_R8.y;   // effect on writting
  commandCartPos[11] -= v3_R8.z;   

  fri2th_cartPos(commandCartPos, &th_R8);
 #ifdef LWR_FRI    
  if (SDI_F->status.simulation == GEN_FALSE) {
    friInstRight->doCartesianImpedanceControl(commandCartPos, NULL, NULL, NULL, NULL, false);
  } else {
    memcpy(&(SDI_F->armRight.currentCartPos), commandCartPos, sizeof(float)*LWR_CART_FRM_DIM);
  }
#else 
  memcpy(&(SDI_F->armRight.currentCartPos), commandCartPos, sizeof(float)*LWR_CART_FRM_DIM);



  // As we are simulating the movement, we need the MGI to obtain current configuration
  // Note that we have the arm pose on Rtool system, so we need to transform it into R7
  Gb_th th_R7, th_8_tool_inv;
  
  Gb_th_inverse(&SDI_F->armRight.th_8_tool, &th_8_tool_inv);
  Gb_th_produit(&th_R8, &th_8_tool_inv, &th_R7);
  
  int e1, e2, e3;
  kukaLBR_gete1e2e3(LWR_R3, LWR_R5, &SDI_F->armRight.currConf, &e1, &e2, &e3);
  
  Gb_q7 q;
  Gb_statusMGI result =
    kukaLBR_mgi_q_e(&th_R7, &SDI_F->armRight.currConf, LWR_R3, LWR_R5, LWR_MGI_EPSILON,
		    e1, e2, e3, &q);
  //               kukaLBR_mgi_q_e(&currTh_R7, &SDI_F->pos.q, LWR_R3, LWR_R5, LWR_MGI_EPSILON, e1, e2, e3, &q);
  
  switch (result) {
  case MGI_ERROR:
    printf("WARN: MGI calculation failed. Coldman model is probably wrong\n");
    break;
  case MGI_SINGULAR:
    printf("WARN: Singularity on MGI calculation. Colman model is probably wrong\n");
    break;
  case MGI_APPROXIMATE:
  case MGI_OK:
    if (result == MGI_APPROXIMATE)
                  printf("WARN: Approximate MGI calculation. Colman model can be inexact\n");
    
    SDI_F->armRight.currentJntVals[0] = q.q1;
    SDI_F->armRight.currentJntVals[1] = q.q2;
    SDI_F->armRight.currentJntVals[2] = q.q3;
    SDI_F->armRight.currentJntVals[3] = q.q4;
    SDI_F->armRight.currentJntVals[4] = q.q5;
    SDI_F->armRight.currentJntVals[5] = q.q6;
    SDI_F->armRight.currentJntVals[6] = q.q7;
    break;
  default:
    printf("WARN: default default default default default default default  %d\n", result);
  }
  

#endif

  lwrPositionsPoster.track_position_index++;
  
  return EXEC;
  
  
  // Get current pose
  //Gb_dep currDep;
  //fri2dep_cartPos(currentCartPos, &currDep);
  //
  //
  ///* Fill currentMotion */
  //// Compute sample time, total time, samples number and time evolution
  //computeTimming(friInst->getSampleTime(),
  //      	 friInst->getSampleTime()*lwrPositionsPoster.nb_points, &currentMotion);
  //
  //currentMotion.samplesNb =  currentMotion.samplesNb -1;
  //
  //if(currentMotion.samplesNb != lwrPositionsPoster.nb_points) {
  //  printf("ERROR currentMotion.samplesNb != lwrPositionsPoster.nb_points %f != %f \n",
  //           currentMotion.samplesNb,lwrPositionsPoster.nb_points );
  //  return ETHER;
  //}
  //
  //for (int smp = 0; smp < currentMotion.samplesNb; smp++) {
  //  
  //  if (smp%50 == 49) friInst->doDataExchange();  // avoid bad quality
  //  
  //  
  //
  //  v3.x = lwrPositionsPoster.armRight[smp].q1;
  //  v3.y = lwrPositionsPoster.armRight[smp].q2;
  //  v3.z = lwrPositionsPoster.armRight[smp].q3;
  //  
  //  Gb_th_x_v3(&th_R8, &v3, &v3_R8);
  //  
  //  
  //  currentMotion.lp[smp] =  lwrPositionsPoster.lp[smp];
  //  
  //  currentMotion.a[0][smp] = 0.0;
  //  currentMotion.v[0][smp] = 0.0;
  //  currentMotion.x[0][smp] = currDep.x + v3_R8.x;
  //  
  //  currentMotion.a[1][smp] = 0.0;
  //  currentMotion.v[1][smp] = 0.0;
  //  currentMotion.x[1][smp] = currDep.y + v3_R8.y;
  //  
  //  currentMotion.a[2][smp] = 0.0;
  //  currentMotion.v[2][smp] = 0.0;
  //  currentMotion.x[2][smp] = currDep.z + v3_R8.z;
  //  
  //  currentMotion.a[3][smp] = 0.0;
  //  currentMotion.v[3][smp] = 0.0;
  //  currentMotion.x[3][smp] = currDep.a * currDep.rx;
  //  
  //  
  //  currentMotion.a[4][smp] = 0.0;
  //  currentMotion.v[4][smp] = 0.0;
  //  currentMotion.x[4][smp] = currDep.a * currDep.ry;
  //  
  //  currentMotion.a[5][smp] = 0.0;
  //  currentMotion.v[5][smp] = 0.0;
  //  currentMotion.x[5][smp] = currDep.a * currDep.rz;
  //    
  //} // for each sample...
  //
  //
  //setMotionStart(LWR_CARTESIAN);
  
  return END;
}

/* lwrTrackXEnd  -  codel END of TrackX
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrTrackXEnd(LWR_TRACK_STR *trackStr, int *report)
{
  /* ... add your code here ... */
  return ETHER;
}

/* lwrTrackXInter  -  codel INTER of TrackX
   Returns:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrTrackXInter(LWR_TRACK_STR *trackStr, int *report)
{
  /* ... add your code here ... */
  return ETHER;
}

/*------------------------------------------------------------------------
 * SetCollisionEnvPoster
 *
 * Description: 
 *
 * Reports:      OK
 *              S_lwr_CANNOT_FIND_POSTER
 *              S_lwr_CANNOT_READ_POSTER
 */

/* lwrSetCollisionEnvPosterMain  -  codel EXEC of SetCollisionEnvPoster
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetCollisionEnvPosterMain(GEN_STRING128 *collisionEnvPosterName, int *report)
{

  /* Find the external poster and try it by reading once */
  if(lwrGENMANIP_COLLISION_STATE_STRPosterFind(collisionEnvPosterName->name,
					       &posterMhpCollideId) == ERROR) {
    posterMhpCollideId = NULL;
    *report =  S_lwr_CANNOT_FIND_POSTER;
  }
  else if(lwrGENMANIP_COLLISION_STATE_STRPosterRead(posterMhpCollideId,
						    &SDI_F->collisionEnvState) == ERROR) {
    posterMhpCollideId = NULL;
    *report =  S_lwr_CANNOT_READ_POSTER;
  }
  else {
    *report = OK;
  }
 
  return ETHER;
}




/*------------------------------------------------------------------------
 * GotoContact
 *
 * Description: 
 *
 * Reports:      OK
 *              S_lwr_NOT_INITIALIZED
 */

/* lwrGotoContactStart  -  codel START of GotoContact
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoContactStart(double *contactForce, int *report)
{

  return ETHER;
}

/* lwrGotoContactMain  -  codel EXEC of GotoContact
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoContactMain(double *contactForce, int *report)
{


  return ETHER;
}

/* lwrGotoContactEnd  -  codel END of GotoContact
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoContactEnd(double *contactForce, int *report)
{

  return ETHER;
}

/* lwrGotoContactInter  -  codel INTER of GotoContact
   Returns:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoContactInter(double *contactForce, int *report)
{

  return ETHER;
}




/*------------------------------------------------------------------------
 * SetScheme
 *
 * Description: 
 *
 * Reports:      OK
 */

/* lwrSetSchemeMain  -  codel EXEC of SetScheme
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetSchemeMain(LWR_FRI_SCHEME *fri_scheme, int *report)
{
  /* ... add your code here ... */
  return ETHER;
}

/* lwrSetSchemeEnd  -  codel END of SetScheme
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetSchemeEnd(LWR_FRI_SCHEME *fri_scheme, int *report)
{
  /* ... add your code here ... */
  return ETHER;
}





/*------------------------------------------------------------------------
 * GotoHome
 *
 * Description: 
 *
 * Reports:      OK
 */

/* lwrGotoHomeMain  -  codel EXEC of GotoHome
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoHomeMain(int *report)
{
  // Move arm to a predefined home position
  //  friInst->setToKRLInt(CMD_CHANNEL, START_POSE);
  return ETHER;
}

/* lwrGotoHomeEnd  -  codel END of GotoHome
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrGotoHomeEnd(int *report)
{
  /* ... add your code here ... */
  return ETHER;
}








/*------------------------------------------------------------------------
 * SetCntrlMode
 *
 * Description: 
 *
 * Reports:      OK
 */

/* lwrSetCntrlModeStart  -  codel START of SetCntrlMode
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetCntrlModeStart(LWR_CNTRL_MODE_ENUM *cntrl_mode, int *report)
{
  if(SDI_F->status.simulation == GEN_TRUE) {
    return EXEC;
  }
  ACTIVITY_EVENT eventLeft, eventRight;  
#ifdef LWR_FRI

  SDI_F->armLeft.cntrl_mode = *cntrl_mode;
  SDI_F->armRight.cntrl_mode = *cntrl_mode;

  switch(*cntrl_mode) {
  case LWR_OPEN_LOOP_CNTRL:
  case LWR_TRAJECTORY_CNTRL:
  case LWR_GOAL_CNTRL:
  case LWR_EXTERNAL_POSITION_CNTRL:
  case LWR_EXTERNAL_VELOCITY_CNTRL:
  case LWR_FORCE_CNTRL:

    if(SDI_F->armLeft.is_used == GEN_TRUE) {
      eventLeft = requestFriStateCmd(report, LWR_ARM_LEFT, LWR_CTRL_POSITION);
    } else {
      eventLeft = EXEC;
    }
    if(SDI_F->armRight.is_used == GEN_TRUE) {
      eventRight = requestFriStateCmd(report, LWR_ARM_RIGHT, LWR_CTRL_POSITION);
    } else {
      eventRight = EXEC;
    }
    
    if(eventLeft == EXEC && eventRight == EXEC) {
      printf("lwrGotoQRelStart OK\n");
      return EXEC;
    }
    if(eventLeft == ETHER || eventRight == ETHER) { 
      return ETHER;
    }
    if(eventLeft == START || eventRight == START) {
      if(eventLeft == START ) {
	printf("lwrGotoQRelStart waiting for arm Left\n");
      }
      if(eventRight == START ) {
	printf("lwrGotoQRelStart waiting for arm Right\n");
      }
      return START;
    }
    break;
  default:

    break;
  }
#else
  return EXEC;
#endif
}

/* lwrSetCntrlModeMain  -  codel EXEC of SetCntrlMode
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetCntrlModeMain(LWR_CNTRL_MODE_ENUM *cntrl_mode, int *report)
{
  currentMotion.curTime = 0.0;
  setMotionStart(LWR_POSITION);
  return ETHER;
}

/* lwrSetCntrlModeEnd  -  codel END of SetCntrlMode
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetCntrlModeEnd(LWR_CNTRL_MODE_ENUM *cntrl_mode, int *report)
{
  /* ... add your code here ... */
  return ETHER;
}

/* lwrSetCntrlModeInter  -  codel INTER of SetCntrlMode
   Returns:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetCntrlModeInter(LWR_CNTRL_MODE_ENUM *cntrl_mode, int *report)
{
  /* ... add your code here ... */
  return ETHER;
}



/*------------------------------------------------------------------------
 * SetSAHandPoster
 *
 * Description: 
 *
 * Reports:      OK
 *              S_lwr_CANNOT_FIND_POSTER
 *              S_lwr_CANNOT_READ_POSTER
 */

/* lwrSetSAHandPosterMain  -  codel EXEC of SetSAHandPoster
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetSAHandPosterMain(GEN_STRING128 *sahandPosterName, int *report)
{
  /* Find the external poster and try it by reading once */
  if(lwrGENMANIP_SAHAND_STRPosterFind(sahandPosterName->name,
				      &posterSAHandId) == ERROR) {
    *report =  S_lwr_CANNOT_FIND_POSTER;
  }
  else if(lwrGENMANIP_SAHAND_STRPosterRead(posterSAHandId,
					   &SDI_F->sahandPos) == ERROR) {
    *report =  S_lwr_CANNOT_READ_POSTER;
  }
  else {
    SDI_F->status.sahandUpdate = GEN_TRUE;  
    *report = OK;
  }
  return ETHER;
}

/*------------------------------------------------------------------------
 * SetHumanDistancePoster
 *
 * Description: 
 *
 * Reports:      OK
 *              S_lwr_CANNOT_FIND_POSTER
 *              S_lwr_CANNOT_READ_POSTER
 */

/* lwrSetHumanDistancePosterMain  -  codel EXEC of SetHumanDistancePoster
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetHumanDistancePosterMain(GEN_STRING128 *humanDistPosterName, int *report)
{
  /* Find the external poster and try it by reading once */
 if(lwrGENHUM_HUMAN_DISTANCEPosterFind(humanDistPosterName->name,
				      &posterHumDistId) == ERROR) {
    *report =  S_lwr_CANNOT_FIND_POSTER;
  }
  else if(lwrGENHUM_HUMAN_DISTANCEPosterRead(posterHumDistId,
					   &SDI_F->humanDist) == ERROR) {
    *report =  S_lwr_CANNOT_READ_POSTER;
  }
  else {
    *report = OK;
  }

  return ETHER;
}




/*------------------------------------------------------------------------
 * SaveForces
 *
 * Description: 
 *
 * Reports:      OK
 *              S_lwr_FILE_NAME_TOO_LONG
 *              S_lwr_FILE_IO_ERROR
 */

/* lwrSaveForcesMain  -  codel EXEC of SaveForces
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSaveForcesMain(GEN_STRING128 *ForcesFileName, int *report)
{
  FILE* filePtr = 0;

  #ifdef LWR_FRI
  if(friInstLeft) {
    friInstLeft->setToKRLInt(CMD_CHANNEL, FRI_IDLE);
  }
  if(friInstRight) {
    friInstRight->setToKRLInt(CMD_CHANNEL, FRI_IDLE);
  }
  #endif
  if ((filePtr = fopen(ForcesFileName->name,"wb")) == NULL) {
    printf("ERROR Open File %s\n", ForcesFileName->name);
    *report = S_lwr_FILE_IO_ERROR;
    return ETHER;
  } else {
    fprintf(filePtr,"# i L_Fx L_Fy L_Fz L_Mx L_My L_Mz R_Fx R_Fy R_Fz R_Mx R_My R_Mz ;\n");
    for(int i=0; i < LWR_FORCES_WINDOW; i++) {
      fprintf(filePtr,"%d %g %g %g %g %g %g %g %g %g %g %g %g ;\n", i, 
	      SDI_F->armLeft.estExtTcpFTHistory[i][0],
	      SDI_F->armLeft.estExtTcpFTHistory[i][1],
	      SDI_F->armLeft.estExtTcpFTHistory[i][2],
	      SDI_F->armLeft.estExtTcpFTHistory[i][3],
	      SDI_F->armLeft.estExtTcpFTHistory[i][4],
	      SDI_F->armLeft.estExtTcpFTHistory[i][5],
	      SDI_F->armRight.estExtTcpFTHistory[i][0],
	      SDI_F->armRight.estExtTcpFTHistory[i][1],
	      SDI_F->armRight.estExtTcpFTHistory[i][2],
	      SDI_F->armRight.estExtTcpFTHistory[i][3],
	      SDI_F->armRight.estExtTcpFTHistory[i][4],
	      SDI_F->armRight.estExtTcpFTHistory[i][5]);
    }
    fclose(filePtr);
  }
  return ETHER;
}




/*------------------------------------------------------------------------
 * SetCollisionEnvMode
 *
 * Description: 
 *
 * Reports:   S_lwr_POSTER_NOT_SETTED    
 *            OK
 */

/* lwrSetCollisionEnvModeMain  -  codel EXEC of SetCollisionEnvMode
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
lwrSetCollisionEnvModeMain(GEN_BOOL *collisionEnvMode, int *report)
{
  if(posterMhpCollideId == NULL) {
    SDI_F->status.collisionEnvMode = GEN_FALSE;
    printf("WARN: lwrSetCollisionEnvModeMain the collisionEnv poster is not setted\n");
    *report =  S_lwr_POSTER_NOT_SETTED;
  } else {
    SDI_F->status.collisionEnvMode = GEN_TRUE;
  }
  return ETHER;
}



