/*
 * Copyright (c) 2010 LAAS/CNRS                       --  Oct 2010
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
/*------------------------------------------------------------
 * 
 *                     --  Module  LWR  --
 * 
 *  Description:
 *  Creation date : Wed May 19 22:35:29 2010 
 *  Author: Xavier Broquere
 *  LAAS-CNRS, Toulouse FRANCE
 *------------------------------------------------------------*/

#include <math.h>
#include <stdio.h>

#ifdef LWR_FRI
#include "kukaFri-libs/friremote.h"
#endif

#include "server/lwrHeader.h"

extern "C" {
  #include "gbStruct.h"
  #include "Proto_gb.h"
  #include "Proto_gbModeles.h"  
}


#include <sys/time.h>
#include "libeuler/pomEuler.h"

#include "lwrStruct.h"
#include "lwrFunctions.h"

extern LWR_MOTION_STR currentMotion;

double deg2rad(double deg)  { return (deg*M_PI)/180.0; }
double rad2deg(double rad)  { return (rad/M_PI)*180.0; }



void fri2th_cartPos(float cartPos[LWR_CART_FRM_DIM], Gb_th* th)
{
  th->vx.x = cartPos[ 0];
  th->vy.x = cartPos[ 1];
  th->vz.x = cartPos[ 2];
  th->vp.x = cartPos[ 3];

  th->vx.y = cartPos[ 4];
  th->vy.y = cartPos[ 5];
  th->vz.y = cartPos[ 6];
  th->vp.y = cartPos[ 7];

  th->vx.z = cartPos[ 8];
  th->vy.z = cartPos[ 9];
  th->vz.z = cartPos[10];
  th->vp.z = cartPos[11];
}


Gb_th fri2dep_cartPos(float cartPos[LWR_CART_FRM_DIM], Gb_dep* dep)
{
  Gb_th th;

  fri2th_cartPos(cartPos, &th);

  normalize( dep->rx,  dep->ry,  dep->rz,
            &dep->rx, &dep->ry, &dep->rz);

  Gb_th_dep(&th, dep);

  return th;
}

void th2fri_cartPos(Gb_th* th, float cartPos[LWR_CART_FRM_DIM])
{
  cartPos[ 0] = th->vx.x;
  cartPos[ 1] = th->vy.x;
  cartPos[ 2] = th->vz.x;
  cartPos[ 3] = th->vp.x;

  cartPos[ 4] = th->vx.y;
  cartPos[ 5] = th->vy.y;
  cartPos[ 6] = th->vz.y;
  cartPos[ 7] = th->vp.y;

  cartPos[ 8] = th->vx.z;
  cartPos[ 9] = th->vy.z;
  cartPos[10] = th->vz.z;
  cartPos[11] = th->vp.z;
}

Gb_th dep2fri_cartPos(Gb_dep* dep, float cartPos[LWR_CART_FRM_DIM])
{
  Gb_th th;
  Gb_dep_th(dep, &th);

  th2fri_cartPos(&th, cartPos);

  return th;
}


/*
 * Setup transformation matrix from reference system 7 to R tool. The second
 * parameter alows projecting the origin of R tool outside the physical tool
 * providing pivoting movements around a target object.
 */
#ifdef LWR_FRI
void setupTh7Tool(friRemote * friInst, LWR_TOOL_ENUM tool, double distance,  Gb_th *th_7_tool,  Gb_th *th_8_tool)
#else
void setupTh7Tool( LWR_TOOL_ENUM tool, double distance,  Gb_th *th_7_tool,  Gb_th *th_8_tool)
#endif
{
  Gb_th th_7_8;

  th_7_8.vx.x =  1;
  th_7_8.vy.x =  0;
  th_7_8.vz.x =  0;
  th_7_8.vp.x =  0;

  th_7_8.vx.y =  0;
  th_7_8.vy.y =  1;
  th_7_8.vz.y =  0;
  th_7_8.vp.y =  0;

  th_7_8.vx.z =  0;
  th_7_8.vy.z =  0;
  th_7_8.vz.z =  1;
  th_7_8.vp.z = LWR_R7;

  switch (tool) {
    case LWR_NO_TOOL:
      th_8_tool->vx.x =  1;
      th_8_tool->vy.x =  0;
      th_8_tool->vz.x =  0;
      th_8_tool->vp.x =  0;

      th_8_tool->vx.y =  0;
      th_8_tool->vy.y =  1;
      th_8_tool->vz.y =  0;
      th_8_tool->vp.y =  0;

      th_8_tool->vx.z =  0;
      th_8_tool->vy.z =  0;
      th_8_tool->vz.z =  1;
      th_8_tool->vp.z = (distance < 0.0)?LWR_R8_NO_TOOL:distance;

      printf("INFO: LWR configured with no tool\n");

      break;
    case LWR_GRIPPER:
      th_8_tool->vx.x =  1;
      th_8_tool->vy.x =  0;
      th_8_tool->vz.x =  0;
      th_8_tool->vp.x =  0;

      th_8_tool->vx.y =  0;
      th_8_tool->vy.y =  1;
      th_8_tool->vz.y =  0;
      th_8_tool->vp.y =  0;

      th_8_tool->vx.z =  0;
      th_8_tool->vy.z =  0;
      th_8_tool->vz.z =  1;
      th_8_tool->vp.z = (distance < 0.0)?LWR_R8_GRIPPER:distance;

      printf("INFO: LWR configured to use the gripper\n");
      
      break;
    case LWR_DLR_HIT:
      th_8_tool->vx.x =  1;
      th_8_tool->vy.x =  0;
      th_8_tool->vz.x =  0;
      th_8_tool->vp.x =  0;

      th_8_tool->vx.y =  0;
      th_8_tool->vy.y =  1;
      th_8_tool->vz.y =  0;
      th_8_tool->vp.y =  0;

      th_8_tool->vx.z =  0;
      th_8_tool->vy.z =  0;
      th_8_tool->vz.z =  1;
      th_8_tool->vp.z = (distance < 0.0)?LWR_R8_DLR_HIT:distance;

      printf("INFO: LWR configured to use the DLR hand\n");
      
      break;
    default:
      printf("ERROR: Unrecogniced LWR tool: %d. No tool will be used instead\n", tool);
#ifdef LWR_FRI
      setupTh7Tool(friInst, LWR_NO_TOOL);
#else
      setupTh7Tool( LWR_NO_TOOL);
#endif
      return;
  }

  Gb_th_produit(&th_7_8, th_8_tool, th_7_tool);
#ifdef LWR_FRI
  friInst->setToKRLInt(CMD_CHANNEL, SET_TOOL);
  friInst->setToKRLInt(ARG_CHANNEL, tool);
#endif
  return;
}


/*
 * Setup transformation matrix from reference system 7 to R tool. The second
 * parameter alows projecting the origin of R tool outside the physical tool
 * providing pivoting movements around a target object.
 */
#ifdef LWR_FRI
void setupThBase0(friRemote * friInst, LWR_BASE_ENUM base, Gb_th *th_base_0)
#else
void setupThBase0( LWR_BASE_ENUM base, Gb_th *th_base_0)
#endif
{
  switch (base) {
    case LWR_NULL_BASE:
      th_base_0->vx.x =  1;
      th_base_0->vy.x =  0;
      th_base_0->vz.x =  0;
      th_base_0->vp.x =  0;

      th_base_0->vx.y =  0;
      th_base_0->vy.y =  1;
      th_base_0->vz.y =  0;
      th_base_0->vp.y =  0;

      th_base_0->vx.z =  0;
      th_base_0->vy.z =  0;
      th_base_0->vz.z =  1;
      th_base_0->vp.z =  0;

      printf("INFO: LWR configured to operate mounted on a table\n");
      
      break;
    case LWR_JIDO_BASE:
      th_base_0->vx.x =  0;
      th_base_0->vy.x =  1;
      th_base_0->vz.x =  0;
      th_base_0->vp.x =  0.165;

      th_base_0->vx.y =  0;
      th_base_0->vy.y =  0;
      th_base_0->vz.y =  -1;
      th_base_0->vp.y =  -.137;//-0.165;

      th_base_0->vx.z = -1;
      th_base_0->vy.z =  0;
      th_base_0->vz.z =  0;
      th_base_0->vp.z =  0.907;

      printf("INFO: LWR configured to operate mounted on JIDO robot\n");
      
      break;
    default:
      printf("ERROR: Unrecogniced LWR base: %d. Null base will be used instead\n", base);
#ifdef LWR_FRI
      setupThBase0(friInst, LWR_NULL_BASE);
#else
     setupThBase0(LWR_NULL_BASE);
#endif
      return;
  }
#ifdef LWR_FRI
  friInst->setToKRLInt(CMD_CHANNEL, SET_BASE);
  friInst->setToKRLInt(ARG_CHANNEL, base);
#endif 
  return;
}

/*
 * Setup velocity, acceleration and jerk limits for both per-joint and
 * cartesian movements. fullSpeedPercent Varies from 0 (no movement) and
 * 1 (movement at the maximum allowed velocity, acceleration and jerk)
 */
void setupMaxVAJ(double fullSpeedPercent,
                 double* q_V_max, double* q_A_max, double* q_J_max,
                 double* x_V_max, double* x_A_max, double* x_J_max)
{
  /* Per joint maximum v, a, j */
  q_V_max[0] = LWR_A1_MAX_VELOCITY*fullSpeedPercent;
  q_V_max[1] = LWR_A2_MAX_VELOCITY*fullSpeedPercent;
  q_V_max[2] = LWR_E1_MAX_VELOCITY*fullSpeedPercent;
  q_V_max[3] = LWR_A3_MAX_VELOCITY*fullSpeedPercent;
  q_V_max[4] = LWR_A4_MAX_VELOCITY*fullSpeedPercent;
  q_V_max[5] = LWR_A5_MAX_VELOCITY*fullSpeedPercent;
  q_V_max[6] = LWR_A6_MAX_VELOCITY*fullSpeedPercent;

  q_V_max[7] = LWR_A1_MAX_VELOCITY*fullSpeedPercent;
  q_V_max[8] = LWR_A2_MAX_VELOCITY*fullSpeedPercent;
  q_V_max[9] = LWR_E1_MAX_VELOCITY*fullSpeedPercent;
  q_V_max[10] = LWR_A3_MAX_VELOCITY*fullSpeedPercent;
  q_V_max[11] = LWR_A4_MAX_VELOCITY*fullSpeedPercent;
  q_V_max[12] = LWR_A5_MAX_VELOCITY*fullSpeedPercent;
  q_V_max[13] = LWR_A6_MAX_VELOCITY*fullSpeedPercent;

  for (int i = 0; i < 2*LWR_JOINTS; i++) {
    q_A_max[i] = q_V_max[i]*2.0;
    q_J_max[i] = q_A_max[i]*3.0;
  }

  /* Cartesian maximum v, a, j */
  x_V_max[0] = LWR_MAX_LIN_VELOCITY*fullSpeedPercent;  // position x
  x_V_max[1] = LWR_MAX_LIN_VELOCITY*fullSpeedPercent;  // position y
  x_V_max[2] = LWR_MAX_LIN_VELOCITY*fullSpeedPercent;  // position z
  x_V_max[3] = LWR_MAX_ANG_VELOCITY*fullSpeedPercent;  // rotation x
  x_V_max[4] = LWR_MAX_ANG_VELOCITY*fullSpeedPercent;  // rotation y
  x_V_max[5] = LWR_MAX_ANG_VELOCITY*fullSpeedPercent;  // rotation z

  x_V_max[6] = LWR_MAX_LIN_VELOCITY*fullSpeedPercent;  // position x
  x_V_max[7] = LWR_MAX_LIN_VELOCITY*fullSpeedPercent;  // position y
  x_V_max[8] = LWR_MAX_LIN_VELOCITY*fullSpeedPercent;  // position z
  x_V_max[9] = LWR_MAX_ANG_VELOCITY*fullSpeedPercent;  // rotation x
  x_V_max[10] = LWR_MAX_ANG_VELOCITY*fullSpeedPercent;  // rotation y
  x_V_max[11] = LWR_MAX_ANG_VELOCITY*fullSpeedPercent;  // rotation z

  for (int i = 0; i < 2*LWR_SM_CART_COORDS; i++) {
    x_A_max[i] = x_V_max[i]*2.0;
    x_J_max[i] = x_A_max[i]*3.0;
  }
  return;
}


/*
 * Compute for a given movement: sample time, total time, samples number
 * and time evolution.
 */
void computeTimming(double sampleTime, double totalTime, LWR_MOTION_STR* motion) {
  motion->samplesNb = 0;
  do {
    motion->t[motion->samplesNb] = motion->samplesNb*sampleTime;
  }
  while (motion->t[motion->samplesNb++] < totalTime); 

  if (motion->t[motion->samplesNb - 1] > totalTime)
    motion->t[motion->samplesNb - 1] = totalTime;

  //#ifdef LWR_DEBUG
  printf("TIMMING : samples = %d, mot. duration = %f, sample time = %f\n", motion->samplesNb, totalTime, sampleTime);
//#endif
}


bool approachingBounds(int joint, double old_q, double new_q)
{
  if ((fabs(LWR_JOINT_RANGE_MIN[joint] - new_q) < LWR_JNT_SECURITY_MARGIN) &&
      (new_q < (old_q - 0.0001)))
    return true;

  if ((fabs(LWR_JOINT_RANGE_MAX[joint] - new_q) < LWR_JNT_SECURITY_MARGIN) &&
      (new_q > (old_q + 0.0001)))
    return true;

  return false;
}

double normalize(double   x, double   y, double   z,
                 double* nx, double* ny, double* nz)
{
  Gb_v3 v, nv;
  v.x = x;
  v.y = y;
  v.z = z;
  double w = Gb_v3_norme(&v, &nv);
  *nx = nv.x;
  *ny = nv.y;
  *nz = nv.z;
  return w;
}

void setMotionError(LWR_MOTION_ERROR_ENUM error)
{
  currentMotion.type  = LWR_NOT_MOVE;
  currentMotion.state = LWR_ABORTED;
  currentMotion.error = error;
}

void setMotionStart(LWR_MOTION_TYPE_ENUM type)
{
  currentMotion.type  = type;
  currentMotion.state  = LWR_STARTED;
  currentMotion.error  = LWR_OK;
  currentMotion.sample = 0;
  currentMotion.curTime = 0.0;
}

void setMotionEnd()
{
  currentMotion.type  = LWR_NOT_MOVE;
  currentMotion.state = LWR_FINISHED;
  currentMotion.error = LWR_OK;
}

void fillMotionReport(LWR_MOTION_ERROR_ENUM error, int *report)
{
  switch (error) {
    case LWR_OK:               *report = OK;                      break;
    case LWR_NOT_INITIALIZED:  *report = S_lwr_NOT_INITIALIZED;   break;
    case LWR_WRONG_Q_VALUE:    *report = S_lwr_WRONG_VALUE;       break;
    case LWR_SOFTMOTION_ERROR: *report = S_lwr_SOFTMOTION_ERROR;  break;
    case LWR_FRI_MONITOR_MODE: *report = S_lwr_FRI_MONITOR_MODE;  break;
    case LWR_MGI_CALC_ERROR:   *report = S_lwr_MGI_CALC_ERROR;    break;
    case LWR_MGI_SINGULARITY:  *report = S_lwr_MGI_SINGULARITY;   break;
    case LWR_JNT_OUT_OF_RANGE: *report = S_lwr_JNT_OUT_OF_RANGE;  break;
    case LWR_COLLISION_RISK:   *report = S_lwr_COLLISION_RISK;    break;
  }
}

// daniel: reste problme calcul temps pour acclration
#ifdef LWR_FRI
int lwrUpdateArmStatus(friRemote *friInst, LWR_ARM_INST* arm, POM_SE_POSTER *pomArmPos, char * pomPosterName)
#else
int lwrUpdateArmStatus( LWR_ARM_INST* arm, POM_SE_POSTER *pomArmPos, char * pomPosterName)
#endif
{
  //struct timeval s_date;
  // struct timeval s_date0 = { arm->s_date0_tv_sec, arm->s_date0_tv_usec };
  double date; 
  double time;

  //float currentCartPos[LWR_CART_FRM_DIM];
  float currentJntVals[LWR_JOINTS];
  T3D t3d;
  Gb_dep currDep;
  Gb_th currTh;

#ifdef LWR_FRI
  if( friInst != NULL) {
    memcpy(arm->currentCartPos, friInst->getMsrCartPosition(), sizeof(float)*LWR_CART_FRM_DIM);
    memcpy(arm->currentMsrJntVals, friInst->getMsrMsrJntPosition(), sizeof(float)*LWR_JOINTS);
    //    memcpy(currentCartPos, friInst->getMsrCmdCartPosition(), sizeof(float)*FRI_CART_FRM_DIM);
    friInst->getCurrentCmdJntPosition(currentJntVals);
    //memcpy(currentJntVals, friInst->getMsrMsrJntPosition(), sizeof(float)*LWR_JOINTS);

  } else {
    memcpy(currentJntVals , arm->currentJntVals, sizeof(float)*LWR_JOINTS);
  }
  // Fill posters with current position and configuration
  fri2dep_cartPos(arm->currentCartPos, &arm->currPose);
#else
 memcpy(currentJntVals , arm->currentJntVals, sizeof(float)*LWR_JOINTS);

#endif
  arm->currConf.q1 = currentJntVals[0];
  arm->currConf.q2 = currentJntVals[1];
  arm->currConf.q3 = currentJntVals[2];
  arm->currConf.q4 = currentJntVals[3];
  arm->currConf.q5 = currentJntVals[4];
  arm->currConf.q6 = currentJntVals[5];
  arm->currConf.q7 = currentJntVals[6];

#ifndef   LWR_FRI
  Gb_th th07;
  kukaLBR_mgd(&(arm->currConf), LWR_R3, LWR_R5, &th07);
  //  Gb_th_print(&th07, "thR7");
  th2fri_cartPos(&th07, arm->currentCartPos);
#endif

  arm->pos.q.q1 = currentJntVals[0];
  arm->pos.q.q2 = currentJntVals[1];
  arm->pos.q.q3 = currentJntVals[2];
  arm->pos.q.q4 = currentJntVals[3];
  arm->pos.q.q5 = currentJntVals[4];
  arm->pos.q.q6 = currentJntVals[5];
  arm->pos.q.q7 = currentJntVals[6];

#ifdef LWR_FRI
  if( friInst != NULL) {
    date = friInst->getTimestamp();
  } else {
    date  = arm->timeStamp;
  }
#endif
  arm->pos.date = date;

  if (arm->date1 == -1.) {
    arm->pos.vq.q1 = 0.;
    arm->pos.vq.q2 = 0.;
    arm->pos.vq.q3 = 0.;
    arm->pos.vq.q4 = 0.;
    arm->pos.vq.q5 = 0.;
    arm->pos.vq.q6 = 0.;
    arm->pos.vq.q7 = 0.;

    arm->pos.vqold.q1 = 0.;
    arm->pos.vqold.q2 = 0.;
    arm->pos.vqold.q3 = 0.;
    arm->pos.vqold.q4 = 0.;
    arm->pos.vqold.q5 = 0.;
    arm->pos.vqold.q6 = 0.;
    arm->pos.vqold.q7 = 0.;

  }
  else {

#ifndef LWR_FRI
   time = (double)LWR_PERIOD;
#else
    if(date > 0.0 ) {
      time = (double)date - (double)arm->date1;
    } else {
      time = (double)LWR_PERIOD;
    }
      time = (double)LWR_PERIOD;
#endif
    //  arm->pos.vq.q1 = (arm->pos.vqold.q1 + ((arm->pos.q.q1 - arm->pos.qold.q1) / (time)))/2.;
    //  arm->pos.vq.q2 = (arm->pos.vqold.q2 + ((arm->pos.q.q2 - arm->pos.qold.q2) / (time)))/2.;
    //  arm->pos.vq.q3 = (arm->pos.vqold.q3 + ((arm->pos.q.q3 - arm->pos.qold.q3) / (time)))/2.;
    //  arm->pos.vq.q4 = (arm->pos.vqold.q4 + ((arm->pos.q.q4 - arm->pos.qold.q4) / (time)))/2.;
    //  arm->pos.vq.q5 = (arm->pos.vqold.q5 + ((arm->pos.q.q5 - arm->pos.qold.q5) / (time)))/2.;
    //  arm->pos.vq.q6 = (arm->pos.vqold.q6 + ((arm->pos.q.q6 - arm->pos.qold.q6) / (time)))/2.;
    //  arm->pos.vq.q7 = (arm->pos.vqold.q7 + ((arm->pos.q.q7 - arm->pos.qold.q7) / (time)))/2.;

      arm->pos.vq.q1 =  (arm->pos.q.q1 - arm->pos.qold.q1) / (time);
      arm->pos.vq.q2 =  (arm->pos.q.q2 - arm->pos.qold.q2) / (time);
      arm->pos.vq.q3 =  (arm->pos.q.q3 - arm->pos.qold.q3) / (time);
      arm->pos.vq.q4 =  (arm->pos.q.q4 - arm->pos.qold.q4) / (time);
      arm->pos.vq.q5 =  (arm->pos.q.q5 - arm->pos.qold.q5) / (time);
      arm->pos.vq.q6 =  (arm->pos.q.q6 - arm->pos.qold.q6) / (time);
      arm->pos.vq.q7 =  (arm->pos.q.q7 - arm->pos.qold.q7) / (time);

  }
  if (arm->date2 == -1.) {
    arm->pos.aq.q1 = 0.;
    arm->pos.aq.q2 = 0.;
    arm->pos.aq.q3 = 0.;
    arm->pos.aq.q4 = 0.;
    arm->pos.aq.q5 = 0.;
    arm->pos.aq.q6 = 0.;
    arm->pos.aq.q7 = 0.;

    arm->pos.aqold.q1 = 0.;
    arm->pos.aqold.q2 = 0.;
    arm->pos.aqold.q3 = 0.;
    arm->pos.aqold.q4 = 0.;
    arm->pos.aqold.q5 = 0.;
    arm->pos.aqold.q6 = 0.;
    arm->pos.aqold.q7 = 0.;
  }
  else {
#ifndef LWR_FRI
    time = (double)LWR_PERIOD;
#else
    if(date > 0.0 ) {
      //time =  (date - arm->date2) * 2.0 ;
      time =  ((double)date - (double)arm->date2);
    } else {
      time = (double)LWR_PERIOD;
    }
    time = (double)LWR_PERIOD;
#endif
  
//    arm->pos.aq.q1 = (arm->pos.aqold.q1 + ( (arm->pos.vq.q1 - arm->pos.vqold.q1) / time)) /2.;
//    arm->pos.aq.q2 = (arm->pos.aqold.q2 + ( (arm->pos.vq.q2 - arm->pos.vqold.q2) / time)) /2.;
//    arm->pos.aq.q3 = (arm->pos.aqold.q3 + ( (arm->pos.vq.q3 - arm->pos.vqold.q3) / time)) /2.;
//    arm->pos.aq.q4 = (arm->pos.aqold.q4 + ( (arm->pos.vq.q4 - arm->pos.vqold.q4) / time)) /2.;
//    arm->pos.aq.q5 = (arm->pos.aqold.q5 + ( (arm->pos.vq.q5 - arm->pos.vqold.q5) / time)) /2.;
//    arm->pos.aq.q6 = (arm->pos.aqold.q6 + ( (arm->pos.vq.q6 - arm->pos.vqold.q6) / time)) /2.;
//    arm->pos.aq.q7 = (arm->pos.aqold.q7 + ( (arm->pos.vq.q7 - arm->pos.vqold.q7) / time)) /2.;

    arm->pos.aq.q1 = (arm->pos.vq.q1 - arm->pos.vqold.q1) / time;
    arm->pos.aq.q2 = (arm->pos.vq.q2 - arm->pos.vqold.q2) / time;
    arm->pos.aq.q3 = (arm->pos.vq.q3 - arm->pos.vqold.q3) / time;
    arm->pos.aq.q4 = (arm->pos.vq.q4 - arm->pos.vqold.q4) / time;
    arm->pos.aq.q5 = (arm->pos.vq.q5 - arm->pos.vqold.q5) / time;
    arm->pos.aq.q6 = (arm->pos.vq.q6 - arm->pos.vqold.q6) / time;
    arm->pos.aq.q7 = (arm->pos.vq.q7 - arm->pos.vqold.q7) / time;

  }



 // arm->pos.vq.q1 =arm->currentCmdVelJntVals[0] ;
 // arm->pos.vq.q2 =arm->currentCmdVelJntVals[1] ;
 // arm->pos.vq.q3 =arm->currentCmdVelJntVals[2] ;
 // arm->pos.vq.q4 =arm->currentCmdVelJntVals[3] ;
 // arm->pos.vq.q5 =arm->currentCmdVelJntVals[4] ;
 // arm->pos.vq.q6 =arm->currentCmdVelJntVals[5] ;
 // arm->pos.vq.q7 =arm->currentCmdVelJntVals[6] ;
  
//  arm->pos.aq.q1 =arm->currentCmdAccJntVals[0] ;
//  arm->pos.aq.q2 =arm->currentCmdAccJntVals[1] ;
//  arm->pos.aq.q3 =arm->currentCmdAccJntVals[2] ;
//  arm->pos.aq.q4 =arm->currentCmdAccJntVals[3] ;
//  arm->pos.aq.q5 =arm->currentCmdAccJntVals[4] ;
//  arm->pos.aq.q6 =arm->currentCmdAccJntVals[5] ;
//  arm->pos.aq.q7 =arm->currentCmdAccJntVals[6] ;


  arm->date2 = arm->date1;
  arm->date1 = date;
  /* Save previous configuration (q and vq) */ 
  arm->pos.qold  = arm->pos.q;
  arm->pos.vqold = arm->pos.vq;
  arm->pos.aqold = arm->pos.aq;

  // POM interface
  strcpy(pomArmPos->seName, pomPosterName);
  fri2th_cartPos(arm->currentCartPos, &currTh);
  t3dSetOk(&t3d);
  t3d.matrix.flags = T3D_ALLOW_CONVERSION;
  t3d.type = T3D_MATRIX;
  t3d.matrix.type = T3D_MATRIX;
  t3d.matrix.matrix[0]  = currTh.vx.x;
  t3d.matrix.matrix[1]  = currTh.vy.x;
  t3d.matrix.matrix[2]  = currTh.vz.x;
  t3d.matrix.matrix[3]  = currTh.vp.x;
  t3d.matrix.matrix[4]  = currTh.vx.y;
  t3d.matrix.matrix[5]  = currTh.vy.y;
  t3d.matrix.matrix[6]  = currTh.vz.y;
  t3d.matrix.matrix[7]  = currTh.vp.y;
  t3d.matrix.matrix[8]  = currTh.vx.z;
  t3d.matrix.matrix[9]  = currTh.vy.z;
  t3d.matrix.matrix[10] = currTh.vz.z;
  t3d.matrix.matrix[11] = currTh.vp.z;
  t3d.matrix.matrix[12] = 0;
  t3d.matrix.matrix[13] = 0;
  t3d.matrix.matrix[14] = 0;
  t3d.matrix.matrix[15] = 1;
  pomEulerVFromT3D(&t3d, &pomArmPos->seConfig);

  return 0;
}

#ifdef LWR_FRI
int lwrCommandArm(friRemote *friInst, GEN_BOOL simulation, GEN_BOOL isTrqCntrlSofMotion, LWR_ARM_ID_ENUM arm_id, LWR_ARM_INST* armInst, GENMANIP_MOTION_STR *curState, LWR_MOTION_STR *motion) 
#else
  int lwrCommandArm(  GEN_BOOL isTrqCntrlSofMotion, LWR_ARM_ID_ENUM arm_id, LWR_ARM_INST* armInst, GENMANIP_MOTION_STR *curState, LWR_MOTION_STR *motion) 
#endif
{
  float commandJntVals[LWR_JOINTS];
  float currentJntVals[LWR_JOINTS];
  float commandCartPos[LWR_CART_FRM_DIM];
  float currentCartPos[LWR_CART_FRM_DIM];
  Gb_dep dep;
  Gb_th th_R8, th_W8, th_Wt;
  int OFFSET_ARM = 0;

  std::vector<SM_COND> cond;

  if ((motion->type == LWR_POSITION) && (motion->state != LWR_ABORTED)) {

    if(arm_id == LWR_ARM_RIGHT) {
      OFFSET_ARM =  LWR_JOINTS;
    }
    /***********************************************/
    /*         Compute The Control Law             */
    /*  output: commandJntVals[]                   */
    /***********************************************/
    switch (armInst->cntrl_mode) {
    case LWR_OPEN_LOOP_TRACK_Q_CNTRL:
      lwrOpenLoopTrackQCntrl(armInst, curState, motion, commandJntVals);
      break;
    case LWR_OPEN_LOOP_CNTRL :
       lwrOpenLoopCntrl(armInst, curState, motion, commandJntVals);
      break;
    case LWR_TRAJECTORY_CNTRL :
      lwrTrajectoryCntrl(OFFSET_ARM, armInst, curState, motion, commandJntVals);
      break;
    case LWR_GOAL_CNTRL :
      lwrGoalCntrl(OFFSET_ARM, curState, motion, commandJntVals);
      break;
    case LWR_EXTERNAL_POSITION_CNTRL :
      break;
    case LWR_EXTERNAL_VELOCITY_CNTRL :
      break;
   case LWR_FORCE_CNTRL :
     lwrJntTrqCntrl(isTrqCntrlSofMotion, OFFSET_ARM, armInst, curState, motion, commandJntVals);
      break;
    default:
      printf("ERROR lwrCommandArm: WRONG CONTROL MODE\n");
      break;
    }
    /***********************************************/

#ifdef LWR_FRI
    if((friInst != NULL) && (simulation != GEN_TRUE)) {
      friInst->doPositionControl(commandJntVals, false);
      memcpy(armInst->currentJntVals, commandJntVals, sizeof(float)*LWR_JOINTS);
    } else {
      memcpy(armInst->currentJntVals, commandJntVals, sizeof(float)*LWR_JOINTS);
    }
#else
    memcpy(armInst->currentJntVals, commandJntVals, sizeof(float)*LWR_JOINTS);
#endif
    motion->state = LWR_MOVING;

  } /*if ((currentMotion.type[arm_id] == LWR_POSITION) && (currentMotion.state != LWR_ABORTED)) */

  else if  ((motion->type == LWR_CARTESIAN) && (motion->state != LWR_ABORTED)){

    //if(arm_id == LWR_ARM_RIGHT) {
    //  OFFSET_ARM =  LWR_CART_VEC;
    //}
    //
    //// Cartesian command: we obtain Q from the MGI
    //dep.x = motion->x[0 + OFFSET_ARM][motion->sample];
    //dep.y = motion->x[1 + OFFSET_ARM][motion->sample];
    //dep.z = motion->x[2 + OFFSET_ARM][motion->sample];
    //dep.a = normalize(motion->x[3 + OFFSET_ARM][motion->sample],
    //    	      motion->x[4 + OFFSET_ARM][motion->sample],
    //    	      motion->x[5 + OFFSET_ARM][motion->sample],
    //    	      &dep.rx, &dep.ry, &dep.rz);
    //
    //Gb_dep_th(&dep, &th_R8);
    //Gb_th_produit(&armInst->th_base_0, &th_R8, &th_W8);
    //Gb_th_produit(&th_W8, &armInst->th_8_tool, &th_Wt);
    //
    //th2fri_cartPos(&th_R8, commandCartPos);
#ifdef LWR_FRI
    //if((friInst != NULL) && (simulation != GEN_TRUE)) {
    // friInst->doCartesianImpedanceControl(commandCartPos, NULL, NULL, //currentCartStiff, currentCartDamp,
    //					 NULL, NULL, false);
    //} else {
     //TODO COMPUTE THE SUIMUALTED JNT VAL
    //}
#else
     //TODO COMPUTE THE SUIMUALTED JNT VAL
#endif
  }/*if  ((currentMotion.type == LWR_CARTESIAN) && (currentMotion.state != LWR_ABORTED)) */
  return 0;
}

int lwrOpenLoopCntrl( LWR_ARM_INST* armInst, GENMANIP_MOTION_STR *curState, LWR_MOTION_STR *motion, float *commandJntVals) {
  std::vector<SM_COND> cond;
  motion->traj.getMotionCond(motion->curTime, cond);
  for (int i = 0; i < LWR_JOINTS; i++) {
    commandJntVals[i] = cond[i+armInst->indexInTraj].x;
  }

  if(armInst->isHandUsed == GEN_TRUE) {
    armInst->hand.thumbRotation  = cond[armInst->indexHandInTraj + 0].x;
    armInst->hand.thumb.base     = cond[armInst->indexHandInTraj + 1].x;
    armInst->hand.thumb.phaProx  = cond[armInst->indexHandInTraj + 2].x;
    armInst->hand.thumb.phaMid   = cond[armInst->indexHandInTraj + 3].x;
    armInst->hand.thumb.phaDist  = cond[armInst->indexHandInTraj + 4].x;
    armInst->hand.fore.base      = cond[armInst->indexHandInTraj + 5].x;
    armInst->hand.fore.phaProx   = cond[armInst->indexHandInTraj + 6].x;
    armInst->hand.fore.phaMid    = cond[armInst->indexHandInTraj + 7].x;
    armInst->hand.fore.phaDist   = cond[armInst->indexHandInTraj + 8].x;
    armInst->hand.middle.base    = cond[armInst->indexHandInTraj + 9].x;
    armInst->hand.middle.phaProx = cond[armInst->indexHandInTraj + 10].x;
    armInst->hand.middle.phaMid  = cond[armInst->indexHandInTraj + 11].x;
    armInst->hand.middle.phaDist = cond[armInst->indexHandInTraj + 12].x;
    armInst->hand.ring.base      = cond[armInst->indexHandInTraj + 13].x;
    armInst->hand.ring.phaProx   = cond[armInst->indexHandInTraj + 14].x;
    armInst->hand.ring.phaMid    = cond[armInst->indexHandInTraj + 15].x;
    armInst->hand.ring.phaDist   = cond[armInst->indexHandInTraj + 16].x;
  }

  return 0;
}


int lwrOpenLoopTrackQCntrl(LWR_ARM_INST* armInst, GENMANIP_MOTION_STR *curState, LWR_MOTION_STR *motion, float *commandJntVals) 
{
  if(armInst->isLeft == GEN_TRUE) {
    commandJntVals[0] = currentMotion.armLeft[motion->sample].q1;
    commandJntVals[1] = currentMotion.armLeft[motion->sample].q2;
    commandJntVals[2] = currentMotion.armLeft[motion->sample].q3;
    commandJntVals[3] = currentMotion.armLeft[motion->sample].q4;
    commandJntVals[4] = currentMotion.armLeft[motion->sample].q5;
    commandJntVals[5] = currentMotion.armLeft[motion->sample].q6;
    commandJntVals[6] = currentMotion.armLeft[motion->sample].q7;
    if(armInst->isHandUsed == GEN_TRUE) {
      armInst->hand.thumbRotation  = motion->handLeft[motion->sample].thumbRotation ;
      armInst->hand.thumb.base     = motion->handLeft[motion->sample].thumb.base    ;
      armInst->hand.thumb.phaProx  = motion->handLeft[motion->sample].thumb.phaProx ;
      armInst->hand.thumb.phaMid   = motion->handLeft[motion->sample].thumb.phaMid  ;
      armInst->hand.thumb.phaDist  = motion->handLeft[motion->sample].thumb.phaDist ;
      armInst->hand.fore.base      = motion->handLeft[motion->sample].fore.base     ;
      armInst->hand.fore.phaProx   = motion->handLeft[motion->sample].fore.phaProx  ;
      armInst->hand.fore.phaMid    = motion->handLeft[motion->sample].fore.phaMid   ;
      armInst->hand.fore.phaDist   = motion->handLeft[motion->sample].fore.phaDist  ;
      armInst->hand.middle.base    = motion->handLeft[motion->sample].middle.base   ;
      armInst->hand.middle.phaProx = motion->handLeft[motion->sample].middle.phaProx;
      armInst->hand.middle.phaMid  = motion->handLeft[motion->sample].middle.phaMid ;
      armInst->hand.middle.phaDist = motion->handLeft[motion->sample].middle.phaDist ;
      armInst->hand.ring.base      = motion->handLeft[motion->sample].ring.base     ;
      armInst->hand.ring.phaProx   = motion->handLeft[motion->sample].ring.phaProx  ;
      armInst->hand.ring.phaMid    = motion->handLeft[motion->sample].ring.phaMid   ;
      armInst->hand.ring.phaDist   = motion->handLeft[motion->sample].ring.phaDist  ; 
    }
  } else {
    commandJntVals[0] = currentMotion.armRight[motion->sample].q1;
    commandJntVals[1] = currentMotion.armRight[motion->sample].q2;
    commandJntVals[2] = currentMotion.armRight[motion->sample].q3;
    commandJntVals[3] = currentMotion.armRight[motion->sample].q4;
    commandJntVals[4] = currentMotion.armRight[motion->sample].q5;
    commandJntVals[5] = currentMotion.armRight[motion->sample].q6;
    commandJntVals[6] = currentMotion.armRight[motion->sample].q7;
    if(armInst->isHandUsed == GEN_TRUE) {
      armInst->hand.thumbRotation  = motion->handRight[motion->sample].thumbRotation ;
      armInst->hand.thumb.base     = motion->handRight[motion->sample].thumb.base    ;
      armInst->hand.thumb.phaProx  = motion->handRight[motion->sample].thumb.phaProx ;
      armInst->hand.thumb.phaMid   = motion->handRight[motion->sample].thumb.phaMid  ;
      armInst->hand.thumb.phaDist  = motion->handRight[motion->sample].thumb.phaDist ;
      armInst->hand.fore.base      = motion->handRight[motion->sample].fore.base     ;
      armInst->hand.fore.phaProx   = motion->handRight[motion->sample].fore.phaProx  ;
      armInst->hand.fore.phaMid    = motion->handRight[motion->sample].fore.phaMid   ;
      armInst->hand.fore.phaDist   = motion->handRight[motion->sample].fore.phaDist  ;
      armInst->hand.middle.base    = motion->handRight[motion->sample].middle.base   ;
      armInst->hand.middle.phaProx = motion->handRight[motion->sample].middle.phaProx;
      armInst->hand.middle.phaMid  = motion->handRight[motion->sample].middle.phaMid ;
      armInst->hand.middle.phaDist = motion->handRight[motion->sample].middle.phaDist ;
      armInst->hand.ring.base      = motion->handRight[motion->sample].ring.base     ;
      armInst->hand.ring.phaProx   = motion->handRight[motion->sample].ring.phaProx  ;
      armInst->hand.ring.phaMid    = motion->handRight[motion->sample].ring.phaMid   ;
      armInst->hand.ring.phaDist   = motion->handRight[motion->sample].ring.phaDist  ;
    }
  } 
  return 0;
}



int lwrTrajectoryCntrl(int offsetArm,  LWR_ARM_INST* armInst, GENMANIP_MOTION_STR *curState, LWR_MOTION_STR *motion, float *commandJntVals){
  double commandJntDes[LWR_JOINTS];

  double timeSwitchMode;

  std::vector<SM_COND> condDes;
  std::vector<SM_COND> condCur;


  double criticalLength;
  SM_MOTION_MONO motionMono[LWR_JOINTS];
  SM_LIMITS limitsGoto;
  int TrajectoryType[LWR_JOINTS];
  SM_TIMES T_Jerk[LWR_JOINTS];
  double maxTime = 0.0;
  double sum = 0.0;
  double maxAxis = 0;

  double jerkNeed[LWR_JOINTS]; 


  double ratioJerkAllowed = 1.5;

  lwrGetCondFromGENMANIP_MOTION_STR(curState, condCur);
  motion->traj.getMotionCond(motion->curTime, condDes);

  maxTime = 0.0;
  sum = 0.0;
  maxAxis = 0;


  /* check if condDes is reachable within one tic according the limits */
  for(unsigned int i=0; i<LWR_JOINTS; i++) {
    jerkNeed[i] = (condDes[i+ offsetArm].a -  condCur[i+ offsetArm].a) / ((double)LWR_PERIOD*2.0);
    if( ABS(jerkNeed[i]) > ratioJerkAllowed*ABS( motion->q_J_max[i+ offsetArm]) ) {
      //std::cout << " WARNING JERK EXCEED (" << i << "){jmax= " <<  motion->q_J_max[i+ offsetArm] << " , jneed= " << jerkNeed[i] << "}"  << " condDes.a: " <<  condDes[i+ offsetArm].a << " condCur.a: " << condCur[i+ offsetArm].a << std::endl;;

    }
  }

  //for(unsigned int i=0 i<LWR_JOINTS; i++) {
  //  limitsGoto.maxJerk = motion->q_J_max[i+ offsetArm];
  //  limitsGoto.maxAcc  = motion->q_A_max[i+ offsetArm];
  //  limitsGoto.maxVel  = motion->q_V_max[i+ offsetArm];
  //
  //  SM_STATUS resp = sm_ComputeSoftMotion(condCur[i+ offsetArm], condDes[i+ offsetArm], limitsGoto, &(T_Jerk[i]), &(TrajectoryType[i]));
  //  if (resp != SM_OK) {
  //    printf("ERROR sm_ComputeSoftMotion\n");
  //  }
  //  sm_sum_motionTimes(&(T_Jerk[i]), &sum);
  //  
  //  if(sum > maxTime) {
  //    maxTime = sum;
  //    maxAxis = i;
  //  }
  //}
  //
  //for(int i=0; i<LWR_JOINTS_USER; i++) {
  //  sm_adjustMotionWith3seg(motion[i].IC, motion[i].FC, maxTime, &(motionMono[i]));
  //}


  



  for (int i = 0; i < LWR_JOINTS; i++) {
    commandJntVals[i] = condDes[i + offsetArm].x;
    armInst->currentCmdVelJntVals[i] = condDes[i + offsetArm].v;
    armInst->currentCmdAccJntVals[i] = condDes[i + offsetArm].a;
  }


  

  return 0;
}

int lwrGoalCntrl(int offsetArm, GENMANIP_MOTION_STR *curState, LWR_MOTION_STR *motion, float *commandJntVals){


  return 0;
}

int lwrJntTrqCntrlPosition(double force, double factor, double maxForce,SM_COND cond, float* jntVal) {
  double step = 0.0;
  step = ABS(force)*0.000375*factor  - 0.001;
  if(ABS(force)> maxForce) {
    step =  maxForce*0.000375*factor  - 0.001;
  }
  if(force >= 0.0) {
    *jntVal =  cond.x - step;
  } else {
    *jntVal =  cond.x + step;
  }  
  return 0;
}

int lwrJntTrqCntrlSoftMotion(SM_LIMITS limitsGoto, double force, double maxForce, SM_COND cond, float* jntVal, float *jntVel, float *jntAcc) {
  double velocity = 0.0, dc = 0.0;
  SM_COND IC, condDes;
  int dir =0;
  SM_TIMES T_Jerk;
  SM_STATUS resp;

  std::vector<double> I(3);
  std::vector<double> T(SM_NB_SEG);
  std::vector<double> J(SM_NB_SEG);
  std::vector<double> t(1);
  std::vector<double> a(1);
  std::vector<double> v(1);
  std::vector<double> x(1);

  if(ABS(force) > ABS(maxForce)) {
    if(force >= 0.0) {
      velocity = -limitsGoto.maxVel;
    } else {
      velocity = limitsGoto.maxVel;
    }
  } else {
    if(force >= 0.0) {
      velocity = - ABS(force * limitsGoto.maxVel / ABS(maxForce));
    } else {
      velocity = ABS(force * limitsGoto.maxVel / ABS(maxForce));
    }
  }
  condDes.a = 0.0;
  condDes.v = velocity;
  condDes.x = 0.0;
  
  IC.a = cond.a;
  IC.v = cond.v;
  IC.x = 0.0;

  /* compute the critical length */
  sm_CalculOfCriticalLength(IC, condDes, limitsGoto, &dc);
  if (isnan(dc)) {
    printf("isnan dc\n");
    printf("IC.a = %f    IC.v = %f    IC.x = %f\n",cond.a,cond.v,cond.x);
    printf("FC.a = %f    FC.v = %f    FC.x = %f\n",condDes.a, condDes.v,condDes.x);
    *jntVal = cond.x;
    *jntVel = 0.0;
    *jntAcc = 0.0;
    return 0;
  }
  condDes.x = dc;

  /* compute the motion */
  resp = sm_ComputeSoftMotion(IC, condDes, limitsGoto, &T_Jerk, &dir);
  if (resp != SM_OK) {
    printf("ERROR sm_ComputeSoftMotion\n");
  }

  /* get the  position at the next tick */       
  I[0] = cond.a;
  I[1] = cond.v;
  I[2] = cond.x;
      
  T[0] = T_Jerk.Tjpa;
  T[1] = T_Jerk.Taca;
  T[2] = T_Jerk.Tjna;
  T[3] = T_Jerk.Tvc;
  T[4] = T_Jerk.Tjnb;
  T[5] = T_Jerk.Tacb;
  T[6] = T_Jerk.Tjpb;
      
  J[0] =   dir*limitsGoto.maxJerk;
  J[1] =   0.0;
  J[2] = - dir*limitsGoto.maxJerk;
  J[3] =   0.0;
  J[4] = - dir*limitsGoto.maxJerk;
  J[5] =   0.0;
  J[6] =   dir*limitsGoto.maxJerk;

  t[0] = (double)LWR_PERIOD;    
  resp = sm_AVX_TimeVar(I, T, J, t, a, v, x);

  *jntVal = x[0];
  *jntVel  = v[0];
  *jntAcc  = a[0];

  return 0;
}


int lwrJntTrqCntrl(GEN_BOOL isTrqCntrlSofMotion, int offsetArm,  LWR_ARM_INST* armInst, GENMANIP_MOTION_STR *curState, LWR_MOTION_STR *motion, float *commandJntVals){

  std::vector<SM_COND> condCur;
  double velocity = 0.0;
  double coef = 2.0;
  double minForce[LWR_JOINTS];
  double maxForce = 20.0;
  SM_LIMITS limitsGoto;
  double force = 0.0;
  double factor[LWR_JOINTS];


  minForce[0] = 1.0;
  minForce[1] = 1.0;
  minForce[2] = 1.0;
  minForce[3] = 1.0;
  minForce[4] = 0.5;
  minForce[5] = 0.5;
  minForce[6] = 0.5;


  factor[0] = coef * 1.0;
  factor[1] = coef * 1.0;
  factor[2] = coef * 1.0;
  factor[3] = coef * 3.0;
  factor[4] = coef * 10.0;
  factor[5] = coef * 20.0;
  factor[6] = coef * 40.0;

  int nbJointStabilized = 0;
  lwrGetCondFromGENMANIP_MOTION_STR(curState, condCur);

  for(unsigned int i=0 ; i<LWR_JOINTS; i++) {    

    force = armInst->estMeanExtJntTrq[i];

    // if(ABS( armInst->desJntTrq[i] - armInst->estExtJntTrq[i] ) < ABS(minForce[i]) &&  ABS(condCur[i+ offsetArm].v) < 0.01) {

    if(ABS( armInst->desJntTrq[i] - force ) < ABS(minForce[i]) ) {

      commandJntVals[i] = condCur[i+ offsetArm].x;
      armInst->currentCmdVelJntVals[i] = 0.0;
      armInst->currentCmdAccJntVals[i] = 0.0;      

      nbJointStabilized ++;
    } else {

      if(isTrqCntrlSofMotion == GEN_FALSE) {

	//step = ABS(force)*0.000375*factor[i]  - 0.001;
	//
	//if(ABS(force)> maxForce) {
	//  step =  maxForce*0.000375*factor[i]  - 0.001;
	//}
	//if(force >= 0.0) {
	//  commandJntVals[i] =  condCur[i+ offsetArm].x - step;
	//} else {
	//  commandJntVals[i] =  condCur[i+ offsetArm].x + step;
	//}
	lwrJntTrqCntrlPosition(force, factor[i], maxForce, condCur[i+ offsetArm], &(commandJntVals[i]));

      } else { 

	limitsGoto.maxJerk = 2.0 * motion->q_J_max[i+ offsetArm];
	limitsGoto.maxAcc  = 2.0 * motion->q_A_max[i+ offsetArm];
	limitsGoto.maxVel  =       motion->q_V_max[i+ offsetArm];

	lwrJntTrqCntrlSoftMotion(limitsGoto, force, maxForce, condCur[i+ offsetArm], &(commandJntVals[i]), &(armInst->currentCmdVelJntVals[i]), &(armInst->currentCmdAccJntVals[i]));
      }
    }
  }

  if(nbJointStabilized == LWR_JOINTS) {
    printf("INFO: Arm is stabilized\n");
    armInst->isStabilized = GEN_TRUE;
  } else {
    armInst->isStabilized = GEN_FALSE;
  }


  return 0;
}


int lwrGetCondFromGENMANIP_MOTION_STR(GENMANIP_MOTION_STR *curState, std::vector<SM_COND> &cond) {
  SM_COND cond_i;
  cond.clear();
  for(int i=0; i<curState->size; i++){
    cond_i.a = curState->aq[i];
    cond_i.v = curState->vq[i];    
    cond_i.x = curState->q[i];
    cond.push_back(cond_i);
  }
  return 0;
}


int lwrComputeSmoothedTimeScale(double timeScale, SM_LIMITS limitsGoto, SM_COND* timeScaleCond) {
  
  double dc = 0.0;
  SM_COND FC;
  SM_STATUS resp;

  SM_TIMES T_Jerk;
  int dir = 0;
  double timeTrajDuration = 0.0;
  std::vector<double> I(3);
  std::vector<double> T(SM_NB_SEG);
  std::vector<double> J(SM_NB_SEG);
  std::vector<double> t(1);
  std::vector<double> a(1);
  std::vector<double> v(1);
  std::vector<double> x(1);

  FC.a = 0.0;
  FC.v = timeScale;
  FC.x = 1.0;

  timeScaleCond->x = 0.0;

  if(ABS(FC.v - timeScaleCond->v) > 0.001) {
    sm_CalculOfCriticalLength(*timeScaleCond, FC, limitsGoto, &dc);
    if (isnan(dc)) {
      printf("isnan dc\n");
      printf("IC.a = %f    IC.v = %f    IC.x = %f\n", timeScaleCond->a, timeScaleCond->v, timeScaleCond->x);
      printf("FC.a = %f    FC.v = %f    FC.x = %f\n",FC.a, FC.v, FC.x);
    }
	   
    FC.x = dc;
	    	    
    /* compute the motion */
    resp = sm_ComputeSoftMotion(*timeScaleCond, FC, limitsGoto, &T_Jerk, &dir);
    if (resp != SM_OK) {
      printf("ERROR sm_ComputeSoftMotion\n");
    }

    /* get the  position at the next tick */       
    I[0] = timeScaleCond->a;
    I[1] = timeScaleCond->v;
    I[2] = timeScaleCond->x;
      
    T[0] = T_Jerk.Tjpa;
    T[1] = T_Jerk.Taca;
    T[2] = T_Jerk.Tjna;
    T[3] = T_Jerk.Tvc;
    T[4] = T_Jerk.Tjnb;
    T[5] = T_Jerk.Tacb;
    T[6] = T_Jerk.Tjpb;
      
    J[0] =   dir*limitsGoto.maxJerk;
    J[1] =   0.0;
    J[2] = - dir*limitsGoto.maxJerk;
    J[3] =   0.0;
    J[4] = - dir*limitsGoto.maxJerk;
    J[5] =   0.0;
    J[6] =   dir*limitsGoto.maxJerk;

    timeTrajDuration = 0.0;
    for (int p=0; p < SM_NB_SEG; p++) {
      timeTrajDuration += T[p];
    }

    if(timeTrajDuration < LWR_PERIOD) {
      t[0] = (double)timeTrajDuration; 
    } else {
      t[0] = (double)LWR_PERIOD; 
    }
    resp = sm_AVX_TimeVar(I, T, J, t, a, v, x);
    timeScaleCond->x = x[0];
    timeScaleCond->v  = v[0];
    timeScaleCond->a  = a[0];
    //printf("timeScaleCond->v %f \n",timeScaleCond->v); 
  } else {
    timeScaleCond->v = timeScale;
  }
  return 0;
}



/********************************************************/
/*                                                      */
/*               FUNCTIONS SPECIFIC TO FRI              */
/*                                                      */
/********************************************************/
#ifdef LWR_FRI



void switchCtrlScheme(friRemote * friInst, LWR_CTRL_SCHEME_ENUM scheme)
{
  switch (scheme) {
    case LWR_CTRL_POSITION:
      friInst->setToKRLInt(CMD_CHANNEL, POSIT_CTRL);
      break;
    case LWR_CTRL_CART_IMP:
      friInst->setToKRLInt(CMD_CHANNEL, CART_CTRL);
      break;
    case LWR_CTRL_JOINT_IMP:
      friInst->setToKRLInt(CMD_CHANNEL, AXIS_CTRL);
      break;
    case LWR_CTRL_GRAV_COMP:
      friInst->setToKRLInt(CMD_CHANNEL, GRAV_COMP);
      break;
    default:
      printf("ERROR: Unrecogniced control scheme: %d. Ignoring command\n", scheme);
      break;
  }
}

const char * friStateStr(int state)
{
  switch (state) {
    case FRI_STATE_OFF:
      return "inactive";
    case FRI_STATE_MON:
      return "monitor";
    case FRI_STATE_CMD:
      return "command";
    default:
      return "ERROR";
  }
}

const char * friSchemeStr(int scheme)
{
  switch (scheme) {
    case LWR_CTRL_POSITION:
      return "joint position";
    case LWR_CTRL_CART_IMP:
      return "cartesian imp.";
    case LWR_CTRL_JOINT_IMP:
      return "axis impedance";
    case LWR_CTRL_GRAV_COMP:
      return "gravity comp.";
    default:
      return "ERROR";
  }
}

const char * friQualityStr(int quality)
{
  switch (quality) {
    case FRI_QUALITY_UNACCEPTABLE:
      return "unacceptable";
    case FRI_QUALITY_BAD:
      return "bad";
    case FRI_QUALITY_OK:
      return "ok";
    case FRI_QUALITY_PERFECT:
      return "perfect";
    default:
      return "ERROR";
  }
}

/*
 * Synchronize FRI commanded position/configuration with the measured one. This
 * is required when switching between position and cartesian control schemes. If
 * not in FRI monitor state, this can cause a FRI interpolation error.
 */
void syncFriWithMeasuredPose(friRemote * friInst, LWR_CTRL_SCHEME_ENUM newScheme)
{
  static int first = 0;


    if (first < 2) {
      //printf("******************************* FIRST ************************\n");
      if (newScheme == LWR_CTRL_CART_IMP) {
	friInst->doCartesianImpedanceControl(friInst->getMsrCartPosition(), NULL, NULL, NULL, NULL, false);
      } else {
	friInst->doPositionControl(friInst->getMsrMsrJntPosition(), false);
      }
      first++;// = false;
      return;
    }
    LWR_CTRL_SCHEME_ENUM currentScheme = (LWR_CTRL_SCHEME_ENUM)friInst->getCurrentControlScheme();

    if (newScheme != currentScheme) {
      if ((newScheme == LWR_CTRL_POSITION) || (newScheme == LWR_CTRL_JOINT_IMP)) {
	friInst->doPositionControl(friInst->getMsrMsrJntPosition(), false); 
      }
      else
	if (newScheme == LWR_CTRL_CART_IMP) {
	  friInst->doCartesianImpedanceControl(friInst->getMsrCartPosition(),
					       NULL, NULL, NULL, NULL, false);
	}
    }
    else
      /*
	if ((ctrlScheme == LWR_CTRL_GRAV_COMP) && (friInst->getCurrentControlScheme() == LWR_CTRL_CART_IMP))    {
	friInst->doCartesianImpedanceControl(friInst->getMsrCartPosition(), NULL, NULL, NULL, NULL, false);
	printf("*******************************CARTESIAN************************\n");}
	else
	if ((ctrlScheme == LWR_CTRL_GRAV_COMP) &&
        ((friInst->getCurrentControlScheme() == LWR_CTRL_POSITION) ||
	(friInst->getCurrentControlScheme() == LWR_CTRL_JOINT_IMP)))         {
	friInst->doPositionControl(friInst->getMsrMsrJntPosition(), false);
	printf("*******************************POSITION************************\n");}
	else
      */
      // TODO esto nuevo por si solo podria valer.....
      if (((newScheme == LWR_CTRL_POSITION) || (newScheme == LWR_CTRL_JOINT_IMP)) && (jointMaxDiff(friInst) > 0.0001)) {
	friInst->doPositionControl(friInst->getMsrMsrJntPosition(), false);
	//printf("*******************************POSITION 2 %f ************************\n", jointMaxDiff());
      }
      else {
	if ((newScheme == LWR_CTRL_CART_IMP) && (cartMaxDiff(friInst) > 0.0001)) {
	  friInst->doCartesianImpedanceControl(friInst->getMsrCartPosition(), NULL, NULL, NULL, NULL, false);
	  //	printf("*******************************CARTESIAN 2 %f ************************\n",cartMaxDiff());
	}
      }

  return;
}

float cartMaxDiff(friRemote * friInst)
{
  float maxDiff = 0.0;
  float cmd[FRI_CART_FRM_DIM];

  float* msr = friInst->getMsrCartPosition();
//  float* cmd = friInst->getMsrCmdCartPosition();
  friInst->getCurrentCmdCartPosition(cmd);

  for (int i = 0; i < FRI_CART_FRM_DIM; i++) {
//printf("%f / %f,  ", msr[i],cmd[i]);
    float diff = fabs(msr[i] - cmd[i]);
    if (diff > maxDiff)
      maxDiff = diff;
  }
//printf("  ### %f ###   ", maxDiff);
  return maxDiff;
}

float jointMaxDiff(friRemote * friInst)
{
  float maxDiff = 0.0;

  float* msr = friInst->getMsrMsrJntPosition();
  float* cmd = friInst->getMsrCmdJntPosition();
  //friInst->getCurrentCmdJntPosition(cmd);

  for (int i = 0; i < LWR_JOINTS; i++) {
    float diff = fabs(msr[i] - cmd[i]);
    if (diff > maxDiff)
      maxDiff = diff;
  }
  return maxDiff;
}



#endif
