/*
 * Copyright (c) 2010 LAAS/CNRS                       --  Oct 2010
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
/*------------------------------------------------------------
 * 
 *                     --  Module  LWR  --
 * 
 *  Description:
 *  Creation date : Wed May 19 22:35:29 2010 
 *  Author: Xavier Broquere
 *  LAAS-CNRS, Toulouse FRANCE
 *------------------------------------------------------------*/


#ifndef _LWRFUNCTIONS_H
#define _LWRFUNCTIONS_H

#include "softMotion/softMotion.h"

#include "lwrFunctionsStruct.h"

/* Some generic functions */
double deg2rad(double deg);
double rad2deg(double rad);

// Rotation range for each joint; required by approachingBounds function
static const double LWR_JOINT_RANGE_MIN[] = { LWR_A1_RANGE_MIN,   /* q1: -170.0 degrees */
                                              LWR_A2_RANGE_MIN,   /* q2: -120.0 degrees */
                                              LWR_E1_RANGE_MIN,   /* q3: -170.0 degrees */
                                              LWR_A3_RANGE_MIN,   /* q4: -120.0 degrees */
                                              LWR_A4_RANGE_MIN,   /* q5: -170.0 degrees */
                                              LWR_A5_RANGE_MIN,   /* q6: -120.0 degrees */
                                              LWR_A6_RANGE_MIN    /* q7: -170.0 degrees */
                                            };

static const double LWR_JOINT_RANGE_MAX[] = { LWR_A1_RANGE_MAX,   /* q1: +170.0 degrees */
                                              LWR_A2_RANGE_MAX,   /* q2: +120.0 degrees */
                                              LWR_E1_RANGE_MAX,   /* q3: +170.0 degrees */
                                              LWR_A3_RANGE_MAX,   /* q4: +120.0 degrees */
                                              LWR_A4_RANGE_MAX,   /* q5: +170.0 degrees */
                                              LWR_A5_RANGE_MAX,   /* q6: +120.0 degrees */
                                              LWR_A6_RANGE_MAX    /* q7: +170.0 degrees */
                                            };

void  fri2th_cartPos(float cartPos[LWR_CART_FRM_DIM], Gb_th* th);
Gb_th fri2dep_cartPos(float cartPos[LWR_CART_FRM_DIM], Gb_dep* dep);
void  th2fri_cartPos(Gb_th* th, float cartPos[LWR_CART_FRM_DIM]);
Gb_th dep2fri_cartPos(Gb_dep* dep, float cartPos[LWR_CART_FRM_DIM]);

/*
 * Setup velocity, acceleration and jerk limits for both per-joint and
 * cartesian movements. fullSpeedPercent Varies from 0 (no movement) and
 * 1 (movement at the maximum allowed velocity, acceleration and jerk)
 */
void setupMaxVAJ(double fullSpeedPercent,
                 double* q_V_max, double* q_A_max, double* q_J_max,
                 double* x_V_max, double* x_A_max, double* x_J_max);

/*
 * Compute for a given movement: sample time, total time, samples number
 * and time evolution.
 */
void computeTimming(double sampleTime, double totalTime, LWR_MOTION_STR* motion);

bool approachingBounds(int joint, double old_q, double new_q);

double normalize(double   x, double   y, double   z,
                 double* nx, double* ny, double* nz);

void setMotionError(LWR_MOTION_ERROR_ENUM error);
void setMotionStart(LWR_MOTION_TYPE_ENUM type);
void setMotionEnd();
void fillMotionReport(LWR_MOTION_ERROR_ENUM error, int *report);

const char * friStateStr(int state);
const char * friSchemeStr(int scheme);
const char * friQualityStr(int quality);


int lwrGetCondFromGENMANIP_MOTION_STR(GENMANIP_MOTION_STR *curState, std::vector<SM_COND> &cond);

int lwrOpenLoopCntrl(LWR_ARM_INST* armInst, GENMANIP_MOTION_STR *curState, LWR_MOTION_STR *motion, float *commandJntVals);
int lwrOpenLoopTrackQCntrl(LWR_ARM_INST* armInst, GENMANIP_MOTION_STR *curState, LWR_MOTION_STR *motion, float *commandJntVals);
int lwrTrajectoryCntrl(int offsetArm,  LWR_ARM_INST* armInst, GENMANIP_MOTION_STR *curState, LWR_MOTION_STR *motion, float *commandJntVals);
int lwrGoalCntrl(int offsetArm, GENMANIP_MOTION_STR *curState, LWR_MOTION_STR *motion, float *commandJntVals);
int lwrJntTrqCntrl(GEN_BOOL isTrqCntrlSofMotion, int offsetArm,  LWR_ARM_INST* armInst, GENMANIP_MOTION_STR *curState, LWR_MOTION_STR *motion, float *commandJntVals);

int lwrComputeSmoothedTimeScale(double timeScale, SM_LIMITS limitsGoto, SM_COND* timeScaleCond);

#ifdef LWR_FRI
/*
 * Setup transformation matrix from reference system 7 to R tool. The second
 * parameter alows projecting the origin of R tool outside the physical tool
 * providing pivoting movements around a target object.
 */
void setupTh7Tool(friRemote * friInst, LWR_TOOL_ENUM tool, double distance = -1,  Gb_th *th_7_tool = NULL,  Gb_th *th_8_tool = NULL);
void setupThBase0(friRemote * friInst, LWR_BASE_ENUM baseth, Gb_th *th_base_0 = NULL);
int lwrCommandArm(friRemote *friInst, GEN_BOOL simulation, GEN_BOOL isTrqCntrlSofMotion, LWR_ARM_ID_ENUM arm_id, LWR_ARM_INST* armInst, GENMANIP_MOTION_STR *curState, LWR_MOTION_STR *motion);
int lwrUpdateArmStatus(friRemote *friInst, LWR_ARM_INST* arm, POM_SE_POSTER *pomArmPos, char * pomPosterName);


float cartMaxDiff(friRemote * friInst);
float jointMaxDiff(friRemote * friInst);
void switchCtrlScheme(friRemote * friInst, LWR_CTRL_SCHEME_ENUM scheme);
/*
 * Synchronize FRI commanded position/configuration with the measured one. This
 * is required when switching between position and cartesian control schemes. If
 * not in FRI monitor state, this can cause a FRI interpolation error.
 */
void syncFriWithMeasuredPose(friRemote * friInst, LWR_CTRL_SCHEME_ENUM ctrlScheme = LWR_CTRL_POSITION);

#else
void setupTh7Tool( LWR_TOOL_ENUM tool, double distance = -1,  Gb_th *th_7_tool = NULL,  Gb_th *th_8_tool = NULL);
void setupThBase0( LWR_BASE_ENUM baseth, Gb_th *th_base_0 = NULL);
int lwrCommandArm( GEN_BOOL isTrqCntrlSofMotion, LWR_ARM_ID_ENUM arm_id, LWR_ARM_INST* armInst, GENMANIP_MOTION_STR *curState, LWR_MOTION_STR *motion);
int lwrUpdateArmStatus( LWR_ARM_INST* arm, POM_SE_POSTER *pomArmPos, char * pomPosterName);

#endif

int lwrUpdateMotion();

#endif // _LWRFUNCTIONS_H
